#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/prom.h>
#include <linux/avm_hw_config.h>


/*------------------------------------------------------------------------------------------*\
 * FRITZ!WLAN Repeater 300E (Virian)
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw173[] = {
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw173);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box ATH DB120
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw96[] = {
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw96);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box 6810 LTE (ab C-Karten)
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw180[] = {
    { .name = "gpio-button_wlan",    .value = 4,  .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-button_dect",    .value = 16, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-led1",           .value = 18, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-led2",           .value = 19, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-led3",           .value = 20, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-led4",           .value = 21, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-led5",           .value = 22, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio_avm_dect_reset", .value = 11 },
    { .name = "gpio_avm_dect_rd",    .value = 9  },
    { .name = "gpio_slic_pcm_fs",    .value = 12 },
    { .name = "gpio_slic_pcm_clk",   .value = 13 },
    { .name = "gpio_slic_pcm_do",    .value = 14 },
    { .name = "gpio_slic_pcm_di",    .value = 15 },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw180);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box 6841 LTE
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw184[] = {
    { .name = "gpio-button_wlan",   .value = 4,  .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-button_dect",   .value = 16, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-usb_reset",     .value = 19, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-shift_clk",     .value = 20, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-shift_din",     .value = 22, .param = avm_hw_param_gpio_in_active_low  },
    { .name = "gpio_avm_dect_reset", .value = 21 },
    { .name = "gpio_avm_dect_rd",    .value = 9  },
    { .name = "gpio_slic_pcm_fs",    .value = 12 },
    { .name = "gpio_slic_pcm_clk",   .value = 13 },
    { .name = "gpio_slic_pcm_do",    .value = 14 },
    { .name = "gpio_slic_pcm_di",    .value = 15 },
    { .name = "usb_hub_external_port1", .value = 3  },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw184);

/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box Powerline_546E
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw190[] = {
    { .name = "gpio-button_wlan",   .value = 4,  .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-button_t1",     .value = 16, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio-button_t2",     .value = 17, .param = avm_hw_param_gpio_out_active_low },
    { .name = "gpio_avm_relay",     .value = 12 },
    { .name = "gpio_avm_reset_plc", .value = 13 },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw190);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config_table avm_hw_config_tables[] = {
    { 173, avm_hardware_config_hw173 }, /*--- FRITZ!WLAN Repeater 300E (Virian) ---*/
    {  96, avm_hardware_config_hw96  }, /*--- FRITZ!Box ATH DB120 ---*/
    { 180, avm_hardware_config_hw180 }, /*--- FRITZ!Box 6810 LTE (ab C-Karten) ---*/
    { 184, avm_hardware_config_hw184 }, /*--- FRITZ!Box 6841 LTE ---*/
    { 190, avm_hardware_config_hw190 }, /*--- FRITZ!Box Powerline_546E ---*/
};
EXPORT_SYMBOL(avm_hw_config_tables);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config *avm_current_hw_config = NULL;
EXPORT_SYMBOL(avm_current_hw_config);

struct _avm_hw_config *avm_get_hw_config_table(void)
{
    unsigned int hwrev, i;
    char *s;

    s = prom_getenv("HWRevision");
    if (s) {
        hwrev = simple_strtoul(s, NULL, 10);
    }     
    if((s == NULL) || (hwrev < 0) || (hwrev > 255)) {
        printk("[%s] error: %s HWRevision detected in environment variables: %s\n", __FUNCTION__, (s ? "invalid" : "no"), s);
        BUG_ON(1);
        return NULL;
    }
    for(i = 0; i < sizeof(avm_hw_config_tables)/sizeof(struct _avm_hw_config *); i++) {
        if(avm_hw_config_tables[i].hwrev == hwrev) 
            return avm_hw_config_tables[i].table;
    }
    printk("[%s] error: No hardware configuration defined for HWRevision %d\n", __FUNCTION__, hwrev);
    BUG_ON(1);
    return NULL;
}
EXPORT_SYMBOL(avm_get_hw_config_table);

