#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/resource.h>
#include <linux/percpu.h>
#include <linux/smp.h>

#include <linux/console.h>
#include <asm/time.h>
#include <linux/miscdevice.h>
#include <asm/delay.h>
#include <asm/cevt-r4k.h>

#include <atheros.h>
#include <asm/mach_avm.h>

#define ATH_CPU_DDR_CLOCK_CONTROL	(ATH_PLL_BASE + CPU_DDR_CLOCK_CONTROL_OFFSET)
#define ATH_CPU_PLL_CONFIG		(ATH_PLL_BASE + CPU_PLL_CONFIG_OFFSET)

unsigned int ath_get_clock(enum _avm_clock_id id);
unsigned int cpu_get_clock(void);
unsigned int cpu_set_clock(unsigned int freq);

/*
 * Leute:
 *
 * Diese Tabelle ist ein uebler Hack.
 * Ich habe sie nur gebaut, damit man von der cmd-line lesebare Frequenzen eingeben kann.
 * Die eingegebene Wunschfrequenz wird rueckwaerts gerechnet um auf die korrekten Teilerwerte zu kommen.
 * Da waere jedenfalls wuenschenswert. Es kommen aber Teiler raus die immer in anderen 
 * Frequenzen resultieren -> deshalb diese Tabelle!
 */
struct clk_translation_descr {
    unsigned int target_freq;
    unsigned int pll_freq;
} clk_trans[] = {
    { .target_freq = 560, .pll_freq = 560 },
    { .target_freq = 480, .pll_freq = 500 },
    { .target_freq = 400, .pll_freq = 400 },
    { .target_freq = 320, .pll_freq = 350 },
    { .target_freq = 280, .pll_freq = 300 },
    { .target_freq = 200, .pll_freq = 200 },
    { .target_freq = 80, .pll_freq = 100 },
    { .target_freq = 0, .pll_freq = 0 },
};

static ssize_t ath_clksw_read(struct file *file, char __user *buf, size_t count, loff_t *ppos) {
    count = 0;
    count += sprintf(buf + count, "Quick'n'Dirty hack to switch clock on atheros based cpus.\n");
    count += sprintf(buf + count, "Target clock calc ist not perfect, so you have to use the\n");
    count += sprintf(buf + count, "supported values below.\n");
    count += sprintf(buf + count, "\n");
    count += sprintf(buf + count, "Supported clk settings\n");
    count += sprintf(buf + count, "======================\n");
    count += sprintf(buf + count, "Command: 'echo xxx > clksw'\n");
    {
        int i = 0;

        while(clk_trans[i].target_freq != 0) {
            count += sprintf(buf + count, "    %d MHz\n", clk_trans[i].target_freq);
            i++;
        }
    }
    count += sprintf(buf + count, "\n");
    count += sprintf(buf + count, "!!! If you know what you are doing, you can pass a custom value !!!\n");
    count += sprintf(buf + count, "\n");
    count += sprintf(buf + count, "\n");
    count += sprintf(buf + count, "Current clk settings\n");
    count += sprintf(buf + count, "====================\n");
    count += sprintf(buf + count, "cpu:  % 3d MHz\n", ath_get_clock(avm_clock_id_cpu)/1000000);
    count += sprintf(buf + count, "ddr:  % 3d MHz\n", ath_get_clock(avm_clock_id_ddr)/1000000);
    count += sprintf(buf + count, "ahb:  % 3d MHz\n", ath_get_clock(avm_clock_id_ahb)/1000000);
    count += sprintf(buf + count, "ref:  % 3d MHz\n", ath_get_clock(avm_clock_id_ref)/1000000);
    count += sprintf(buf + count, "uart: % 3d MHz\n", ath_get_clock(avm_clock_id_peripheral)/1000000);


    if(*ppos >= count) {
        return 0;
    }

    *ppos += count;
   
	return count;
}

static ssize_t ath_clksw_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos) {
    char my_buf[3];
    int clock = 0;
    unsigned int set_clock = 0;

    if (copy_from_user(my_buf, buf, 3)) {
        return -EFAULT;
    }

    {
        int i = 0;

        while((my_buf[i] >= '0') && (my_buf[i] <= '9')) {
            clock *= 10;
            clock += (my_buf[i] - '0');

            /*--- printk("clock: %d, i: %d, my_buf[i]: %c\n", clock, i, my_buf[i]); ---*/

            i++;
        }

        i = 0;
        while(clk_trans[i].target_freq != 0) {
            if(clk_trans[i].target_freq == clock) {
                set_clock = clk_trans[i].pll_freq;
            }
            i++;
        }

    }
    /*--- printk("clock: %d\n", clock); ---*/

    if(set_clock == 0) {
        cpu_set_clock(clock * 1000000);
    } else {
        cpu_set_clock(set_clock * 1000000);
    }
    return count;
}
//static ssize_t
//ath_clksw_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
//{
//	struct clock_event_device *cd;
//	unsigned int cpu = smp_processor_id();
//	uint8_t setting;
//	uint32_t val;
//	struct { uint32_t freq, pre, post; } pll[] = {
//		{ 300,	CPU_PLL_CONFIG_NINT_SET(0x18)	|
//			CPU_PLL_CONFIG_REFDIV_SET(1)	|
//			CPU_PLL_CONFIG_RANGE_SET(1)	|
//			CPU_PLL_CONFIG_OUTDIV_SET(1),
//			CPU_PLL_CONFIG_OUTDIV_SET(1)		},
//		{ 400,	CPU_PLL_CONFIG_NINT_SET(32)	|
//			CPU_PLL_CONFIG_REFDIV_SET(1)	|
//			CPU_PLL_CONFIG_RANGE_SET(0)	|
//			CPU_PLL_CONFIG_OUTDIV_SET(1),
//			CPU_PLL_CONFIG_OUTDIV_SET(1)		},
//		{ 500,	CPU_PLL_CONFIG_NINT_SET(20)	|
//			CPU_PLL_CONFIG_REFDIV_SET(1)	|
//			CPU_PLL_CONFIG_RANGE_SET(3)	|
//			CPU_PLL_CONFIG_OUTDIV_SET(1),
//			CPU_PLL_CONFIG_OUTDIV_SET(0)		},
//		{ 600,	CPU_PLL_CONFIG_NINT_SET(24)	|
//			CPU_PLL_CONFIG_REFDIV_SET(1)	|
//			CPU_PLL_CONFIG_RANGE_SET(0)	|
//			CPU_PLL_CONFIG_OUTDIV_SET(1),
//			CPU_PLL_CONFIG_OUTDIV_SET(0)		},
//	};
//
//	if (copy_from_user(&setting, buf, 1)) {
//		return -EFAULT;
//	}
//
//	setting = setting - '0';
//
//	if (setting < 0 ||
//	    setting >= (sizeof(pll) / sizeof(pll[0]))) {
//		return -EINVAL;
//	}
//
//	printk("%s: setting: %d | freq: %d | pre: 0x%x | post: 0x%x\n", __func__, setting,
//        pll[setting].freq,
//		pll[setting].pre, pll[setting].post);
//
//	// bypass for cpu pll
//	val = ath_reg_rd(ATH_CPU_DDR_CLOCK_CONTROL);
//	val &= ~CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_MASK;
//	val |= CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_SET(1);
//	ath_reg_wr(ATH_CPU_DDR_CLOCK_CONTROL, val);
//	udelay(10);
//
//	// pll settings...
//	ath_reg_wr(ATH_CPU_PLL_CONFIG,
//		pll[setting].pre | CPU_PLL_CONFIG_PLLPWD_SET(1));
//	udelay(10);
//
//	// clear pll power
//	val = ath_reg_rd(ATH_CPU_PLL_CONFIG);
//	val &= ~CPU_PLL_CONFIG_PLLPWD_MASK;
//	val |= CPU_PLL_CONFIG_PLLPWD_SET(0);
//	ath_reg_wr(ATH_CPU_PLL_CONFIG, val);
//	udelay(100);
//
//	// reset out div
//	val = ath_reg_rd(ATH_CPU_PLL_CONFIG);
//	val &= ~CPU_PLL_CONFIG_OUTDIV_MASK;
//	val |= pll[setting].post;
//	ath_reg_wr(ATH_CPU_PLL_CONFIG, val);
//	udelay(10);
//
//	// unset bypass for cpu pll
//	val = ath_reg_rd(ATH_CPU_DDR_CLOCK_CONTROL);
//	val &= ~CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_MASK;
//	val |= CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_SET(0);
//	ath_reg_wr(ATH_CPU_DDR_CLOCK_CONTROL, val);
//	udelay(10);
//
//	// reconfigure kernel's notion of time
//	mips_hpt_frequency = pll[setting].freq * 1000000 / 2;
//
//	// see r4k_clockevent_init()
//	cd = &per_cpu(mips_clockevent_device, cpu);
//	cd->mult		= div_sc((unsigned long) mips_hpt_frequency, NSEC_PER_SEC, 32);
//	cd->max_delta_ns	= clockevent_delta2ns(0x7fffffff, cd);
//	cd->min_delta_ns	= clockevent_delta2ns(0x300, cd);
//	printk("%s: mult = %lu\n", __func__, cd->mult);
//
//	return 1;
//}

static struct file_operations ath_clksw_fops = {
	.read		= ath_clksw_read,
	.write		= ath_clksw_write,
};

static struct miscdevice ath_clksw_miscdev = {
	ATH_CLKSW_MINOR,
	"clksw",
	&ath_clksw_fops
};


static int __init
ath_clksw_init(void)
{
	u32	tdata;

	printk("%s: Registering Clock Switch Interface ", __func__);

	if ((tdata = misc_register(&ath_clksw_miscdev))) {
		printk("failed %d\n", tdata);
		return tdata;
	} else {
		printk("success\n");
	}

	return 0;
}
late_initcall(ath_clksw_init);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void ath_internel_set_clk(unsigned int NINT, unsigned int REFDIV, unsigned int RANGE, unsigned int OUTDIV) {
	struct clock_event_device *cd;
	unsigned int cpu = smp_processor_id();
	uint32_t val;

	// bypass for cpu pll
	val = ath_reg_rd(ATH_CPU_DDR_CLOCK_CONTROL);
	val &= ~CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_MASK;
	val |= CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_SET(RANGE);
	ath_reg_wr(ATH_CPU_DDR_CLOCK_CONTROL, val);
	udelay(10);

	// pll settings...
	ath_reg_wr(ATH_CPU_PLL_CONFIG,
        CPU_PLL_CONFIG_NINT_SET(NINT)     |
        CPU_PLL_CONFIG_REFDIV_SET(REFDIV) |
        CPU_PLL_CONFIG_RANGE_SET(1)       |
        CPU_PLL_CONFIG_OUTDIV_SET(OUTDIV)
		| CPU_PLL_CONFIG_PLLPWD_SET(1));
	udelay(10);

	// clear pll power
	val = ath_reg_rd(ATH_CPU_PLL_CONFIG);
	val &= ~CPU_PLL_CONFIG_PLLPWD_MASK;
	val |= CPU_PLL_CONFIG_PLLPWD_SET(0);
	ath_reg_wr(ATH_CPU_PLL_CONFIG, val);
	udelay(100);

	// reset out div
	val = ath_reg_rd(ATH_CPU_PLL_CONFIG);
	val &= ~CPU_PLL_CONFIG_OUTDIV_MASK;
	val |= CPU_PLL_CONFIG_OUTDIV_SET(OUTDIV);
	ath_reg_wr(ATH_CPU_PLL_CONFIG, val);
	udelay(10);

	// unset bypass for cpu pll
	val = ath_reg_rd(ATH_CPU_DDR_CLOCK_CONTROL);
	val &= ~CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_MASK;
	val |= CPU_DDR_CLOCK_CONTROL_CPU_PLL_BYPASS_SET(0);
	ath_reg_wr(ATH_CPU_DDR_CLOCK_CONTROL, val);
	udelay(10);

	// reconfigure kernel's notion of time
    {
        unsigned int freq = cpu_get_clock();
        mips_hpt_frequency = freq * 1000000 / 2;

        // see r4k_clockevent_init()
        cd = &per_cpu(mips_clockevent_device, cpu);
        cd->mult		= div_sc((unsigned long) mips_hpt_frequency, NSEC_PER_SEC, 32);
        cd->max_delta_ns	= clockevent_delta2ns(0x7fffffff, cd);
        cd->min_delta_ns	= clockevent_delta2ns(0x300, cd);
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpu_set_clock(unsigned int freq) {
    volatile unsigned int *pcpu_pll_config = (volatile unsigned int *)KSEG1ADDR(0x18050000);
    unsigned int cpu_pll_config = *pcpu_pll_config;
    unsigned int NINT   = CPU_PLL_CONFIG_NINT_GET(cpu_pll_config);
    unsigned int REFDIV = CPU_PLL_CONFIG_REFDIV_GET(cpu_pll_config);
    unsigned int OUTDIV = CPU_PLL_CONFIG_OUTDIV_GET(cpu_pll_config);
    unsigned int RANGE = CPU_PLL_CONFIG_RANGE_GET(cpu_pll_config);
    unsigned int vcoout;
    unsigned int REFCLK;

	printk("%s: Called with: %d\n", __func__, freq);

	if ((ath_reg_rd(ATH_BOOTSTRAP_REG) & ATH_REF_CLK_40)) {
		REFCLK = (40 * 1000000);
	} else {
		REFCLK = (25 * 1000000);
	}

    // Annahme: NINT ist variable
    vcoout = freq * (1 << OUTDIV);
    NINT = vcoout / (REFCLK / REFDIV);

    ath_internel_set_clk(NINT, REFDIV, RANGE, OUTDIV);

    return cpu_get_clock();
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpu_get_clock(void) {
    volatile unsigned int *pcpu_pll_config = (volatile unsigned int *)KSEG1ADDR(0x18050000);
    unsigned int cpu_pll_config = *pcpu_pll_config;
    unsigned int NINT   = CPU_PLL_CONFIG_NINT_GET(cpu_pll_config);
    unsigned int REFDIV = CPU_PLL_CONFIG_REFDIV_GET(cpu_pll_config);
    unsigned int OUTDIV = CPU_PLL_CONFIG_OUTDIV_GET(cpu_pll_config);
    unsigned int freq, vcoout;
    unsigned int REFCLK;
	if ((ath_reg_rd(ATH_BOOTSTRAP_REG) & ATH_REF_CLK_40)) {
		REFCLK = (40 * 1000000);
	} else {
		REFCLK = (25 * 1000000);
	}
    vcoout =(REFCLK / REFDIV) * NINT;
    freq = vcoout / (1 << OUTDIV);
    /*--- printk("%d %d %d %d freq=%d %x %x%x\n", REFCLK, REFDIV, NINT, OUTDIV, freq, cpu_pll_config, CPU_PLL_CONFIG_NINT_LSB, CPU_PLL_CONFIG_NINT_MASK); ---*/
    return freq;
}
EXPORT_SYMBOL(cpu_get_clock);

/*--------------------------------------------------------------------------------*\
 * todo: dynamische Ermittlung
\*--------------------------------------------------------------------------------*/
unsigned int ath_get_clock(enum _avm_clock_id id) {
    switch(id) {
        case avm_clock_id_cpu:
            return cpu_get_clock();
        case avm_clock_id_peripheral:
            return ath_uart_freq;
        case avm_clock_id_ahb:
            return ath_ahb_freq;
        case avm_clock_id_ddr:
            return ath_ddr_freq;
        case avm_clock_id_ref:
            return ath_ref_freq;
        default:
            break;
    }
    return 0;
}
EXPORT_SYMBOL(ath_get_clock);
