#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/prom.h>
#include <linux/avm_hw_config.h>

/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box WLAN 3370           
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw175[] = {
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw175);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box 6840 LTE                    
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw177[] = {
    { .name = "gpio_avm_dect_reset", .value = 30 },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw177);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box Fon WLAN 7360 SL            
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw181[] = {
    { .name = "gpio_avm_dect_reset", .value = 30 },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw181);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box Fon WLAN 7360               
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw183[] = {
    { .name = "gpio_avm_dect_reset", .value = 30 },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw183);


/*------------------------------------------------------------------------------------------*\
 * FRITZ!Box Fon WLAN 7391               
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config avm_hardware_config_hw185[] = {
    { .name = "gpio-wlan_offload_wasp_reset",  .value = 21,    .param = avm_hw_param_gpio_out_active_low },
    { .name = "wlan_offload_mac",              .value = 5,     .param = avm_hw_param_no_param },
    { .name = "gpio_avm_piglet_noemif_data",   .value = 17 },
    { .name = "gpio_avm_piglet_noemif_done",   .value =  5 },
    { .name = "gpio_avm_piglet_noemif_clk",    .value = 18 },
    { .name = "gpio_avm_piglet_noemif_prog",   .value = 10 },
    { .name = "gpio_avm_dect_reset",           .value = 30 },
    { .name = NULL }
};
EXPORT_SYMBOL(avm_hardware_config_hw185);



/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config_table avm_hw_config_tables[] = {
    { 175, avm_hardware_config_hw175 }, /*--- FRITZ!Box WLAN 3370 ---*/
    { 177, avm_hardware_config_hw177 }, /*--- FRITZ!Box 6840 LTE ---*/
    { 181, avm_hardware_config_hw181 }, /*--- FRITZ!Box Fon WLAN 7360 SL ---*/
    { 183, avm_hardware_config_hw183 }, /*--- FRITZ!Box Fon WLAN 7360 ---*/
    { 185, avm_hardware_config_hw185 }, /*--- FRITZ!Box Fon WLAN 7391 ---*/
};
EXPORT_SYMBOL(avm_hw_config_tables);



/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct _avm_hw_config *avm_current_hw_config = NULL;
EXPORT_SYMBOL(avm_current_hw_config);

struct _avm_hw_config *avm_get_hw_config_table(void)
{
    unsigned int hwrev, i;
    char *s;

    s = prom_getenv("HWRevision");
    if (s) {
        hwrev = simple_strtoul(s, NULL, 10);
    }     
    if((s == NULL) || (hwrev < 0) || (hwrev > 255)) {
        printk("[%s] error: %s HWRevision detected in environment variables: %s\n", __FUNCTION__, (s ? "invalid" : "no"), s);
        BUG_ON(1);
        return NULL;
    }
    for(i = 0; i < sizeof(avm_hw_config_tables)/sizeof(struct _avm_hw_config *); i++) {
        if(avm_hw_config_tables[i].hwrev == hwrev) 
            return avm_hw_config_tables[i].table;
    }
    printk("[%s] error: No hardware configuration defined for HWRevision %d\n", __FUNCTION__, hwrev);
    BUG_ON(1);
    return NULL;
}
EXPORT_SYMBOL(avm_get_hw_config_table);

