/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#if defined(CONFIG_AR9) || defined(CONFIG_VR9)

#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/sched.h>
#include "cpmac_if.h"
#include "cpmac_main.h"
#include "cpmac_ar9_if.h"
#include "cpphy_types.h"
#include "cpmac_debug.h"
#include "cpphy_ar8216.h"
#include "cpphy_mdio.h"
#include "cpphy_if.h"
#include "cpmac_eth.h"
#include <asm/mach_avm.h>
#include "cpphy_mgmt.h"
#include "ifx_ar9/ifxmips_3port_sw_reg.h"
/*--- #include <mach/hw_gpio.h> ---*/

/* TODO list */
/* - Get MAC address from environment and assign it to the PHY (from cpphy_if_init) */
 

/*------------------------------------------------------------------------------------------*\
 * Global resources
\*------------------------------------------------------------------------------------------*/
void *ar9_phy_dev = NULL;
extern cpmac_global_t cpmac_global;


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mdio_user_access_write(unsigned short regadr,
                                  unsigned short phyadr,
                                  unsigned short data) {
#   if defined(CONFIG_AR9)
    ar9_switch_write_mdio(phyadr, regadr, data);
#   elif defined(CONFIG_VR9) /*--- #if defined(CONFIG_AR9) ---*/
    vr9_switch_write_mdio(phyadr, regadr, data);
#   endif /*--- #elif defined(CONFIG_VR9) ---*/
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpphy_mdio_user_access_read(unsigned short regadr, unsigned short phyadr) {
    unsigned int result;

#   if defined(CONFIG_AR9)
    result = ar9_switch_read_mdio(phyadr, regadr);
#   elif defined(CONFIG_VR9) /*--- #if defined(CONFIG_AR9) ---*/
    result = vr9_switch_read_mdio(phyadr, regadr);
#   endif /*--- #elif defined(CONFIG_VR9) ---*/
    return result;
}


/*------------------------------------------------------------------------------------------*\
 * FIXME *
\*------------------------------------------------------------------------------------------*/
void vr9_switch_write(unsigned int uRegAddr __attribute__ ((unused)),
                      unsigned int data __attribute__ ((unused))) {
    return;
}
unsigned int vr9_switch_read(unsigned int uRegAddr __attribute__ ((unused))) {
    return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int cpmac_ar9_set_switch_register(unsigned int addr, unsigned int dat) {
#   if defined(CONFIG_AR9)
    ar9_switch_write(addr, dat);
#   elif defined(CONFIG_VR9) /*--- #if defined(CONFIG_AR9) ---*/
    vr9_switch_write(addr, dat);
#   endif /*--- #elif defined(CONFIG_VR9) ---*/
    return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpmac_ar9_get_switch_register(unsigned int addr) {
#   if defined(CONFIG_AR9)
    return ar9_switch_read(addr);
#   elif defined(CONFIG_VR9) /*--- #if defined(CONFIG_AR9) ---*/
    return vr9_switch_read(addr);
#   endif /*--- #elif defined(CONFIG_VR9) ---*/
}


/*------------------------------------------------------------------------------------------*\
 * TODO: The receive functions here and in cpmac_if.c should be unified
\*------------------------------------------------------------------------------------------*/
void cpmac_ar9_if_rx(struct sk_buff *skb) {
    cpphy_mdio_t *mdio = &cpmac_global.instance[0].mdio;

    if(mdio->switch_config.is_being_configured) {
        dev_kfree_skb_any(skb);
        /* TODO Count the dropped packets */
        return;
    }
    cpmac_if_data_from_phy(cpmac_global.instance[0].cpmac_priv, skb, skb->len);
}
EXPORT_SYMBOL(cpmac_ar9_if_rx);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if defined(CONFIG_AR9)
void cpphy_mdio_init_ar9(phy_number_t phy_number __attribute__ ((unused))) {
    unsigned char port;

    DEB_INFOTRC("[%s] Setting default switch register values\n", __FUNCTION__);
    cpmac_ar9_set_switch_register(0x4   , 0x00000001);
    cpmac_ar9_set_switch_register(0x8   , 0x00000001);
    cpmac_ar9_set_switch_register(0xC   , 0x00000001);
    cpmac_ar9_set_switch_register(0x10  , 0x04070001);
    cpmac_ar9_set_switch_register(0x14  , 0x04070001);
    cpmac_ar9_set_switch_register(0x18  , 0x04070001);
    cpmac_ar9_set_switch_register(0x20  , 0x00001BE8);
    cpmac_ar9_set_switch_register(0x24  , 0x00001BE8);
    cpmac_ar9_set_switch_register(0x28  , 0x00001BE8);
    cpmac_ar9_set_switch_register(0x2C  , 0x07070707);
    cpmac_ar9_set_switch_register(0x30  , 0x00000000);
    cpmac_ar9_set_switch_register(0x34  , 0x00000000);
    cpmac_ar9_set_switch_register(0x38  , 0x00080004);
    cpmac_ar9_set_switch_register(0x3C  , 0x00020001);
    cpmac_ar9_set_switch_register(0x40  , 0x00000000);
    cpmac_ar9_set_switch_register(0x44  , 0x00000000);
    cpmac_ar9_set_switch_register(0x48  , 0x00080004);
    cpmac_ar9_set_switch_register(0x4C  , 0x00020001);
    cpmac_ar9_set_switch_register(0x50  , 0x00000000);
    cpmac_ar9_set_switch_register(0x54  , 0x00000000);
    cpmac_ar9_set_switch_register(0x58  , 0x00080004);
    cpmac_ar9_set_switch_register(0x5C  , 0x00020001);
    cpmac_ar9_set_switch_register(0x60  , 0x00000000);
    cpmac_ar9_set_switch_register(0x68  , 0x80080800);
    cpmac_ar9_set_switch_register(0x6C  , 0x001F0000);
    cpmac_ar9_set_switch_register(0x70  , 0x00000000);
    cpmac_ar9_set_switch_register(0x74  , 0x00000000);
    cpmac_ar9_set_switch_register(0x78  , 0x0C002C0B);
    cpmac_ar9_set_switch_register(0x7C  , 0x0000FA50);
    cpmac_ar9_set_switch_register(0x80  , 0x00001000);
    cpmac_ar9_set_switch_register(0x84  , 0x00001800);
    cpmac_ar9_set_switch_register(0x88  , 0x00000000);
    cpmac_ar9_set_switch_register(0x8C  , 0x00040000);
    cpmac_ar9_set_switch_register(0x90  , 0x00000000);
    cpmac_ar9_set_switch_register(0x94  , 0x00000000);
    cpmac_ar9_set_switch_register(0x98  , 0x00000000);
    cpmac_ar9_set_switch_register(0x9C  , 0x00000000);
    cpmac_ar9_set_switch_register(0xA0  , 0x00000000);
    cpmac_ar9_set_switch_register(0xA4  , 0x00000050);
    cpmac_ar9_set_switch_register(0xA8  , 0x0000013B); /* Lantiq-DMA-Fix */
    cpmac_ar9_set_switch_register(0xD4  , 0x00000000);
    cpmac_ar9_set_switch_register(0xD8  , 0x00000000);
    cpmac_ar9_set_switch_register(0x100 , 0x00000000);
    cpmac_ar9_set_switch_register(0x104 , 0x00000000);
    cpmac_ar9_set_switch_register(0x108 , 0x00000000);
    cpmac_ar9_set_switch_register(0x10C , 0x00000000);
    cpmac_ar9_set_switch_register(0x110 , 0x00000000);
    cpmac_ar9_set_switch_register(0x114 , 0x00000000);
    cpmac_ar9_set_switch_register(0x118 , 0x00000000);
    cpmac_ar9_set_switch_register(0x11C , 0x00000000);
    cpmac_ar9_set_switch_register(0x120 , 0x00000000);
    cpmac_ar9_set_switch_register(0x124 , 0x00000000);
    cpmac_ar9_set_switch_register(0x128 , 0x00000000);
    cpmac_ar9_set_switch_register(0x12C , 0x00000000);
    cpmac_ar9_set_switch_register(0x130 , 0x00000000);
    cpmac_ar9_set_switch_register(0x134 , 0x00000000);
    cpmac_ar9_set_switch_register(0x138 , 0x00000000);
    cpmac_ar9_set_switch_register(0x13C , 0x00000000);
    cpmac_ar9_set_switch_register(0x140 , 0x00000000);
    cpmac_ar9_set_switch_register(0x144 , 0x00000000);
    cpmac_ar9_set_switch_register(0x148 , 0x00000000);
    cpmac_ar9_set_switch_register(0x14C , 0x00000000);
    cpmac_ar9_set_switch_register(0x150 , 0x00000000);
    cpmac_ar9_set_switch_register(0x154 , 0x00000000);
    cpmac_ar9_set_switch_register(0x158 , 0x00000000);
    cpmac_ar9_set_switch_register(0x15C , 0x00000000);
    cpmac_ar9_set_switch_register(0x160 , 0x00000000);
    cpmac_ar9_set_switch_register(0x164 , 0x00000000);
    cpmac_ar9_set_switch_register(0x168 , 0x00000000);
    cpmac_ar9_set_switch_register(0x16C , 0x00000000);
    cpmac_ar9_set_switch_register(0x170 , 0x00000000);
    cpmac_ar9_set_switch_register(0x174 , 0x00000000);
    cpmac_ar9_set_switch_register(0x178 , 0x00000000);
    cpmac_ar9_set_switch_register(0x17C , 0x00000000);
    cpmac_ar9_set_switch_register(0x180 , 0x00000000);
    cpmac_ar9_set_switch_register(0x184 , 0x00000000);
    cpmac_ar9_set_switch_register(0x188 , 0x00000000);
    cpmac_ar9_set_switch_register(0x18C , 0x00000000);
    cpmac_ar9_set_switch_register(0x190 , 0x00000000);
    cpmac_ar9_set_switch_register(0x194 , 0x00000000);
    cpmac_ar9_set_switch_register(0x198 , 0x00000000);
    cpmac_ar9_set_switch_register(0x19C , 0x00000000);
    cpmac_ar9_set_switch_register(0x1A0 , 0x00000000);
    cpmac_ar9_set_switch_register(0x1A4 , 0x00000000);
    cpmac_ar9_set_switch_register(0x1A8 , 0x00000000);
    cpmac_ar9_set_switch_register(0x1AC , 0x00000000);
    cpmac_ar9_set_switch_register(0x1B0 , 0x00000000);
    cpmac_ar9_set_switch_register(0x1B4 , 0x00000000);

    for(port = 0; port < 2; port++) {
        unsigned int RegData;
        unsigned int mode = cpmac_global.instance[0].mii_mode[port];
        switch(port) {
            case 0:
                RegData = cpmac_ar9_get_switch_register(0x78);
                RegData &= ~(  RGMII_CTL_P0CKIO 
                             | RGMII_CTL_P0IS_MASK
                             | RGMII_CTL_P0TDLY_MASK
                             | RGMII_CTL_P0RDLY_MASK
                             | RGMII_CTL_P0SPD_MASK
                             | RGMII_CTL_P0DUP);
                RegData |= mode << 8;
                cpmac_ar9_set_switch_register(0x78, RegData);
                DEB_TRC("[%s] Configure MII %u as %s (%#x)\n", 
                        __FUNCTION__,
                        port,
                        (mode == AR9_MODE_RGMII) ? "RGMII" :  
                        (mode == AR9_MODE_MII) ? "MII" :  
                        (mode == AR9_MODE_REV_MII) ? "REV_MII" :  
                        (mode == AR9_MODE_RED_MII) ? "RED_MII" : "unknown!",
                        RegData);

                /*--- Force link up (no VLAN): ---*/
                RegData = cpmac_ar9_get_switch_register(0x4);
                RegData |= 0x1 << 18; /* Force MAC link up */
                cpmac_ar9_set_switch_register(0x4, RegData);
                /*--- REG32(AMAZON_S_SW_P0_VLAN) &= ~(0x7 << 16); ---*/
                /*--- REG32(AMAZON_S_SW_P0_VLAN) |=  ((0x1|0x4) << 16); ---*/
                break;
            case 1:
                RegData = cpmac_ar9_get_switch_register(0x78);
                RegData &= ~(  RGMII_CTL_P1CKIO 
                             | RGMII_CTL_P1IS_MASK
                             | RGMII_CTL_P1TDLY_MASK
                             | RGMII_CTL_P1RDLY_MASK
                             | RGMII_CTL_P1SPD_MASK
                             | RGMII_CTL_P1DUP);
                RegData |= mode << 18;
                cpmac_ar9_set_switch_register(0x78, RegData);
                DEB_TRC("[%s] Configure MII %u as %s (%#x)\n", 
                        __FUNCTION__,
                        port,
                        (mode == AR9_MODE_RGMII) ? "RGMII" :  
                        (mode == AR9_MODE_MII) ? "MII" :  
                        (mode == AR9_MODE_REV_MII) ? "REV_MII" :  
                        (mode == AR9_MODE_RED_MII) ? "RED_MII" : "unknown!",
                        RegData);


                /*--- Force link up (no VLAN): ---*/
                RegData = cpmac_ar9_get_switch_register(0x8);
                RegData |= 0x1 << 18; /* Force MAC link up */
                cpmac_ar9_set_switch_register(0x8, RegData);
                /*--- REG32(AMAZON_S_SW_P1_VLAN) &= ~(0x7 << 16); ---*/
                /*--- REG32(AMAZON_S_SW_P1_VLAN) |=  ((0x2|0x4) << 16); ---*/
                break;
        }
    }
    ssleep(1);
}
#endif /*--- #if defined(CONFIG_AR9) ---*/


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_ar9_start(void) {
    cpphy_mgmt_global_power_set(CPPHY_POWER_GLOBAL_ON);
    cpphy_mgmt_work_schedule(CPMAC_WORK_TICK, 0);
    (void) wait_event_interruptible(cpmac_global.power.waitqueue, (cpmac_global.power.global == CPPHY_POWER_GLOBAL_NONE));
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_ar9_set_dev(struct net_device *dev) {
    cpmac_priv_t *cpmac_priv;
    DEB_INFOTRC("[%s]\n", __FUNCTION__);

    cpmac_priv = kzalloc(sizeof(cpmac_priv_t), GFP_KERNEL);
    assert(cpmac_priv); /* Can not work without memory */
    cpmac_global.instance[0].cpmac_priv = cpmac_priv;
    cpmac_global.instance[0].mdio.cpmac_priv = cpmac_priv;
    cpmac_global.instance[0].dev = dev;
    ar9_phy_dev = dev;
    if(cpmac_global.instance[0].is_open == 1) {
        DEB_TEST("[%s] start cpmac\n", __FUNCTION__); /* FIXME */
        cpmac_ar9_start();
    }
}
EXPORT_SYMBOL(cpmac_ar9_set_dev);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_ar9_open(struct net_device *dev) {
    static unsigned int any_opened = 0;

    DEB_INFOTRC("[%s]\n", __FUNCTION__);

    if(strcmp(dev->name, "eth0") == 0) {
        cpmac_global.instance[0].is_open = 1;
    } else if(strcmp(dev->name, "eth1") == 0) {
        cpmac_global.instance[1].is_open = 1;
    } else if(strcmp(dev->name, "eth2") == 0) {
        cpmac_global.instance[2].is_open = 1;
    } else if(strcmp(dev->name, "eth3") == 0) {
        cpmac_global.instance[3].is_open = 1;
    }
#ifdef CONFIG_MII5_PORT_ENABLED 
    else if(strcmp(dev->name, "eth4") == 0) {
        cpmac_global.instance[4].is_open = 1;
    }
#endif // CONFIG_MII5_PORT_ENABLED

    if(any_opened == 0) {
        DEB_TEST("[%s] start cpmac\n", __FUNCTION__); /* FIXME */
        cpphy_mgmt_global_power_set(CPPHY_POWER_GLOBAL_ON);
        cpphy_mgmt_work_schedule(CPMAC_WORK_TICK, 0);
        any_opened = 1;
    }
}
EXPORT_SYMBOL(cpmac_ar9_open);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int cpmac_ar9_ioctl(struct net_device *dev __attribute__ ((unused)),
                    struct ifreq *ifr,
                    int cmd __attribute__ ((unused))) {
    struct avm_cpmac_ioctl_struct ioctl_struct;
    cpmac_err_t err;

    if(copy_from_user(&ioctl_struct, ifr->ifr_data, sizeof(ioctl_struct))) {
        DEB_ERR("[%s] copy from user failed\n", __FUNCTION__);
        return -EFAULT;
    }
    err = cpphy_switch_ioctl(&cpmac_global.instance[0].mdio, &ioctl_struct);
    if(err != CPMAC_ERR_NOERR) {
        return -EFAULT;
    }
    if(copy_to_user(ifr->ifr_data, &ioctl_struct, sizeof(ioctl_struct))) {
        DEB_ERR("[%s] copy to user failed\n", __FUNCTION__);
        return -EFAULT;
    }
    return 0;
}
EXPORT_SYMBOL(cpmac_ar9_ioctl);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_ar9_set_port_speed(unsigned char port,
                              unsigned char speed,
                              unsigned char fullduplex) {
    static unsigned char last_speed[2] = {7, 7};

    if(last_speed[port] != speed) {
        unsigned int port_control = cpmac_ar9_get_switch_register(0x78);

        if(port == 0) {
            port_control &= ~(0xf << 0);
            port_control |= ((speed << 2) | (fullduplex << 1) | 1) << 0;
        } else {
            port_control &= ~(0xf << 10);
            port_control |= ((speed << 2) | (fullduplex << 1) | 1) << 10;
        }
        DEB_TRC("[%s] Set port %u speed to %s %s (reg = %#x)\n", 
                __FUNCTION__, 
                port, 
                  (speed == 2) ? "1 GBit/s"
                : (speed == 1) ? "100 MBit/s"
                : "10 MBit/s",
                fullduplex ? "FD" : "HD",
                port_control);
        cpmac_ar9_set_switch_register(0x78, port_control);
        last_speed[port] = speed;
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mdio_check_link_arx188(phy_number_t phy_number __attribute__ ((unused))) { /* FIXME */
    unsigned char port;

    /*--- DEB_DEBUG("[%s] Check link ports (PHY nummer %u)\n", __FUNCTION__, mdio->phy_num); ---*/

    for(port = 0; port < cpmac_global.ports; port++) {
        if(!cpmac_global.power.port_on[port]) {
            /* Without power, the link information would be wrong */
            DEB_TRC("[%s] Port %u is powered down\n", __FUNCTION__, port);
        } else {
            phy_number = cpmac_global.map_port_to_phy[port];
            if(cpmac_global.phy[phy_number].f->check_link) {
                cpmac_global.phy[phy_number].f->check_link(phy_number);
            }
        }
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_ar9_printk(const char *text __attribute__ ((unused)),
                       unsigned int line __attribute__ ((unused))) {
    DEB_TEST("[%s] %s - %u\n", __FUNCTION__, text, line);
}

#endif /*--- #if defined(CONFIG_AR9) || defined(CONFIG_VR9) ---*/

