/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#include <linux/avm_event.h>
#include "cpmac_debug.h"
#include "cpphy_types.h"
#include "cpphy_mdio.h"
#include "cpmac_ar9_if.h"
#include "cpmac_interface.h"
#include "cpphy_mgmt.h"
#include "mii_reg.h"

#if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE)
#include <linux/avm_led_event.h>
#endif /*--- #if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE) ---*/
#if defined(CONFIG_AVM_LED)
#include <linux/avm_led.h>
#endif /*--- #if defined(CONFIG_AVM_LED) ---*/


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_AR8030(unsigned char phy_number) {
#   if defined(CONFIG_AR9)
    unsigned int value, shift;
    value = (unsigned int) *((volatile unsigned int *) 0xbf20302c);
    shift = (phy_number == 1) ? 9 : 18;
    value &= ~(0x7 << shift);
    value |= 0x5 << shift;
    *((volatile unsigned int *) 0xbf20302c) = (volatile unsigned int) value;
#   endif /*--- #if defined(CONFIG_AR9) ---*/
    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;
    DEB_INFOTRC("[%s] PHY %u, mdio address %u\n", __FUNCTION__, phy_number, 
                cpmac_global.phy[phy_number].mdio_address);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_AR8035(unsigned char phy_number) {
#   if defined(CONFIG_AR9)
    unsigned int value, shift;
    value = (unsigned int) *((volatile unsigned int *) 0xbf20302c);
    shift = (phy_number == 1) ? 9 : 18;
    value &= ~(0x7 << shift);
    value |= 0x5 << shift;
    *((volatile unsigned int *) 0xbf20302c) = (volatile unsigned int) value;
#   endif /*--- #if defined(CONFIG_AR9) ---*/
    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;
    DEB_INFOTRC("[%s] PHY %u, mdio address %u\n", __FUNCTION__, phy_number, 
                cpmac_global.phy[phy_number].mdio_address);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_link_status_11G(unsigned char  mdio_address,
                                unsigned char *link,
                                unsigned char *speed,
                                unsigned char *fullduplex) {
    DEB_DEBUG("[%s]\n", __FUNCTION__);

    *link = (mdio_read(mdio_address, MII_REG_STATUS) & MII_REG_STATUS_LINK) ? 1 : 0;
    if(*link == 1) {
        unsigned int status = mdio_read(mdio_address, 0x18);
        *speed      = status & 0x3;
        *fullduplex = (status & 0x8) ? 1 : 0;
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_link_status_AR8035(unsigned char  mdio_address,
                                   unsigned char *link,
                                   unsigned char *speed,
                                   unsigned char *fullduplex) {
    unsigned int status = 0;

    DEB_DEBUG("[%s]\n", __FUNCTION__);

    status = mdio_read(mdio_address, 0x11);
    *link       = (status & (0x1 << 10)) ? 1 : 0;
    *speed      = (status >> 14) & 0x3;
    *fullduplex = (status & (0x1 << 13)) ? 1 : 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_link_status_vitesse(unsigned char  mdio_address,
                                    unsigned char *link,
                                    unsigned char *speed,
                                    unsigned char *fullduplex) {
    unsigned int status;

    DEB_DEBUG("[%s]\n", __FUNCTION__);

    *link = (mdio_read(mdio_address, MII_REG_STATUS) & MII_REG_STATUS_LINK) ? 1 : 0;

    if(*link) {
        status = mdio_read(mdio_address, 0x1c);
        *speed = (status >> 3) & 0x3;
        *fullduplex = (status >> 5) & 0x1;
        if(*speed == avm_event_ethernet_speed_10M) { /* FIXME */
            DEB_WARN("[%s] 10 MBit/s does not work yet!\n", __FUNCTION__);
        }
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_check_link_PHY(phy_number_t phy_number) {
    cpmac_phy_t *phy = &cpmac_global.phy[phy_number];
    unsigned char port = phy->port_offset;
    enum _avm_event_ethernet_speed speed = avm_event_ethernet_speed_no_link;
    unsigned char link = 0, fullduplex = 0;

    DEB_TRC("[%s] Check link port %u (PHY number %u)\n", __FUNCTION__, port, phy_number);

    if(!cpmac_global.power.port_on[port]) {
        /* Without power, the link information would be wrong */
        DEB_TRC("[%s] Port %u is powered down\n", __FUNCTION__, port);
    } else {
        unsigned char speedvalue = 0xff;
        /* Check real time link information */
        if(phy->f->link_status) {
            phy->f->link_status(phy->mdio_address, &link, &speedvalue, &fullduplex);
        }
        switch(speedvalue) {
            case 0:  speed = avm_event_ethernet_speed_10M;     break;
            case 1:  speed = avm_event_ethernet_speed_100M;    break;
            case 2:  speed = avm_event_ethernet_speed_1G;      break;
            default: speed = avm_event_ethernet_speed_no_link; break;
        }
#       if defined(CONFIG_AR9)
        cpmac_ar9_set_port_speed(port, speed, fullduplex);
#       endif /*--- #if defined(CONFIG_AR9) ---*/
#       if defined(CONFIG_FUSIV_VX180)
        mdio->cpmac_priv->fusiv.set_port_speed(port, speed, fullduplex);
#       endif /*--- #if defined(CONFIG_FUSIV_VX180) ---*/
    }
    cpmac_interface_port_update(port, link, speed, fullduplex);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_11G(unsigned char phy_number) {
    cpmac_phy_t *phy = &cpmac_global.phy[phy_number];
    /*--- unsigned int tries = 0; ---*/

    DEB_INFOTRC("[%s] PHY %u, mdio address %u\n", __FUNCTION__, phy_number, phy->mdio_address);
    cpmac_global.power.port_on[phy->port_offset] = 1;
#   if 0
    while(1) {
        if(0xffff != mdio_read(phy->mdio_address, MII_REG_PHY_ID1)) {
            break;
        }
        tries++;
        if(tries >= 3) {
            panic("[%s] Could not revive PHY after %u tries!\n", __FUNCTION__, tries);
        }
        DEB_WARN("[%s] Resetting PHY %u again\n", __FUNCTION__, phy_number);
        cpphy_mgmt_reset_hw(phy->reset_bit, 1);
        ssleep(1);
        cpphy_mgmt_reset_hw(phy->reset_bit, 0);
        ssleep(1);
    }

    mdio_write(phy->mdio_address, MII_REG_CONTROL,   MII_REG_CONTROL_RESET 
                                                   | MII_REG_CONTROL_AUTONEG_ENABLE);
    msleep(500);
    mdio_write(phy->mdio_address, 
               MII_REG_AUTONEG_ADV,
                 MII_REG_AUTONEG_ADV_SF(1)
               | MII_REG_AUTONEG_ADV_10BT_HDX
               | MII_REG_AUTONEG_ADV_10BT_FDX
               | MII_REG_AUTONEG_ADV_100BT_HDX
               | MII_REG_AUTONEG_ADV_100BT_FDX
               | MII_REG_AUTONEG_ADV_PAUSE_SYM
               | MII_REG_AUTONEG_ADV_PAUSE_ASYM);
    /* Declare PHY as part of a multi port device. Set MSPT according to Lantiq hint */
    /* (concerning link problems with GBit switches */
    /* Set MSPT according to Lantiq hint (concerning link problems with GBit switches */
    /* and activate GBit FD and HD for now */
    mdio_write(phy->mdio_address, MII_REG_1000BASET_CTRL,   MII_REG_1000BASET_CTRL_ADV_HD
                                                          | MII_REG_1000BASET_CTRL_ADV_FD
                                                          | MII_REG_1000BASET_CTRL_MSPT);
    /* Enable auto downshift from GBit/s to 100 MBit/s after three autoneg failures */
    mdio_write(cpmac_global.phy[phy_number].mdio_address, 0x14, 0x8006);
    /* Fix against bad links */
    mdio_write(cpmac_global.phy[phy_number].mdio_address, 0x1d, 0x100);
#   endif /*--- #if 0 ---*/
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_phys_init_vitesse(unsigned char phy_number) {
    DEB_INFOTRC("[%s] PHY %u\n", __FUNCTION__, phy_number);
    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;
}


