/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2006,2007, ... ,2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#include <stdarg.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/ioport.h>
#if !defined(CONFIG_NETCHIP_ADM69961)
#define CONFIG_NETCHIP_ADM69961
#endif
#include <linux/avm_event.h>
#if defined(CONFIG_MIPS_OHIO)
#include <asm/mips-boards/ohio.h>
#endif /*--- #if defined(CONFIG_MIPS_OHIO) ---*/
#include <linux/version.h>
#if (KERNEL_VERSION(2,6,28) <= LINUX_VERSION_CODE)
#include <linux/env.h>
#else /*--- #if (KERNEL_VERSION(2,6,28) <= LINUX_VERSION_CODE) ---*/
#include <asm/mips-boards/prom.h>
#endif /*--- #else ---*/ /*--- #if (KERNEL_VERSION(2,6,28) <= LINUX_VERSION_CODE) ---*/
#if defined(CONFIG_ARCH_PUMA5)
#include <mach/hw_gpio.h>
#endif /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/
#if defined(CONFIG_MIPS_UR8)

#if (KERNEL_VERSION(2,6,32) <= LINUX_VERSION_CODE)
#include <ur8.h>
#include <asm/prom.h>
#else
#include <asm/mips-boards/ur8.h>
#endif

#include <asm/mach-ur8/hw_nwss.h>
#endif /*--- #if defined(CONFIG_MIPS_UR8) ---*/
#include <asm/mach_avm.h>

#include "cpmac_if.h"
#include "cpmac_const.h"
#include "cpmac_debug.h"
#include "cpphy_const.h"
#include "cpphy_types.h"
#include "cpphy_main.h"
#include "cpmac_main.h"
#include "cpphy_adm6996.h"
#include "cpphy_ar8216.h"
#include "cpphy_mdio.h"
#include "cpgmac_f.h"
#include "cpmac_product_conf.h"
#include "cpphy_mgmt.h"
#include "cpphy_if.h"
#include "cpmac_ar9_if.h"
#include "cpmac_fusiv_if.h"
#include "cpmac_interface.h"
#include "cpmac_multicast.h"
#include "cpmac_phys.h"
#include "cpmac_proc.h"
#include "cpmac_puma_if.h"
#if defined(CONFIG_AVM_POWER)
#include <linux/avm_power.h>
#endif /*--- #if defined(CONFIG_AVM_POWER) ---*/


MODULE_AUTHOR("Maintainer: AVM GmbH");
MODULE_DESCRIPTION("Ethernet driver for FRITZ!Box");
MODULE_LICENSE("GPL v2");

#if defined(CONFIG_MIPS_AR7)
#error "AR7 not supported anymore!"
#endif /*--- #if defined(CONFIG_MIPS_AR7) ---*/


/*----------------------------------------------------------------------------------*\
 * Zugriff auf cpmac_cpphy_global aus anderen Modulen ueber phy_handle              *
\*----------------------------------------------------------------------------------*/
cpmac_global_t cpmac_global;
extern cpmac_product_struct cpmac_products;


/*------------------------------------------------------------------------------------------*\
 * Global resources                                                                         *
\*------------------------------------------------------------------------------------------*/
#if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_FUSIV_VX180) || defined(CONFIG_ARCH_PUMA5)
static struct resource cpmac_resource_reset_gpio_phy0 = {
    .name = "CPMAC reset phy 0",
    .flags = IORESOURCE_IO		
};

static struct resource cpmac_resource_reset_gpio_phy1 = {
    .name = "CPMAC reset phy 1",
    .flags = IORESOURCE_IO		
};
#endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_FUSIV_VX180) || defined(CONFIG_ARCH_PUMA5) ---*/


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpphy_entry_invalid_product_warning(void) {
    DEB_ERR("*****************************************************************************\n");
    DEB_ERR("*****************************************************************************\n");
    DEB_ERR("*****************************************************************************\n");
    DEB_ERR("***                                                                       ***\n");
    DEB_ERR("***                                                                       ***\n");
    DEB_ERR("***                                                                       ***\n");
    DEB_ERR("*** ATTENTION! There is no valid ethernet configuration for this product! ***\n");
    DEB_ERR("***                                                                       ***\n");
    DEB_ERR("***                                                                       ***\n");
    DEB_ERR("***                                                                       ***\n");
    DEB_ERR("*****************************************************************************\n");
    DEB_ERR("*****************************************************************************\n");
    DEB_ERR("*****************************************************************************\n");
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_init_power_values(enum _avm_event_ethernet_speed maxspeed) {
    unsigned int power[avm_event_ethernet_speed_items] = {  0, /* no link      */
                                                           50, /* 10 MiBits/s  */
                                                          100, /* 100 MiBits/s */
                                                          550  /* 1 GiBits/s   */
                                                         };
    enum _avm_event_ethernet_speed speed;

    if(cpmac_global.power.maxspeed > maxspeed) {
        return;
    }
    for(speed = avm_event_ethernet_speed_no_link; speed < avm_event_ethernet_speed_items; speed++) {
        cpmac_global.power.usage_percentage[speed] = (unsigned char) (min(100u, (100u * power[speed]) / power[maxspeed]));
    }
    DEB_INFOTRC("[%s] Adjusted power percentages for %s: %u%% - %u%% - %u%% - %u%%\n", __FUNCTION__,
                (maxspeed == avm_event_ethernet_speed_10M) ? "10 MBits" :
                (maxspeed == avm_event_ethernet_speed_100M) ? "100 MBits" :
                (maxspeed == avm_event_ethernet_speed_1G) ? "1 GBits" : "unknown",
                cpmac_global.power.usage_percentage[avm_event_ethernet_speed_1G],
                cpmac_global.power.usage_percentage[avm_event_ethernet_speed_100M],
                cpmac_global.power.usage_percentage[avm_event_ethernet_speed_10M],
                cpmac_global.power.usage_percentage[avm_event_ethernet_speed_no_link]
               );
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_device_name_init(unsigned char *name) {
    unsigned int i;
    
    for(i = 0; i < AVM_CPMAC_MAX_INSTANCES; i++) {
        cpmac_global.device_name_length[i] = sprintf(cpmac_global.device_name[i], "%s%u", name, i);
        DEB_TRC("[%s] %s with length %u defined\n", 
                __FUNCTION__, 
                cpmac_global.device_name[i],
                cpmac_global.device_name_length[i]);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned char *cpmac_device_name_get(unsigned int device_number) {
    assert(device_number < AVM_CPMAC_MAX_INSTANCES);
    return cpmac_global.device_name[device_number];
}
EXPORT_SYMBOL(cpmac_device_name_get);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpmac_device_name_cmp(unsigned char *name, unsigned int device_number) {
    assert(device_number < AVM_CPMAC_MAX_INSTANCES);
    return (strncmp(name, cpmac_global.device_name[device_number], cpmac_global.device_name_length[device_number]) == 0);
}
EXPORT_SYMBOL(cpmac_device_name_cmp);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpmac_get_number_of_instances(void) {
    return cpmac_global.instances;
}
EXPORT_SYMBOL(cpmac_get_number_of_instances);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static unsigned int cpphy_entry_check_for_emv(void) {
    unsigned long long m0, m1, m2, m3, m4, m5, mac;
    unsigned char *macstring = prom_getenv("maca");
    if(macstring == NULL)
        return 0;

    sscanf(macstring, "%llx:%llx:%llx:%llx:%llx:%llx", &m0, &m1, &m2, &m3, &m4, &m5);
    mac =   ((m0 & 0xff) << 40)
          | ((m1 & 0xff) << 32)
          | ((m2 & 0xff) << 24)
          | ((m3 & 0xff) << 16)
          | ((m4 & 0xff) <<  8)
          | ((m5 & 0xff) <<  0);

    return (   0
            || ((mac >= 0x001C4A000000ull) && (mac <= 0x001C4AFFFFFFull))
            || ((mac >= 0x001F3F000000ull) && (mac <= 0x001F3F345DC9ull))
            || ((mac >= 0x001F3F3AE817ull) && (mac <= 0x001F3F6DE11Cull))
            || ((mac >= 0x0024FE000000ull) && (mac <= 0x0024FE1EE1F9ull)) 
           );
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpphy_entry_set_config(unsigned char                instance_number, 
                                   cpmac_product_config_struct *config,
                                   unsigned char                phy_number) {
    cpmac_instance_t *instance = &cpmac_global.instance[instance_number];
    cpphy_mdio_t *mdio = &instance->mdio;
    cpmac_phy_t *phy = &cpmac_global.phy[instance_number];
    unsigned char mode, device, external_ports = 0, port, port_modes_possible = 0;
    enum _avm_event_ethernet_speed max_port_speed = avm_event_ethernet_speed_no_link;

    phy->Mode = 0 | CPPHY_SWITCH_MODE_WRITE | CPPHY_SWITCH_MODE_READ; /* 32 Bit, read, write */

    mdio->switch_config.used_portset = config->phy[phy_number].ports;
    mdio->switch_config.mii_portmask = config->phy[phy_number].mii_portmask;
    mdio->mode_pppoa = 0;

    // TKL: pointer auf urspruengliches config-struct im phy speichern
    phy->config = &(config->phy[phy_number]);

    phy->port_offset = cpmac_global.ports;
    phy->mdio_address = config->phy[phy_number].mdio_address;
    config->phy[phy_number].reset_pin = 0xffff; /* FIXME */
    if(config->phy[phy_number].reset_pin != 0xffff) {
#       if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_FUSIV_VX180) || defined(CONFIG_ARCH_PUMA5)
        struct resource *resource_ptr;
        if(phy_number == 0) {
            resource_ptr = &cpmac_resource_reset_gpio_phy0;
        } else {
            resource_ptr = &cpmac_resource_reset_gpio_phy1;
        }
        resource_ptr->start = resource_ptr->end = config->phy[phy_number].reset_pin;
        if(request_resource(&gpio_resource, resource_ptr)) {
            DEB_ERR("[%s] Error! Could not reserve GPIO %u for reset!\n", __FUNCTION__, config->phy[phy_number].reset_pin);
            return;
        }
#       endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_FUSIV_VX180) || defined(CONFIG_ARCH_PUMA5) ---*/
#       if defined(CONFIG_AR9)
        phy->reset_bit = config->phy[phy_number].reset_pin;
        ar9_setup_ext_phy_reset(phy->reset_bit, 1);
        ifx_gpio_output_clear(phy->reset_bit, IFX_GPIO_MODULE_INTERNAL_SWITCH);
        DEB_INFO("[%s] Put PHY %u into reset\n", __FUNCTION__, phy_number);
#       endif /*--- #if defined(CONFIG_AR9) ---*/
#       if defined(CONFIG_FUSIV_VX180) || defined(CONFIG_ARCH_PUMA5)
        phy->reset_bit = config->phy[phy_number].reset_pin;
        avm_gpio_ctrl(phy->reset_bit, GPIO_PIN, GPIO_OUTPUT_PIN);
        avm_gpio_out_bit(phy->reset_bit, 0);
#       else /*--- #if defined(CONFIG_FUSIV_VX180) ---*/
#       if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_ARCH_PUMA5)
        phy->reset_bit = (unsigned int) avm_register_reset_gpio(config->phy[phy_number].reset_pin, LOW_ACTIVE);
        if(phy->reset_bit == (unsigned int) -1) {
            DEB_ERR("[%s] Error! Could not register GPIO %u for reset!\n", __FUNCTION__, config->phy[phy_number].reset_pin);
            return;
        }
#       endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_ARCH_PUMA5) ---*/
#       endif /*--- #else ---*/ /*--- #if defined(CONFIG_FUSIV_VX180) ---*/
    }

    switch(config->phy[phy_number].type) {
        case CPMAC_PHY_TYPE_NONE:
            phy->Mode           = 0;
            port_modes_possible = 1 << ADM_PHY_POWER_OFF;
            phy->ports          = 0;
            external_ports      = 0;
            break;
#       if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8)
        case CPMAC_PHY_TYPE_ADM6996FC:
            DEB_INFOTRC("[%s] Configure functions for ADM6996xC\n", __FUNCTION__);
            phy->f->check_external_tagging  = cpphy_if_check_external_tagging;
            phy->f->check_link              = adm_6996xc_check_link;
            phy->f->config_ports            = adm_config_ports;
            phy->f->configure_special       = cpphy_switch_configure_special;
            phy->f->config_vlan_group       = adm_adm6996_config_vlan_group;
            phy->f->get_new_vid             = cpphy_adm_get_new_vid;
            phy->f->get_switch_register     = adm_read_32;
            phy->f->init_phy                = adm_adm6996xc_init;
            phy->f->deinit_phy              = 0;
            phy->f->mgmt_check_link         = cpphy_mgmt_check_link_orig;
            phy->f->prepare_reboot          = adm_prepare_reboot;
            phy->f->set_default_vid         = cpphy_adm_set_default_vid;
            phy->f->set_good_vid_bitshift   = adm_set_good_vid_bitshift;
            phy->f->set_port_vid_mapping    = cpphy_adm_set_port_vid_mapping;
            phy->f->set_switch_register     = adm_write_32;
            phy->f->set_wan_keep_tagging    = adm_set_wan_keep_tagging;
            phy->f->switch_dump             = switch_dump_adm6996;
            phy->f->switch_port_power       = adm_switch_port_power_6996;
            phy->f->switch_wan_keep_tagging = adm_switch_wan_keep_tagging_adm6996;
            phy->f->update_hw_status        = cpphy_adm_update_hw_status;
            phy->f->vlan_clear              = adm_vlan_clear;
            phy->ports                               = 6;
            mdio->switch_config.cpu_port             = 5;
            mdio->switch_config.is_switch            = 1;
            mdio->switch_config.ports                = 6;
            mdio->switch_config.external_port_offset = 0;
            mdio->switch_config.max_vid_groups       = 16;
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON)
                                  | (1 << ADM_PHY_POWER_SAVE);
            external_ports = 4;
            max_port_speed = avm_event_ethernet_speed_100M;
            cpmac_init_power_values(avm_event_ethernet_speed_100M);
            break;
        case CPMAC_PHY_TYPE_TANTOS:
            DEB_INFOTRC("[%s] Configure functions for Tantos\n", __FUNCTION__);
            phy->f->add_port_to_wan        = adm_add_port_to_wan;
            phy->f->check_link             = adm_tantos_check_link;
            phy->f->config_ports           = adm_tantos_config_ports;
            phy->f->configure_special      = cpphy_switch_configure_special;
            phy->f->config_vlan_group      = adm_tantos_config_vlan_group;
            phy->f->deinit_phy             = 0;
            phy->f->get_new_vid            = cpphy_tantos_get_new_vid;
            phy->f->get_switch_register    = adm_read_32;
            phy->f->init_phy               = adm_tantos_init;
            phy->f->mgmt_check_link        = cpphy_mgmt_check_link_orig;
            phy->f->set_default_vid        = cpphy_tantos_set_default_vid;
            phy->f->set_switch_register    = adm_write_32;
            phy->f->switch_dump            = switch_dump_tantos;
            phy->f->switch_port_power      = adm_switch_port_power_tantos;
            phy->ports                               = 7;
            mdio->switch_config.cpu_port             = 5;
            mdio->switch_config.is_switch            = 1;
            mdio->switch_config.ports                = 7;
            mdio->switch_config.external_port_offset = 0;
            mdio->switch_config.max_vid_groups       = 16;
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON)
                                  | (1 << ADM_PHY_POWER_SAVE);
            external_ports = 4;
            max_port_speed = avm_event_ethernet_speed_100M;
            cpmac_init_power_values(avm_event_ethernet_speed_100M);
            break;
#       endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) ---*/
        case CPMAC_PHY_TYPE_AR8216:
            DEB_INFOTRC("[%s] Configure functions for AR8216\n", __FUNCTION__);
#           if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8)
            phy->f->set_wan_keep_tagging    = adm_set_wan_keep_tagging;
            phy->f->switch_wan_keep_tagging = adm_switch_wan_keep_tagging_ar8x16;
#           endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) ---*/
            /*--- phy->f->switch_port_power       = ar_switch_port_power_orig; ---*/
            /*--- phy->f->power_on_boot           = ar_switch_port_power_on_boot_orig; ---*/
            phy->f->check_link              = cpmac_global.power.use_EMV ? cpmac_interface_check_link_dummy
                                                                         : ar_check_link_orig;
            phy->f->dump_counters           = ar8216_dump_counters;
            phy->f->init_phy                = ar_ar8216_init;
            phy->f->deinit_phy              = ar_ar8216_deinit;
            phy->f->mgmt_check_link         = cpmac_global.power.use_EMV ? cpphy_mgmt_check_link_emv
                                                                         : cpphy_mgmt_check_link_orig;
            mdio->switch_config.max_vid_groups       = 16;
            port_modes_possible = 1 << ADM_PHY_POWER_ON;
            max_port_speed = avm_event_ethernet_speed_100M;
            cpmac_init_power_values(avm_event_ethernet_speed_100M);
            goto generic_atheros;
        case CPMAC_PHY_TYPE_AR8226:
            DEB_INFOTRC("[%s] Configure functions for AR8226\n", __FUNCTION__);
#           if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8)
            phy->f->set_wan_keep_tagging    = adm_set_wan_keep_tagging;
            phy->f->switch_wan_keep_tagging = adm_switch_wan_keep_tagging_ar8x16;
#           endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) ---*/
            phy->f->check_link             = ar_check_link_orig;
            phy->f->init_phy               = ar_ar8226_init;
            phy->f->deinit_phy             = ar_ar8226_deinit;
            phy->f->mgmt_check_link        = cpphy_mgmt_check_link_orig;
            phy->f->switch_port_power      = ar8316_switch_port_power;
            mdio->switch_config.max_vid_groups       = 16;
            port_modes_possible = 1 << ADM_PHY_POWER_ON;
            max_port_speed = avm_event_ethernet_speed_100M;
            cpmac_init_power_values(avm_event_ethernet_speed_100M);
            goto generic_atheros;
        case CPMAC_PHY_TYPE_AR8316:
            DEB_INFOTRC("[%s] Configure functions for AR8316\n", __FUNCTION__);
            phy->f->check_link             = ar_check_link_orig;
            phy->f->dump_counters          = ar8316_dump_counters;
            phy->f->init_phy               = ar_ar8316_init;
            phy->f->deinit_phy             = ar_ar8316_deinit;
            phy->f->mgmt_check_link        = cpphy_mgmt_check_link_orig;
            phy->f->print_mac_table        = ar8316_print_table;
            phy->f->switch_port_power      = ar8316_switch_port_power;
#           if defined(POWERMANAGEMENT_THROTTLE_ETH)
            phy->f->switch_port_speed_throttle = ar8316_switch_port_speed_throttle;
#           endif /*--- #if defined(POWERMANAGEMENT_THROTTLE_ETH) ---*/
            mdio->switch_config.max_vid_groups = 32;
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON)
                                  | (1 << ADM_PHY_POWER_SAVE);
            max_port_speed = avm_event_ethernet_speed_1G;
            cpmac_init_power_values(avm_event_ethernet_speed_1G);
generic_atheros:
#           if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8)
            phy->f->check_external_tagging = cpphy_if_check_external_tagging;
#           endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) ---*/
            phy->f->ar_work_item            = ar8216_ar_work_item;
            phy->f->config_ports            = ar_config_ports;
            phy->f->configure_special       = cpphy_switch_configure_special;
            phy->f->config_vlan_group       = ar_config_vlan_group;
            phy->f->get_new_vid             = cpphy_ar_get_new_vid;
            phy->f->get_phy_port            = ar8216_get_phy_port;
            phy->f->get_switch_register     = ar8216_mdio_read32;
            phy->f->set_default_vid         = cpphy_ar_set_default_vid;
            phy->f->set_switch_register     = ar8216_mdio_write32;
            phy->f->switch_dump             = switch_dump_atheros;
            phy->f->vlan_clear              = ar_vlan_clear;
            phy->ports                               = 6;
            mdio->switch_config.cpu_port             = 0;
            mdio->switch_config.external_port_offset = 1;
            mdio->switch_config.is_switch            = 1;
            mdio->switch_config.ports                = 6;
            external_ports = 4;
            break;
        case CPMAC_PHY_TYPE_VITESSE:
            DEB_INFOTRC("[%s] Configure functions for VITESSE\n", __FUNCTION__);
#           if defined(CONFIG_FUSIV_VX180)
            cpmac_device_name_init("eth");
#           endif /*--- #if defined(CONFIG_FUSIV_VX180) ---*/
            phy->f->check_link             = cpmac_phys_check_link_PHY;
            phy->f->link_status            = cpmac_phys_link_status_vitesse;
            phy->f->init_phy               = cpmac_phys_init_vitesse;
            phy->f->deinit_phy             = 0;
            phy->f->mgmt_check_link        = cpphy_mgmt_check_link_orig;
            phy->ports                 = 1;
            mdio->switch_config.ports  = 1;
            external_ports = 1;
            max_port_speed = avm_event_ethernet_speed_1G;
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON);
            cpmac_init_power_values(avm_event_ethernet_speed_1G);
            break;
        case CPMAC_PHY_TYPE_11G:
#           if defined(CONFIG_VR9) || defined(CONFIG_MACH_AR7240) || defined(CONFIG_MACH_AR724x)
            cpmac_device_name_init("eth");
#           endif /*--- #if defined(CONFIG_VR9) || defined(CONFIG_MACH_AR7240) || defined(CONFIG_MACH_AR724x) ---*/
            /*--- phy->f->check_link             = cpmac_phys_check_link_PHY; ---*/
            /*--- phy->f->link_status            = cpmac_phys_link_status_11G; ---*/
            phy->f->check_link             = cpmac_interface_check_link_dummy;
            phy->f->init_phy               = cpmac_phys_init_11G;
            phy->f->deinit_phy             = 0;
#           if defined(CONFIG_VR9)
            phy->f->switch_port_power      = ifx_avm_7port_power_ctrl;
#           if defined(POWERMANAGEMENT_THROTTLE_ETH)
            phy->f->switch_port_speed_throttle = ifx_avm_7port_speed_throttle;
#           endif /*--- #if defined(POWERMANAGEMENT_THROTTLE_ETH) ---*/
#           else /*--- #if defined(CONFIG_VR9) ---*/
            phy->f->switch_port_power      = ar8316_switch_port_power;
#           if defined(POWERMANAGEMENT_THROTTLE_ETH)
            phy->f->switch_port_speed_throttle = ar8316_switch_port_speed_throttle;
#           endif /*--- #if defined(POWERMANAGEMENT_THROTTLE_ETH) ---*/
#           endif /*--- #else ---*/ /*--- #if defined(CONFIG_VR9) ---*/
            /*--- port_modes_possible = 1 << ADM_PHY_POWER_ON; ---*/
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_SAVE)
                                  | (1 << ADM_PHY_POWER_ON);
            phy->ports        = 1;
            external_ports = 1;
            max_port_speed = avm_event_ethernet_speed_1G;
#           if defined(CONFIG_AR9)
            cpmac_global.instance[0].mii_mode[phy_number - 1] = AR9_MODE_RGMII;
#           endif /*--- #if defined(CONFIG_AR9) ---*/
            cpmac_init_power_values(avm_event_ethernet_speed_1G);
            break;
        case CPMAC_PHY_TYPE_AR8030:
            phy->f->check_link             = cpmac_phys_check_link_PHY;
            phy->f->link_status            = cpmac_phys_link_status_AR8035;
            phy->f->init_phy               = cpmac_phys_init_AR8030;
            phy->f->deinit_phy             = 0;
            phy->f->switch_port_power      = ar8316_switch_port_power;
            /*--- port_modes_possible = 1 << ADM_PHY_POWER_ON; ---*/
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON)
                                  | (1 << ADM_PHY_POWER_SAVE);
            phy->ports        = 1;
            phy->mdio_address = cpmac_global.ports; /* FIXME Temporary solution */
            external_ports    = 1;
            max_port_speed = avm_event_ethernet_speed_100M;
#           if defined(CONFIG_AR9)
            cpmac_global.instance[0].mii_mode[phy_number - 1] = AR9_MODE_RED_MII;
#           endif /*--- #if defined(CONFIG_AR9) ---*/
            break;
        case CPMAC_PHY_TYPE_AR8035:
            phy->f->check_link             = cpmac_phys_check_link_PHY;
            phy->f->link_status            = cpmac_phys_link_status_AR8035;
            phy->f->init_phy               = cpmac_phys_init_AR8035;
            phy->f->deinit_phy             = 0;
            phy->f->switch_port_power      = ar8316_switch_port_power;
#           if defined(POWERMANAGEMENT_THROTTLE_ETH)
            phy->f->switch_port_speed_throttle = ar8316_switch_port_speed_throttle;
#           endif /*--- #if defined(POWERMANAGEMENT_THROTTLE_ETH) ---*/
            /*--- port_modes_possible = 1 << ADM_PHY_POWER_ON; ---*/
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON)
                                  | (1 << ADM_PHY_POWER_SAVE);
            phy->ports        = 1;
            phy->mdio_address = cpmac_global.ports; /* FIXME Temporary solution */
            external_ports    = 1;
            max_port_speed = avm_event_ethernet_speed_1G;
#           if defined(CONFIG_AR9)
            cpmac_global.instance[0].mii_mode[phy_number - 1] = AR9_MODE_RGMII;
#           endif /*--- #if defined(CONFIG_AR9) ---*/
            break;
        case CPMAC_PHY_TYPE_ARX188:
#           if defined(CONFIG_AR9)
#           if 0
            phy->f->check_link             = cpphy_mdio_check_link_arx188;
            phy->f->config_ports           = adm_tantos_ar9_config_ports;
            phy->f->config_vlan_group      = adm_tantos_ar9_config_vlan_group;
            phy->f->get_new_vid            = cpphy_tantos_get_new_vid;
            phy->f->get_switch_register    = cpmac_ar9_get_switch_register;
            phy->f->init_phy               = cpphy_mdio_init_ar9;
            phy->f->mgmt_check_link        = cpphy_mgmt_check_link_orig;
            phy->f->set_default_vid        = cpphy_tantos_set_default_vid;
            phy->f->set_switch_register    = cpmac_ar9_set_switch_register;
            /*--- phy->f->set_wan_keep_tagging    = adm_set_wan_keep_tagging; ---*/
            phy->f->switch_dump            = switch_dump_tantos_ar9;
            /*--- phy->f->update_hw_status        = cpphy_adm_update_hw_status; ---*/
            phy->f->vlan_clear             = adm_tantos_ar9_vlan_clear;
#           endif /*--- #if 0 ---*/
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON)
                                  | (1 << ADM_PHY_POWER_SAVE);
            phy->ports                               = 3;
            mdio->switch_config.cpu_port             = 2;
            mdio->switch_config.external_port_offset = 0;
            mdio->switch_config.is_switch            = 1;
            mdio->switch_config.max_vid_groups       = 16;
            mdio->switch_config.ports                = 3;
#           endif /*--- #if defined(CONFIG_AR9) ---*/
            cpmac_init_power_values(avm_event_ethernet_speed_1G);
            break;
        case CPMAC_PHY_TYPE_VRX288:
#           if defined(CONFIG_VR9)
            /*--- phy->f->check_link             = cpphy_mdio_check_link_arx188; ---*/
            /*--- phy->f->config_ports           = adm_tantos_ar9_config_ports; ---*/
            /*--- phy->f->config_vlan_group      = adm_tantos_ar9_config_vlan_group; ---*/
            /*--- phy->f->get_new_vid            = cpphy_tantos_get_new_vid; ---*/
            /*--- phy->f->get_switch_register    = cpmac_ar9_get_switch_register; ---*/
            /*--- phy->f->init_phy               = cpphy_mdio_init_ar9; ---*/
            /*--- phy->f->mgmt_check_link        = cpphy_mgmt_check_link_orig; ---*/
            /*--- phy->f->set_default_vid        = cpphy_tantos_set_default_vid; ---*/
            /*--- phy->f->switch_port_power       = adm_switch_port_power_6996; ---*/
            /*--- phy->f->set_switch_register    = cpmac_ar9_set_switch_register; ---*/
            /*--- phy->f->set_wan_keep_tagging    = adm_set_wan_keep_tagging; ---*/
            /*--- phy->f->switch_dump            = switch_dump_tantos_ar9; ---*/
            /*--- phy->f->update_hw_status        = cpphy_adm_update_hw_status; ---*/
            /*--- phy->f->vlan_clear             = adm_tantos_ar9_vlan_clear; ---*/
            port_modes_possible =   (1 << ADM_PHY_POWER_ON);
            /*--- port_modes_possible =   (1 << ADM_PHY_POWER_OFF) ---*/
                                  /*--- | (1 << ADM_PHY_POWER_ON) ---*/
                                  /*--- | (1 << ADM_PHY_POWER_SAVE); ---*/
            phy->ports                               = 7;
            mdio->switch_config.cpu_port             = 0;
            mdio->switch_config.external_port_offset = 0;
            mdio->switch_config.is_switch            = 1;
            mdio->switch_config.max_vid_groups       = 16;
            mdio->switch_config.ports                = 3;
#           endif /*--- #if defined(CONFIG_VR9) ---*/
            cpmac_init_power_values(avm_event_ethernet_speed_1G);
            break;
        case CPMAC_PHY_TYPE_ADM7001:
            max_port_speed = avm_event_ethernet_speed_100M;
            cpmac_init_power_values(avm_event_ethernet_speed_1G);
            // kein break
        case CPMAC_PHY_TYPE_INTERNAL:
            DEB_INFOTRC("[%s] Configure functions for PHY\n", __FUNCTION__);
            phy->f->init_phy        = cpphy_mdio_init_phy;
            phy->f->deinit_phy      = 0;
            phy->f->mgmt_check_link = cpphy_mgmt_check_link_phy;
            mdio->switch_config.external_port_offset = 0;
            port_modes_possible =   (1 << ADM_PHY_POWER_OFF)
                                  | (1 << ADM_PHY_POWER_ON);
            phy->ports     = 1;
            external_ports = 1;
            break;
#ifdef CONFIG_MII5_PORT_ENABLED 
        case CPMAC_PHY_TYPE_WASP:
            DEB_INFOTRC("[%s] Configure functions for WASP\n", __FUNCTION__);
            phy->f->init_phy            = cpmac_phys_init_wasp;
            phy->f->deinit_phy          = NULL;
            phy->f->switch_port_power   = ifx_avm_7port_power_ctrl; 
            phy->f->check_link          = cpmac_interface_check_link_dummy; 
            
            mdio->switch_config.external_port_offset = 0;
            port_modes_possible = (1 << ADM_PHY_POWER_ON);
            phy->ports     = 1;
            external_ports = 1;
            break;
#endif // CONFIG_MII5_PORT_ENABLED
        default:
            mdio->switch_config.external_port_offset = 0;
            port_modes_possible = 1 << ADM_PHY_POWER_ON;
            phy->ports     = 1;
            external_ports = 1;
            break;
    }
    switch(config->phy[phy_number].config) {
        case CPMAC_PHY_MODE_7170:
            mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_7170][0]);
            break;
        case CPMAC_PHY_MODE_VINAX5:
            mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_VINAX5][0]);
            break;
        case CPMAC_PHY_MODE_VINAX7:
            mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_VINAX7][0]);
#           if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8)
            phy->f->set_mode_pppoa = adm_set_mode_pppoa;
#           endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) ---*/
            break;
        case CPMAC_PHY_MODE_PROFI2:
#           if defined(CONFIG_MIPS_UR8)
            {
                char *cpu_nr, *cmdline = prom_getcmdline();

                mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_PROFIVOIP2][0]);
                cpu_nr = strstr(cmdline, "CPU_NR=");
                if(cpu_nr == NULL) {
                    DEB_INFO("[%s] ProfiVoIP: running on first CPU\n", __FUNCTION__);
                    mdio->switch_predefined_configs = switch_configuration[CPMAC_SWITCH_CONF_PROFIVOIP2];
                } else {
                    DEB_INFO("[%s] ProfiVoIP: running on second CPU\n", __FUNCTION__);
                    mdio->Mode = 0; /* No access at all! */
                    mdio->linked = 1;
                }
            }
#           else /*--- #if defined(CONFIG_MIPS_UR8) ---*/
            DEB_ERR("[%s] No Profi mode possible!\n", __FUNCTION__);
#           endif /*--- #else ---*/ /*--- #if defined(CONFIG_MIPS_UR8) ---*/
            break;
        case CPMAC_PHY_MODE_AR8216:
            mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_AR8216][0]);
            break;
        case CPMAC_PHY_MODE_MAGPIE:
            mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_MAGPIE][0]);
            break;
        case CPMAC_PHY_MODE_ARX188:
            mdio->switch_predefined_configs = &(switch_configuration[CPMAC_SWITCH_CONF_ARX188][0]);
            break;
        case CPMAC_PHY_MODE_PROFI1:
        case CPMAC_PHY_MODE_DEFAULT:
        case CPMAC_PHY_MODE_PASSIVE:
        case CPMAC_PHY_MODE_WASP:
            mdio->switch_predefined_configs = NULL;
            break;
        default:
            DEB_ERR("[%s] Illegal config %u for this architecture!\n", __FUNCTION__, config->phy[phy_number].config);
            return;
    }
    DEB_INFOTRC("[%s] Adjusting portsets to product specific portset\n", __FUNCTION__);
    if(mdio->switch_predefined_configs != NULL) {
        for(mode = 0; mode < CPMAC_MODE_MAX_NO; mode++) {
            for(device = 0; device < mdio->switch_predefined_configs[mode].number_of_devices; device++) {
                unsigned short oldset = mdio->switch_predefined_configs[mode].device[device].target_mask;

                mdio->switch_predefined_configs[mode].device[device].target_mask &= mdio->switch_config.used_portset;
                if(oldset != mdio->switch_predefined_configs[mode].device[device].target_mask) {
                    DEB_TRC("[%s] Adjusted mode %#x, device %u: portset %#x -> %#x\n", 
                            __FUNCTION__, 
                            mode, 
                            device, 
                            oldset,
                            mdio->switch_predefined_configs[mode].device[device].target_mask);
                }
            }
        }
    }
    for(port = 0; port < external_ports; port++) {
        cpmac_global.map_port_to_phy[cpmac_global.ports] = instance_number;
        cpmac_global.event_data.port[cpmac_global.ports].maxspeed = max_port_speed;
        /* Check, whether the port should be used */
        if((1 << port) & config->phy[phy_number].ports) {
            cpmac_global.power.port_modes_possible[cpmac_global.ports] = port_modes_possible;
        } else {
            cpmac_global.power.port_modes_possible[cpmac_global.ports] = (1 << ADM_PHY_POWER_OFF);
        }
        DEB_INFOTRC("[%s] Mapping port %u to PHY %u\n", __FUNCTION__, cpmac_global.ports, phy_number);
        cpmac_global.ports++;
    }

    if(config->phy[phy_number].type == CPMAC_PHY_TYPE_NONE) {
        DEB_INFO("[%s] PHY %u: none\n", __FUNCTION__, phy_number);
    } else {
        DEB_INFO("[%s] PHY %u: %s, reset %u, mode %s, ports %#x, mii_port %#x, mdio %#x, port modes %s%s%s\n", __FUNCTION__,
                 phy_number,
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_ADM6996FC) ? "ADM6996LC/FC switch" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_TANTOS)    ? "Tantos switch" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_AR8216)    ? "AR8216 switch" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_AR8226)    ? "AR8226 switch" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_AR8316)    ? "AR8316 switch" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_ADM7001)   ? "ADM7001 PHY" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_VITESSE)   ? "VITESSE PHY" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_AR8030)    ? "AR8030 PHY" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_AR8035)    ? "AR8035 PHY" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_AR8035)    ? "AR8035 PHY" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_WASP)      ? "WASP" :
                 (config->phy[phy_number].type == CPMAC_PHY_TYPE_11G)       ? "11G PHY" 
                 : "unknown",
                 (config->phy[phy_number].reset_pin == 0xffff) ? 9999 : config->phy[phy_number].reset_pin,
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_DEFAULT) ? "default" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_PASSIVE) ? "passive" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_7170)    ? "7170" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_VINAX5)  ? "VINAX5" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_VINAX7)  ? "VINAX7" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_PROFI1)  ? "PROFI1" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_WASP)    ? "WASP" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_PROFI2)  ? "PROFI2" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_AR8216)  ? "AR8216" :
                 (config->phy[phy_number].config == CPMAC_PHY_MODE_MAGPIE)  ? "MAGPIE" : "unknown",
                 config->phy[phy_number].ports,
                 config->phy[phy_number].mii_portmask,
                 config->phy[phy_number].mdio_address,
                 (port_modes_possible & (1 << ADM_PHY_POWER_ON)) ? "ON " : "",
                 (port_modes_possible & (1 << ADM_PHY_POWER_SAVE)) ? "SAVE " : "",
                 (port_modes_possible & (1 << ADM_PHY_POWER_OFF)) ? "OFF " : ""
                );
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int __init cpphy_entry_probe(void) {
    char *hwrev, buff[AVM_CPMAC_MAX_HWREV_LENGTH + 1], *p;
    unsigned char instance_number = 0;
    unsigned char phy_number, phy_count;
    cpmac_product_config_struct *config = NULL;

    /* Print driver version */
    DEB_ERR("Version: %s\n", AVM_CPMAC_VERSION);

    /* Initialize administrative array */
    memset(&cpmac_global, 0, sizeof(cpmac_global_t));
    cpmac_global.products = &cpmac_products;

    /*--------------------------------------------------------------------------------------*\
    \*--------------------------------------------------------------------------------------*/
    hwrev = prom_getenv("HWRevision");
    /*--- DEB_TRC("HWRevision=\"%s\"\n", hwrev); ---*/
    if(!hwrev) {
        return -ENODEV;
    }

    strncpy(buff, hwrev, 17);
    buff[AVM_CPMAC_MAX_HWREV_LENGTH - 1] = 0;
    assert(strlen(buff) < AVM_CPMAC_MAX_HWREV_LENGTH - 2); /* To ease Klocwork checking */
    strncat(buff, " ", 1);
    assert(strlen(buff) < AVM_CPMAC_MAX_HWREV_LENGTH); /* To ease Klocwork checking */

    /* These are not supported anymore! */
    if(   (!(strncmp( "94", hwrev, 2)) && (strlen(hwrev) < 4)) /* First 7170 revision */
       || (!(strncmp( "95", hwrev, 2)) && (strlen(hwrev) < 4)) /* First 7140 revision */
       || (!(strncmp("107", hwrev, 3)) && (strlen(hwrev) < 5)) /* First 7140 Annex A revision */
      ) {
        cpphy_entry_invalid_product_warning();
        return -ENODEV;
    }
    /*----------------------------------------------------------------------------------*\
     * aus "94.0.0.1 "  -->   "94 " erzeugen
    \*----------------------------------------------------------------------------------*/
    p = strchr(buff, '.');
    if(p) {
        *p++ = ' ';
        *p   = '\0';
    }

    cpmac_device_name_init("cpmac");
    for(cpmac_global.product_id = 0; cpmac_global.product_id < cpmac_global.products->number_of_products; cpmac_global.product_id++) {
        if(!strncmp(cpmac_global.products->product[cpmac_global.product_id].hwrev, buff, AVM_CPMAC_MAX_HWREV_LENGTH)) {
            config = &cpmac_global.products->product[cpmac_global.product_id];

            break; /* Product found */
        }
    }
    /* Was a product found? */
    if(config == NULL) {
        cpphy_entry_invalid_product_warning();
        return -ENODEV;
    }

    /* Reset and prepare structures */
    cpmac_global.ports = 0;
    for(instance_number = 0; instance_number < AVM_CPMAC_MAX_INSTANCES; instance_number++) {
        memset(&cpmac_global.instance[instance_number], 0, sizeof(cpmac_instance_t));
    }
    for(phy_number = 0; phy_number < AVM_CPMAC_MAX_PHYS; phy_number++) {
        memset(&cpmac_global.phy[phy_number], 0, sizeof(cpmac_phy_t));
    }

    instance_number = 0;
    phy_count = 0;
    for(phy_number = 0; phy_number < config->phys; phy_number++) {
        cpmac_global.phy[phy_count].is_used = (config->phy[phy_number].type == CPMAC_PHY_TYPE_NONE) ? 0 : 1;
        if(cpmac_global.phy[phy_count].is_used == 0) {
           continue;
        }
        cpmac_global.phy[phy_count].f = kzalloc(sizeof(cpmac_funcs_t), GFP_KERNEL);
        if(cpmac_global.phy[phy_count].f == NULL) {
            unsigned int i;

            DEB_ERR("[%s] Out of memory!\n", __FUNCTION__);
            for(i = 0; i < phy_count; i++) {
                DEB_INFOTRC("[%s] Freeing function array for PHY %u.\n", __FUNCTION__, i);
                kfree(cpmac_global.phy[i].f);
            }
            return -ENOMEM;
        }
        if(config->phy[phy_number].is_an_instance) {
            cpmac_global.instance[instance_number].is_used = 1;
            cpmac_global.instance[instance_number].instance = instance_number;
            /* Create dependency information */
            cpmac_global.instance[instance_number].phy = phy_count;
            cpmac_global.phy[phy_count].instance = instance_number;

            /* Check, whether we need EMV handling for the W503V */
            if(!(strncmp("136 ", hwrev, 3))) {
                cpmac_global.power.use_EMV = cpphy_entry_check_for_emv() ? 1 : 0;
            }
            cpmac_global.instance[instance_number].mdio.inst = instance_number;

            /* TODO Temporary solution */
            cpmac_global.instance[instance_number].mdio.f = cpmac_global.phy[phy_count].f;
            cpmac_global.instance[instance_number].cppi.f = cpmac_global.phy[phy_count].f;
            instance_number++;
        } else {
            cpmac_global.phy[phy_count].instance = config->phy[phy_number].connected_to_instance;
        }
        cpphy_entry_set_config(phy_count, config, phy_number);
        phy_count++;
    }

    /* TODO This should be stored somewhere else */
#   if 0
#   if defined(CONFIG_ARCH_PUMA5)
    cpmac_global.phy[0].f->cpmac_transfer_startstop = cpmac_puma_transfer_startstop;
#   elif defined(CONFIG_AR9) || defined(CONFIG_VR9) /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/
    cpmac_global.phy[0].f->cpmac_transfer_startstop = cpmac_puma_transfer_startstop;
#   endif /*--- #elif defined(CONFIG_AR9) || defined(CONFIG_VR9) ---*/
#   endif /*--- #if 0 ---*/

    if(cpmac_proc_init() != CPMAC_ERR_NOERR) {
        DEB_ERR("[%s] Proc initialisation failed!\n", __FUNCTION__);
    }
    init_waitqueue_head(&cpmac_global.power.waitqueue); /* TODO Move to power initialization */
    cpphy_mgmt_global_power_set(CPPHY_POWER_GLOBAL_OFF);
    cpphy_mgmt_global_power(0);

    cpmac_global.instances = instance_number;
    cpmac_global.phys = phy_count;

    DEB_INFO("[%s] HWRev %s: This hardware has %u PHY(s) and %u instance(s) (with %u external ports):\n", 
             __FUNCTION__, hwrev, cpmac_global.phys, cpmac_global.instances, cpmac_global.ports);

    cpphy_mgmt_init();
    cpphy_mgmt_power_init();
    cpphy_mgmt_work_start();
    cpphy_mgmt_work_add(CPMAC_WORK_TICK, cpphy_mdio_tick, 0);
    return 0;
}
arch_initcall(cpphy_entry_probe);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int cpphy_entry_device_init(void) {
    unsigned char instance_number;
    DEB_INFOTRC("[%s]\n", __FUNCTION__);

    /* Register all instances */
    for(instance_number = 0; instance_number < cpmac_global.instances; instance_number++) {
        if(cpmac_global.instance[instance_number].is_used) {
#           if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8)
            if(CPMAC_ERR_NOERR != cpphy_main_register(&cpmac_global.instance[instance_number])) {
                DEB_ERR("[%s] Could not register instance %u!\n", __FUNCTION__, instance);
                return -ENODEV;
            }
#           endif /*--- #if defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) ---*/
#           if defined(CONFIG_IP_MULTICAST_FASTFORWARD) && (defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_ARCH_PUMA5) || defined(CONFIG_AR9) || defined(CONFIG_VR9) || defined(CONFIG_MACH_AR7240) || defined(CONFIG_MACH_AR724x))
            if(cpmac_mcfw_init(instance_number, cpmac_device_name_get(instance_number)) != 0) {
                DEB_ERR("[%s] mcfw_init for instance %u failed!\n", __FUNCTION__, instance_number);
            }
#           endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) && (defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_ARCH_PUMA5) || defined(CONFIG_AR9) || defined(CONFIG_VR9) || defined(CONFIG_MACH_AR7240) || defined(CONFIG_MACH_AR724x)) ---*/
        }
    }
    cpmac_interface_event_init();

    /* Continuation via callback in cpphy_if_init() */
    return 0;
}
device_initcall(cpphy_entry_device_init);


/*----------------------------------------------------------------------------------*\
\*----------------------------------------------------------------------------------*/
static void cpphy_entry_exit(void) {
    unsigned char instance_number;

    DEB_INFO("[%s]\n", __FUNCTION__);

    cpmac_interface_event_deinit();
    cpphy_mgmt_deinit();
    for(instance_number = 0; instance_number < AVM_CPMAC_MAX_INSTANCES; instance_number++) {
#   if defined(CONFIG_IP_MULTICAST_FASTFORWARD) && (defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_ARCH_PUMA5) || defined(CONFIG_AR9))
        cpmac_mcfw_exit(instance_number);
#   endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) && (defined(CONFIG_MIPS_OHIO) || defined(CONFIG_MIPS_UR8) || defined(CONFIG_ARCH_PUMA5) || defined(CONFIG_AR9)) ---*/

        cpmac_if_release(cpmac_global.instance[instance_number].cpmac_priv);
    }
    cpmac_proc_deinit();
}
module_exit(cpphy_entry_exit);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int cpphy_entry_prepare_reboot(struct notifier_block *self __attribute__ ((unused)),
                               unsigned long          kind __attribute__ ((unused)),
                               void                  *cmd  __attribute__ ((unused))) {
    unsigned char instance_number;

    /* We are neither interested in the kind of powerdown nor the command */

    for(instance_number = 0; instance_number < AVM_CPMAC_MAX_INSTANCES; instance_number++) {
        if(cpmac_global.phy[instance_number].f->prepare_reboot) {
            DEB_INFO("System is going down. Resetting VLAN.\n");
            cpmac_global.phy[instance_number].f->prepare_reboot(&cpmac_global.instance[instance_number].mdio);
        }
    }

    return NOTIFY_DONE;
}

