/******************************************************************************
**
** FILE NAME    : ifxmips_7port_eth_sw.c
** PROJECT      : IFX UEIP
** MODULES      : IFX ethenret device driver
** DATE         : 30 July  2009
** AUTHOR       : Reddy Mallikarjuna
** DESCRIPTION  : IFX Cross-Platform ethernet device driver
** COPYRIGHT    :       Copyright (c) 2009
**                      Infineon Technologies AG
**                      Am Campeon 1-12, 85579 Neubiberg, Germany
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date                $Author                 $Comment
** 30 July 2009         Reddy Mallikarjuna      Initial UEIP release
*******************************************************************************/
/*!
  \file ifxmips_7port_eth_sw.c
  \ingroup IFX_ETH_DRV
  \brief ETH driver source file for VRx.
*/

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/moduleparam.h>
#include <linux/types.h>  /* size_t */
#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/ethtool.h>
#include <linux/proc_fs.h>
#include <linux/mii.h>
#include <linux/if_ether.h>
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/if_vlan.h>
#include <linux/sched.h>
#include <linux/netdevice.h>
#include <linux/avm_hw_config.h>

#if defined(CONFIG_RELOAD_PHY_FW)
#include <linux/fs.h>
#include <linux/vmalloc.h>
#endif /*--- #if defined(CONFIG_RELOAD_PHY_FW) ---*/

#include <asm/delay.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/prom.h>

#include <ifx_types.h>
#include <ifx_regs.h>
#include <common_routines.h>
#include <ifx_pmu.h>
#include <ifx_dma_core.h>

#include "ifx_eth_framework.h"
#include "ifxmips_eth.h"
#include "ifxmips_7port_sw_reg.h"
#include "ifxmips_7port_eth_sw.h"
#include <switch_api/ifx_ethsw_pce.h>
#include <switch_api/ifx_ethsw_flow_ll.h>
#include <switch_api/ifx_ethsw_api.h>
#include <switch_api/ifx_ethsw_kernel_api.h>

#include "../cpmac_debug.h"
#include "../cpmac_ar9_if.h"
#include "../cpmac_interface.h"
#include "../cpmac_test.h"
#include "../mii_reg.h"

#define IFX_DRV_MODULE_NAME             "ifxmips_7port_eth_sw"
#define IFX_DRV_MODULE_VERSION          "1.0.4"
static char version[] = IFX_DRV_MODULE_NAME ".c:v" IFX_DRV_MODULE_VERSION " \n";

/* length of time before we decide the hardware is borked,
 * and dev->tx_timeout() should be called to fix the problem
 */
#define IFX_TX_TIMEOUT                  (10 * HZ)

#define DMA_TX_BURST_LEN                DMA_BURSTL_4DW
#define DMA_RX_BURST_LEN                DMA_BURSTL_4DW
#define ETH_PKT_BUF_SIZE                1568

extern void *avm_switch_api_device;
struct dma_device_info *g_dma_device=NULL;
static struct net_device *ifx_eth_dev[NUM_ETH_INF ];
/** PMAC header structure */
static int g_dma_pmac;
/* /proc file to debug */
#define SNMP_COUNTERS_DEBUG
//#undef SNMP_COUNTERS_DEBUG
struct proc_dir_entry* g_ifx_eth_proc_dir;
/** Start the  network device interface queue */
static int ifx_eth_open(struct net_device *dev);
/** Stop the  network device interface queue */
static int ifx_eth_close(struct net_device *dev);
/** Transmit packet from Tx Queue to MAC */
int ifx_start_xmit(struct sk_buff *skb, struct net_device *dev) __attribute__ ((hot));
static void ifx_eth_rx(struct net_device *dev, int len,struct sk_buff* skb)  __attribute__ ((hot));
static void switch_hw_receive(struct dma_device_info* dma_dev) __attribute__ ((hot));
/** Hardware specific IOCTL's  */
static int ifx_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
/** Get the network device statistics */
static struct net_device_stats *ifx_get_stats (struct net_device *dev);
/** change MTU values */
static int ifx_change_mtu (struct net_device *dev, int new_mtu);
/**  Set mac address*/
static int ifx_set_mac_address(struct net_device *dev, void *p);
/** Transmit timeout*/
static void ifx_tx_timeout (struct net_device *dev);
/* select the tx dma channel */
static int ifx_select_tx_chan (struct sk_buff *skb, struct net_device *dev);
/*open dma rx channel*/
static void enable_dma_channel(void);
/*close the dma rx channel*/
static void disable_dma_channel(void);
/** Init of the network device */
static int switch_init(struct net_device *dev);
/* Get the ether addr from u-boot */
static unsigned char my_ethaddr[MAX_ADDR_LEN];

volatile unsigned int phy_is_power_down[NUM_ETH_INF];
void ifx_avm_7port_power_ctrl(unsigned char port, unsigned char power_on);
static void set_mdio_polling(int on);

void vr9_write_mdio_sched(unsigned int phyAddr, unsigned int regAddr,unsigned int data);
unsigned short vr9_read_mdio_sched(unsigned int phyAddr, unsigned int regAddr);
static inline unsigned int get_mdio_addr_from_phy_nr(unsigned int phy_nr);
static inline unsigned int get_network_dev_num_from_phy_nr(unsigned int phy_nr);
static inline unsigned int get_phy_nr_from_mdio_addr(unsigned int mdio_addr);
static inline unsigned int get_phy_nr_from_mii_nr(unsigned int mii_nr);

static struct semaphore mdio_semaphore;

static cpu_ingress_pkt_header_t default_hdr = {
    .resv_7_3      = 0,    
    .SPID          = 2,
    .CRCGEN_DIS    = 0,
    .resv_14_11    = 0,
    .DPID          = 0,
    .Port_Map_EN   = 0,
    .Port_Map_Sel  = 0,
    .LRN_DIS       = 0,
    .CLASS_EN      = 0,
    .Ig_Tr_CLASS   = 0,
    .resv_31       = 0,
    .Dest_Port_Map = 0,
    .DPID_EN       = 0,
};

static char switch_dev[]="/dev/switch/0";
static IFX_ETHSW_HANDLE switch_handle;

static struct wlan_offload_cfg offload_cfg;

/*--- #define DUMP_PACKET ---*/

#ifdef DUMP_PACKET
/*
* \brief dump skb data
* \param[in] len length of the data buffer
* \param[in] pData Pointer to data to dump
*
* \return void No Value
*/
static inline void dump_skb(u32 len, char *pData){
    int i;
    for(i=0;i<len;i++){
        printk("%2.2x ",(u8)(pData[i]));
        if (i % 16 == 15)
            printk("\n");
    }
    printk("\n");
}
#endif

static void init_offload_cfg(void)
{
    struct _avm_hw_config cfg;
    int ret;

    ret = avm_get_hw_config(AVM_HW_CONFIG_VERSION, "wlan_offload_mac", &cfg.value, &cfg.param);
    if(ret != 0){
        goto error_out;
    }

    offload_cfg.offload_mac = cfg.value;

    cfg.name = "gpio-wlan_offload_wasp_reset";
    ret = avm_get_hw_config(AVM_HW_CONFIG_VERSION, cfg.name, &cfg.value, &cfg.param);
    if(ret != 0){
        goto error_out;
    }
    
    offload_cfg.reset_gpio.name = cfg.name;
    offload_cfg.reset_gpio.value = cfg.value;
    offload_cfg.reset_gpio.param = cfg.param;
    
    // TODO: short cut, should be determined in a proper, dynamic way
    offload_cfg.phy_nr      = get_phy_nr_from_mii_nr(offload_cfg.offload_mac);
    offload_cfg.mdio_addr   = get_mdio_addr_from_phy_nr(offload_cfg.phy_nr);
    offload_cfg.port_nr     = get_network_dev_num_from_phy_nr(offload_cfg.phy_nr);
    
    return;

error_out:
    offload_cfg.offload_mac = -1;
    return;
}

static unsigned int wasp_in_reset = 1;

void cpmac_magpie_mdio_write(unsigned short regadr, unsigned short phyadr __attribute__ ((unused)), unsigned short data)
{
    if(offload_cfg.offload_mac < 0){
        DEB_ERR("[%s]: No configuration for WLAN offloading found, aborting\n", __func__);
        return;
    }
    
    vr9_write_mdio_sched(cpmac_global.phy[4].mdio_address, regadr, data);
}
EXPORT_SYMBOL(cpmac_magpie_mdio_write);

unsigned int cpmac_magpie_mdio_read(unsigned short regadr, unsigned short phyadr __attribute__ ((unused)))
{
    if(offload_cfg.offload_mac < 0){
        DEB_ERR("[%s]: No configuration for WLAN offloading found, aborting\n", __func__);
        return 0xffff; // simulate dead phy
    }
    
    return vr9_read_mdio_sched(cpmac_global.phy[4].mdio_address, regadr);
}
EXPORT_SYMBOL(cpmac_magpie_mdio_read);

void cpmac_magpie_reset(enum avm_cpmac_magpie_reset value)
{
    static int gpio_reserved = 0;
    static int reset_pin = 0xffffffff;
    static enum _avm_hw_param pin_param = avm_hw_param_no_param;
    struct net_device *dev;
    int result; 
    
    if(offload_cfg.offload_mac < 0){
        DEB_ERR("[%s]: No configuration for WLAN offloading found, aborting\n", __func__);
        return;
    }

    dev = ifx_eth_dev[offload_cfg.port_nr];

    if(dev == NULL){
        DEB_ERR("[%s]: net device not initialised yet, aborting.\n", __func__);
        return;
    }

    
    if(!gpio_reserved){
        result = avm_get_hw_config(AVM_HW_CONFIG_VERSION, "gpio-wlan_offload_wasp_reset", 
                                    &reset_pin, &pin_param);

        if(result < 0){
            DEB_ERR("[%s]: Could not read WASP reset pin data from avm_hw_config!\n", __func__);
            return;
        }

        if(pin_param != avm_hw_param_gpio_out_active_low 
                && pin_param != avm_hw_param_gpio_out_active_high)
        {
            DEB_ERR("[%s]: Received bogus GPIO pin parameter from avm_hw_config: %x\n", __func__, pin_param);
            return;
        }
        
        if(ifx_gpio_register(IFX_GPIO_MODULE_WLAN_OFFLOAD_WASP_RESET)){
            DEB_ERR("[%s]: Could not register GPIO WLAN_OFFLOAD_WASP_RESET\n", __func__);
            return;
        }

        gpio_reserved = 1;
    }

    if(   (value == CPMAC_MAGPIE_RESET_ON)
       || (value == CPMAC_MAGPIE_RESET_PULSE)) {
        DEB_INFO("[%s] Putting Magpie into reset\n", __func__);
        
        netif_carrier_off(dev);

        if(pin_param == avm_hw_param_gpio_out_active_low){
            ifx_gpio_output_clear(reset_pin, IFX_GPIO_MODULE_WLAN_OFFLOAD_WASP_RESET);
        }else{
            ifx_gpio_output_set(reset_pin, IFX_GPIO_MODULE_WLAN_OFFLOAD_WASP_RESET);
        }
        
        wasp_in_reset = 1;
        msleep(1000);
    }

    if(   (value == CPMAC_MAGPIE_RESET_OFF)
       || (value == CPMAC_MAGPIE_RESET_PULSE)) {
        
        DEB_INFO("[%s] Getting Magpie out of reset\n", __func__);
        if(pin_param == avm_hw_param_gpio_out_active_low){
            ifx_gpio_output_set(reset_pin, IFX_GPIO_MODULE_WLAN_OFFLOAD_WASP_RESET);
        }else{
            ifx_gpio_output_clear(reset_pin, IFX_GPIO_MODULE_WLAN_OFFLOAD_WASP_RESET);
        }
        
        msleep(500);
        wasp_in_reset = 0;
        
        // netif_dormant_on(dev);
        // netif_carrier_on(dev);
    }
}
EXPORT_SYMBOL(cpmac_magpie_reset);


void cpmac_phys_init_wasp(unsigned char phy_number) 
{
    if(offload_cfg.offload_mac < 0){
        DEB_ERR("[%s]: No configuration for WLAN offloading found, aborting\n", __func__);
        return;
    }

    cpmac_global.power.port_on[cpmac_global.phy[phy_number].port_offset] = 1;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline void write_vlan_header(void *start, unsigned short vlanId) {
    unsigned short *ptr = (unsigned short *) start;
    ptr[0] = 0x8100;
    ptr[1] = vlanId;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline void insert_vlan_tag(void *new_start, unsigned short vlanId) {
    unsigned short *ps = (unsigned short*)new_start;
    if ((unsigned int)new_start & 0x2){
        unsigned int *pi = (unsigned int *)((unsigned short*)new_start + 1);
        ps[0] = ps[2];
        pi[0] = pi[1];
        pi[1] = pi[2];
        ps[5] = ps[7];
    } else {
        unsigned int *pi = (unsigned int *)new_start;
        pi[0] = pi[1];
        pi[1] = pi[2];
        pi[2] = pi[3];
    }
    write_vlan_header((void *) &(ps[6]), vlanId);
    return;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned short remove_vlan_tag(void *src){
    unsigned short *ps = (unsigned short*)src;
    unsigned short vlanid = ps[7];
    if ((unsigned int)src & 0x2){
        unsigned int *pi = (unsigned int *)((unsigned short*)src + 1);

        ps[7] = ps[5];
        pi[2] = pi[1];
        pi[1] = pi[0];
        ps[2] = ps[0];

    } else {
        unsigned int *pi = (unsigned int *)src;

        pi[3] = pi[2];
        pi[2] = pi[1];
        pi[1] = pi[0];
            
    }
    return vlanid;
}

/*------------------------------------------------------------------------------------------*\
 * Get the network device/interface number
\*------------------------------------------------------------------------------------------*/
inline unsigned char get_network_dev_num(struct net_device *dev) {
    IFX_switch_priv_t *priv = netdev_priv(dev);

    if(unlikely((priv->dev_index < 0) || (priv->dev_index >= cpmac_global.ports))) {
        DEB_ERR("[%s] illegal device index %d\n", __FUNCTION__, priv->dev_index);
        return 0;
    }
    return priv->dev_index;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void check_phy_irq_pending(struct work_struct *work) ;
DECLARE_DELAYED_WORK(phy_irq_work_struct, check_phy_irq_pending);
struct workqueue_struct *phy_irq_workqueue;

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_mdio_addr_from_phy_nr(unsigned int phy_nr) {
    static unsigned short mdio_addr[NUM_ETH_INF] = { 0x00, 0x01, 0x11, 0x13, 0x07 };

    if(unlikely(phy_nr >= cpmac_global.ports)) {
        DEB_ERR("[%s] illegal input value 0x%x\n", __FUNCTION__, phy_nr);
        return (unsigned int) -1;
    }
    /*--- DEB_DEBUG("[%s] phy_nr(%x) ==> mdio_addr(0x%x) called by %pF\n", ---*/
              /*--- __FUNCTION__, phy_nr, mdio_addr[phy_nr], __builtin_return_address(0)); ---*/
    return mdio_addr[phy_nr];
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_phy_nr_from_mdio_addr(unsigned int mdio_addr) {
    unsigned int phy_nr = (unsigned int)-1;
    switch(mdio_addr) {
        case 0x00: phy_nr = 0; break;
        case 0x01: phy_nr = 1; break;
        case 0x11: phy_nr = 2; break;
        case 0x13: phy_nr = 3; break;
        case 0x07: phy_nr = 4; break;
        default: 
           DEB_ERR("[%s] illegal input value 0x%x\n", __FUNCTION__, mdio_addr);
    }
    /*--- DEB_DEBUG("[%s] mdio_addr(%x) ==> phy_nr(0x%x) called by %pF\n", ---*/
              /*--- __FUNCTION__, mdio_addr, phy_nr, __builtin_return_address(0)); ---*/
    return phy_nr;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_phy_nr_from_mii_nr(unsigned int mii_nr) {
    unsigned int phy_nr = (unsigned int)-1;
    switch(mii_nr) {
        case 0x00: phy_nr = 0; break;
        case 0x01: phy_nr = 1; break;
        case 0x02: phy_nr = 2; break;
        case 0x04: phy_nr = 3; break;
        case 0x05: phy_nr = 4; break; 
        default: 
           DEB_ERR("[%s] illegal input value 0x%x\n", __FUNCTION__, mii_nr);
    }
    /*--- DEB_DEBUG("[%s] mii_nr(%x) ==> phy_nr(0x%x) called by %pF\n", ---*/ 
              /*--- __FUNCTION__, mii_nr, phy_nr, __builtin_return_address(0)); ---*/
    return phy_nr;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_mii_nr_from_phy_nr(unsigned int phy_nr) {
    static unsigned char mii_nr[] = { 0, 1, 2, 4, 5 };
    if(unlikely(phy_nr >= cpmac_global.ports)) {
        DEB_ERR("[%s] illegal input value 0x%x\n", __FUNCTION__, phy_nr);
        return (unsigned int) -1;
    }
    /*--- DEB_DEBUG("[%s] phy_nr(%x) ==> mii_nr(0x%x) called by %pF\n", ---*/ 
              /*--- __FUNCTION__, phy_nr, mii_nr[phy_nr], __builtin_return_address(0)); ---*/
    return mii_nr[phy_nr];
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_mii_nr_from_mdio_addr(unsigned int mdio_addr) {
    return get_mii_nr_from_phy_nr(get_phy_nr_from_mdio_addr(mdio_addr));
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_phy_nr_from_network_index(unsigned int net_dev_number) {
    static unsigned char phy_nr[] = { 3, 2, 0, 1, 4 };
    if(unlikely(net_dev_number >= cpmac_global.ports)) {
        DEB_ERR("[%s] illegal input value 0x%x\n", __FUNCTION__, net_dev_number);
        return net_dev_number;
    }
    /*--- DEB_DEBUG("[%s] net_dev_nr(%x) ==> phy_nr(0x%x) called by %pF\n", ---*/
              /*--- __FUNCTION__, net_dev_number, phy_nr[net_dev_number], __builtin_return_address(0)); ---*/
    return phy_nr[net_dev_number];
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned int get_network_dev_num_from_phy_nr(unsigned int phy_nr) {
    unsigned int net_dev_number = phy_nr;
    switch(phy_nr) {
        case 0: net_dev_number = 0x02; break;
        case 1: net_dev_number = 0x03; break;
        case 2: net_dev_number = 0x01; break;
        case 3: net_dev_number = 0x00; break;
        case 4: net_dev_number = 0x04; break;
        default: 
           DEB_ERR("[%s] illegal input value 0x%x\n", __FUNCTION__, phy_nr);
    }
    /*--- DEB_DEBUG("[%s] phy_nr(%x) ==> net_dev_number(0x%x) called by %pF\n", ---*/ 
              /*--- __FUNCTION__, phy_nr, net_dev_number, __builtin_return_address(0)); ---*/
    return net_dev_number;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline unsigned short get_vlan_tag_from_network_dev(struct net_device *dev) {
    unsigned short dev_num, phy_num, tag;

    dev_num = get_network_dev_num(dev);
    phy_num = get_phy_nr_from_network_index(dev_num);

    if(phy_num >= cpmac_global.ports) {
        DEB_ERR("[%s] unknown device number!\n", __FUNCTION__);
        return 0;
    }

    tag = 10 + phy_num;
    /*--- DEB_DEBUG("[%s] dev_num %u -> tag %#x \n", __FUNCTION__, dev_num, tag); ---*/

    return tag;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline struct net_device *get_network_dev_from_vlan_tag(unsigned int vlan_tag) {
    unsigned int dev_num, phy_num;

    phy_num = vlan_tag - 10;
    dev_num = get_network_dev_num_from_phy_nr(phy_num);

    if(dev_num >= cpmac_global.ports) {
        DEB_ERR("[%s] unknown vlan id %d\n", __FUNCTION__, vlan_tag);
        return NULL;
    }
    /*--- DEB_DEBUG("[%s] tag %#x -> dev_num %u\n", __FUNCTION__, vlan_tag, dev_num); ---*/

    return ifx_eth_dev[dev_num];
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void vr9_write_mdio_sched(unsigned int phyAddr, unsigned int regAddr,unsigned int data ) {
    unsigned int reg;

    if(down_interruptible(&mdio_semaphore)) {
        DEB_WARN("[%s] Could not obtain semaphore! Running anyway.\n", __FUNCTION__);
    }
    do {
        schedule();
        reg = SW_READ_REG32(MDIO_CTRL_REG);
    } while (reg & MDIO_CTRL_MBUSY);

    reg = MDIO_READ_WDATA(data);
    SW_WRITE_REG32( reg, MDIO_WRITE_REG);
    reg = (MDIO_CTRL_OP_WR | MDIO_CTRL_PHYAD_SET(phyAddr) | MDIO_CTRL_REGAD(regAddr));
    reg |= MDIO_CTRL_MBUSY;
    SW_WRITE_REG32( reg, MDIO_CTRL_REG);

    do {
        schedule();
        reg = SW_READ_REG32(MDIO_CTRL_REG);
    } while (reg & MDIO_CTRL_MBUSY);
    up(&mdio_semaphore);

    /*--- DEB_DEBUG("[%s] phyAddr %#x, regAddr %#x, data %#x\n", __FUNCTION__, phyAddr, regAddr, data); ---*/
    return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned short vr9_read_mdio_sched(unsigned int phyAddr, unsigned int regAddr ) {
    unsigned int reg;
    unsigned short data;

    if(down_interruptible(&mdio_semaphore)) {
        DEB_WARN("[%s] Could not obtain semaphore! Running anyway.\n", __FUNCTION__);
    }
    do {
        schedule();
        reg = SW_READ_REG32(MDIO_CTRL_REG);
    } while (reg & MDIO_CTRL_MBUSY);

    reg = (MDIO_CTRL_OP_RD | MDIO_CTRL_PHYAD_SET(phyAddr) | MDIO_CTRL_REGAD(regAddr));
    reg |= MDIO_CTRL_MBUSY;

    SW_WRITE_REG32( reg, MDIO_CTRL_REG);
    do {
        schedule();
        reg = SW_READ_REG32(MDIO_CTRL_REG);
    } while (reg & MDIO_CTRL_MBUSY);

    reg = SW_READ_REG32(MDIO_READ_REG);
    data = MDIO_READ_RDATA(reg);
    up(&mdio_semaphore);

    /*--- DEB_DEBUG("[%s] phyAddr %#x, regAddr %#x, data %#x\n", __FUNCTION__, phyAddr, regAddr, data); ---*/

    return data;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct _phy_link_status {
    unsigned int link_status;
    unsigned int flow_control;
    unsigned int speed;
    unsigned int duplex;  /*--- 1 full duplex ---*/
} phy_link_status[NUM_ETH_INF];

/*------------------------------------------------------------------------------------------*\
 * Link Status abfragen
\*------------------------------------------------------------------------------------------*/
unsigned int ifx_eth_dev_check_link_status(int phy_nr) {
    unsigned int phy_mdio_addr = get_mdio_addr_from_phy_nr(phy_nr);
    unsigned int net_dev_number = get_network_dev_num_from_phy_nr(phy_nr);

    if(ifx_eth_dev[net_dev_number]) {
        struct net_device *dev = ifx_eth_dev[net_dev_number];
        unsigned int phy_status;
        unsigned int mmi_rate = MII_CFG_MIIRATE_AUTO;
        unsigned int speedReg;

        if(offload_cfg.offload_mac >= 0 && offload_cfg.phy_nr == (unsigned int) phy_nr){
            if(wasp_in_reset){
                phy_status = 0;
                speedReg = 0;
            } else {
                phy_status = MII_REG_STATUS_LINK | MII_REG_STATUS_AUTONEG_COMPLETE;
                speedReg = 0x0C; // 1000BASE-T Full Duplex
            }
            printk(KERN_ERR "%s, phy_nr: 0x%x, phy_status: 0x%x, speedReg: 0x%x\n", __func__, phy_nr, phy_status, speedReg);
        } else {
            phy_status = vr9_read_mdio_sched(phy_mdio_addr, MII_REG_STATUS); // Falls Phy tot -> 0xffff abfangen? 
            speedReg = vr9_read_mdio_sched(phy_mdio_addr, 0x18);
        }

        if((  phy_status 
            & (MII_REG_STATUS_LINK | MII_REG_STATUS_AUTONEG_COMPLETE))
           == (MII_REG_STATUS_LINK | MII_REG_STATUS_AUTONEG_COMPLETE)) {

            if(!netif_dormant(dev)) {
                DEB_INFOTRC("[%s] dormant on (%s)\n", __FUNCTION__, dev->name);
                netif_dormant_on(dev);
                return 1;
            }
            
            DEB_INFOTRC("[%s] carrier on (%s)\n", __FUNCTION__, dev->name);
            netif_dormant_off(dev);
            netif_carrier_on(dev);
            netif_wake_queue(dev);
            phy_link_status[phy_nr].link_status = 1;
            if(speedReg & 0x08) { 
                DEB_INFOTRC("[%s] MII - full duplex\n", __FUNCTION__);
                phy_link_status[phy_nr].duplex = 1;
            } else {
                DEB_INFOTRC("[%s] MII - half duplex \n", __FUNCTION__);
                phy_link_status[phy_nr].duplex = 0;
            }

            switch(speedReg & 0x03) {
                case 0:
                    DEB_INFOTRC("[%s] 10BASE-T \n", __FUNCTION__);
                    phy_link_status[phy_nr].speed = 10;
                    mmi_rate = MII_CFG_MIIRATE_2_5MHZ;
                    break;
                case 1:
                    DEB_INFOTRC("[%s] 100BASE-T \n", __FUNCTION__);
                    phy_link_status[phy_nr].speed = 100;
                    mmi_rate = MII_CFG_MIIRATE_25MHZ;
                    break;
                case 2:
                    DEB_INFOTRC("[%s] 1000BASE-T \n", __FUNCTION__);
                    phy_link_status[phy_nr].speed = 1000;
                    mmi_rate = MII_CFG_MIIRATE_125MHZ;
                    break;
                default:
                    phy_link_status[phy_nr].speed = 0;
                    DEB_INFOTRC("[%s] unknown speed \n", __FUNCTION__);
                    break;
            }
        } else {
            if(phy_status & MII_REG_STATUS_LINK) {
               DEB_ERR("[%s] Error? Link, but auto negotiation not completed!\n", __FUNCTION__);
            }
            DEB_INFOTRC("[%s] carrier off (%s)\n", __FUNCTION__, dev->name);
            netif_stop_queue(dev);
            netif_dormant_off(dev);
            netif_carrier_off(dev);
            phy_link_status[phy_nr].link_status = 0;
        }
    }
    cpmac_interface_port_update((unsigned char) get_network_dev_num_from_phy_nr(phy_nr), 
                                phy_link_status[phy_nr].link_status,
                                (phy_link_status[phy_nr].speed == 1000) ? avm_event_ethernet_speed_1G 
                                : ((phy_link_status[phy_nr].speed == 100) ? avm_event_ethernet_speed_100M 
                                : avm_event_ethernet_speed_10M),
                                phy_link_status[phy_nr].duplex);

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int gphy_11g_irq_pending(unsigned int mdio_addr) {
    unsigned int status = 0;
    unsigned int mask = 0;

    if(offload_cfg.offload_mac >= 0 && offload_cfg.mdio_addr == mdio_addr){
        if(phy_link_status[offload_cfg.phy_nr].link_status == wasp_in_reset){
            status = mask = 0x0001; // simulate status change
        }
    } else {
        status = vr9_read_mdio_sched(mdio_addr, 0x1A);
        if(status) {
            mask = vr9_read_mdio_sched(mdio_addr, 0x19);
        }
    }

    if(status & mask)  {
        DEB_INFOTRC("[%s] phy_addr%#x, status=%#x mask=%#x\n", __FUNCTION__, mdio_addr, status, mask);
        DEB_INFOTRC("[%s] " "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s", __FUNCTION__, 
                    (1 << 0)  & status ? " Link state change\n" : "",
                    (1 << 1)  & status ? " Link speed change\n" : "",
                    (1 << 2)  & status ? " Duplex Mode change\n" : "",
                    (1 << 3)  & status ? " MDIX Change\n" : "",
                    (1 << 4)  & status ? " MDI Polarity Change\n" : "",
                    (1 << 5)  & status ? " Link Speed Auto Down Shift Detect\n" : "",
                    (1 << 10) & status ? " Auto Negotiation Complete\n" : "",
                    (1 << 11) & status ? " Auto Negotioation Error\n" : "",
                    (1 << 12) & status ? " Next Page Transmitter\n" : "",
                    (1 << 13) & status ? " Next Page Received\n" : "",
                    (1 << 14) & status ? " Master-Slave Resolution Error\n" : "",
                    (1 << 15) & status ? " Wake On Lan Event\n" : "\n");
        return 1;
    }
    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void check_phy_irq_pending(struct work_struct *work __attribute__ ((unused))) {
    int phy_nr;

    for(phy_nr = 0; phy_nr < cpmac_global.ports; phy_nr++) {
        unsigned int phy_mdio_addr;
        int irq_pending;
        static int irq_emulate_pending = 0;
        if(phy_is_power_down[phy_nr]) {
            continue;
        }
        phy_mdio_addr = get_mdio_addr_from_phy_nr(phy_nr);

        irq_pending = gphy_11g_irq_pending(phy_mdio_addr);
        irq_pending |= irq_emulate_pending & (1 << phy_nr);
        if(irq_pending) {
            irq_emulate_pending &= ~(1 << phy_nr);
            if(1 == ifx_eth_dev_check_link_status(phy_nr)) {
                irq_emulate_pending |= (1 << phy_nr);
            }
        }
    }
    queue_delayed_work(phy_irq_workqueue, &phy_irq_work_struct, CONFIG_HZ / 10);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void init_virt_phy_irq_workqueue(void) {
    phy_irq_workqueue = create_singlethread_workqueue("VirtualPhyIrqQueue");
    if(phy_irq_workqueue == NULL) {
        DEB_ERR("[%s] Creating workqueue failed!\n", __FUNCTION__);
        return;
    }
    queue_delayed_work(phy_irq_workqueue, &phy_irq_work_struct, 0);
}


/*------------------------------------------------------------------------------------------*\
 * Get the driver information, used by ethtool_ops
\*------------------------------------------------------------------------------------------*/
static void ifx_get_drvinfo(struct net_device *dev __attribute__ ((unused)),
                            struct ethtool_drvinfo *info) {
    /* driver driver short name (Max 32 characters) */
    strcpy (info->driver, IFX_DRV_MODULE_NAME);
    /* driver version (Max 32 characters) */
    strcpy (info->version, IFX_DRV_MODULE_VERSION);
}

/*------------------------------------------------------------------------------------------*\
 * Get the network device settings
\*------------------------------------------------------------------------------------------*/
static int ifx_get_settings(struct net_device *dev, struct ethtool_cmd *cmd) {
    IFX_switch_priv_t *priv = netdev_priv(dev);
    unsigned int  port_status;
    int dev_index = get_network_dev_num(dev);
    unsigned int phy_nr = get_phy_nr_from_network_index(dev_index);
    unsigned int mii_nr = get_mii_nr_from_phy_nr(phy_nr);
    /*--- printk("[%s] dev '%s' \n", __FUNCTION__, dev->name); ---*/
    if (dev_index == -1 ) {
        /* should not be here*/
        DEB_ERR("[%s] Dev index error (%d)!\n", __FUNCTION__, dev_index);
        return IFX_ERROR;
    }

    cmd->supported = (SUPPORTED_Autoneg | SUPPORTED_TP);
    cmd->supported |= (SUPPORTED_100baseT_Half |    \
                    SUPPORTED_100baseT_Full |       \
                    SUPPORTED_10baseT_Half |        \
                    SUPPORTED_10baseT_Full |        \
                    /*--- SUPPORTED_1000baseT_Half |      \ ---*/
                    SUPPORTED_1000baseT_Full |      \
                    SUPPORTED_MII);

    cmd->advertising = 0;
    if (priv->flags & IFX_FLAG_ADV_10HALF) {
        cmd->advertising |= ADVERTISED_10baseT_Half;
        /*--- printk("[%s] dev '%s' Advertised: 10MBit Half\n", __FUNCTION__, dev->name); ---*/
    }
    if (priv->flags & IFX_FLAG_ADV_10FULL) {
        cmd->advertising |= ADVERTISED_10baseT_Full;
        /*--- printk("[%s] dev '%s' Advertised: 10MBit Full\n", __FUNCTION__, dev->name); ---*/
    }
    if (priv->flags & IFX_FLAG_ADV_100HALF) {
        cmd->advertising |= ADVERTISED_100baseT_Half;
        /*--- printk("[%s] dev '%s' Advertised: 100MBit Half\n", __FUNCTION__, dev->name); ---*/
    }
    if (priv->flags & IFX_FLAG_ADV_100FULL) {
        cmd->advertising |= ADVERTISED_100baseT_Full;
        /*--- printk("[%s] dev '%s' Advertised: 100MBit Full\n", __FUNCTION__, dev->name); ---*/
    }
    if (priv->flags & IFX_FLAG_ADV_1000HALF) {
        cmd->advertising |= ADVERTISED_1000baseT_Half;
        /*--- printk("[%s] dev '%s' Advertised: 1000MBit Half\n", __FUNCTION__, dev->name); ---*/
    }
    if (priv->flags & IFX_FLAG_ADV_1000FULL) {
        cmd->advertising |= ADVERTISED_1000baseT_Full;
        /*--- printk("[%s] dev '%s' Advertised: 1000MBit Full\n", __FUNCTION__, dev->name); ---*/
    }

    cmd->advertising |= ADVERTISED_Pause | ADVERTISED_Asym_Pause;
    /*--- printk("[%s] dev '%s' Advertised: Paus Asym-Pause\n", __FUNCTION__, dev->name); ---*/
    

    port_status = SW_READ_REG32(MDIO_STAT_0_REG + (mii_nr * 4));

    switch ( MDIO_STAT_SPEED(port_status)) {
        case 0:
            priv->current_speed=SPEED_10;
            /*--- printk("[%s] dev '%s' 10 MBIT\n", __FUNCTION__, dev->name); ---*/
            break;
        case 1:
            priv->current_speed=SPEED_100;
            /*--- printk("[%s] dev '%s' 100 MBIT\n", __FUNCTION__, dev->name); ---*/
            break;
        case 2:
            priv->current_speed=SPEED_1000;
            /*--- printk("[%s] dev '%s' 1000 MBIT\n", __FUNCTION__, dev->name); ---*/
            break;
        default: /* should not be here */
            DEB_ERR("[%s] Port status error(%d)!\n", __FUNCTION__, port_status);
    }
    cmd->speed = priv->current_speed;

    if ( port_status & MDIO_STAT_FDUP ) {
        priv->full_duplex = DUPLEX_FULL;
        /*--- printk("[%s] dev '%s' DUPLEX_FULL\n", __FUNCTION__, dev->name); ---*/
    } else {
        priv->full_duplex = DUPLEX_HALF;
        /*--- printk("[%s] dev '%s' DUPLEX_HALF\n", __FUNCTION__, dev->name); ---*/
    }
    cmd->duplex = priv->full_duplex;

    if ( port_status & MDIO_STAT_LSTAT) {
        cmd->reserved[0]    = 0x1;  /*link up bit:0*/
        /*--- printk("[%s] dev '%s' LINK UP\n", __FUNCTION__, dev->name); ---*/
    } else {
        cmd->reserved[0]    = 0x0;  /*link down */
        /*--- printk("[%s] dev '%s' LINK DOWN\n", __FUNCTION__, dev->name); ---*/
    }

    if (port_status & MDIO_STAT_RXPAUEN ) {
        cmd->reserved[0]    |= 0x2; /*flow control enable Bit1:1*/
        /*--- printk("[%s] dev '%s' RX PAUSE enable\n", __FUNCTION__, dev->name); ---*/
    } else {
        cmd->reserved[0]    |= 0x0; /*flow control disable*/
        /*--- printk("[%s] dev '%s' RX PAUSE disable\n", __FUNCTION__, dev->name); ---*/
    }
    /* TODO*/
    cmd->phy_address = get_mdio_addr_from_phy_nr(phy_nr);

    cmd->port = PORT_MII;
    cmd->transceiver = (priv->flags & IFX_FLAG_INTERNAL_PHY) ? XCVR_INTERNAL : XCVR_EXTERNAL;
    /*--- printk("[%s] dev '%s' transceiver %s\n", __FUNCTION__, dev->name, cmd->transceiver == XCVR_INTERNAL ? "internal" : "external"); ---*/
    return 0;
}

/*------------------------------------------------------------------------------------------*\
 * Set the network device settings
\*------------------------------------------------------------------------------------------*/
static int ifx_set_settings(struct net_device *dev, struct ethtool_cmd *cmd) {
    IFX_switch_priv_t *priv = netdev_priv(dev);
    unsigned int  phy_reg;
    /*--- printk("[%s] dev '%s' \n", __FUNCTION__, dev->name); ---*/
    int dev_index = get_network_dev_num(dev);
    unsigned int phy_nr = get_phy_nr_from_network_index(dev_index);
    unsigned int mii_nr = get_mii_nr_from_phy_nr(phy_nr);
    if (dev_index == -1 ) {
        /* should not be here*/
        /*--- DEB_ERR("%s[%d]: Dev index error(%d)!!!\n", __func__,__LINE__,dev_index); ---*/
        return IFX_ERROR;
    }
    if (cmd->autoneg == AUTONEG_ENABLE) {
        /*TODO*/
        return 0;
    } else if ((cmd->speed != SPEED_100 && cmd->speed != SPEED_10 && cmd->speed != SPEED_1000) \
            ||(cmd->duplex != DUPLEX_HALF && cmd->duplex != DUPLEX_FULL)) {
            return -EINVAL;
    }

    spin_lock_irq(&priv->lock);
    if (cmd->autoneg != AUTONEG_ENABLE) {
        phy_reg = SW_READ_REG32(PHY_ADDR_0 - (mii_nr * 4));
        phy_reg &= ~(PHY_ADDR_LINKST_MASK | PHY_ADDR_SPEED_MASK | PHY_ADDR_FDUP_MASK);
        priv->flags |= IFX_FLAG_FORCE_LINK;
        priv->flags &= ~(IFX_FLAG_100_BASE_T |IFX_FLAG_1000_BASE_T | IFX_FLAG_FULL_DUPLEX);
        if (cmd->speed == SPEED_100) {
            priv->flags |= IFX_FLAG_100_BASE_T;
            phy_reg |= PHY_ADDR_SPEED_100;
        } else if(cmd->speed == SPEED_1000) {
            priv->flags |= IFX_FLAG_1000_BASE_T;
            phy_reg |= PHY_ADDR_SPEED_1000;
        }
        if (cmd->duplex == DUPLEX_FULL) {
            priv->flags |= IFX_FLAG_FULL_DUPLEX;
            phy_reg |= PHY_ADDR_FDUP_EN;
        }
        SW_WRITE_REG32(phy_reg, (PHY_ADDR_0 - (mii_nr * 4)) );
    }
    spin_unlock_irq(&priv->lock);
    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void init_gpio(void){
    /*--- printk("[%s]\n", __FUNCTION__); ---*/
    /*--- ifx_gpio_register( IFX_GPIO_MODULE_EXTPHY_INT ); ---*//* Wurde falsch verkabelt -> muss jetzt gepollt werden */
    ifx_gpio_register( IFX_GPIO_MODULE_EXTPHY_MDIO );
}

/*------------------------------------------------------------------------------------------*\
 * Reset the Phy
\*------------------------------------------------------------------------------------------*/

#define EXT_GPHY_0_GPIO 44
#define EXT_GPHY_1_GPIO 37
void reset_external_phys(unsigned int phy, unsigned int reset){
    unsigned int gpio = (phy == 0 ? EXT_GPHY_0_GPIO : EXT_GPHY_1_GPIO);
    static unsigned int init_done = 0;
    if(init_done == 0) {
        ifx_gpio_register(IFX_GPIO_MODULE_EXTPHY_RESET); 
        init_done = 1;
    }
    if(reset) {
        DEB_INFO("[%s] put external PHY %d to reset\n", __FUNCTION__, phy);
        ifx_gpio_output_clear( gpio, IFX_GPIO_MODULE_EXTPHY_RESET);
    } else {
        DEB_INFO("[%s] start external PHY %d \n", __FUNCTION__, phy);
        ifx_gpio_output_set ( gpio ,IFX_GPIO_MODULE_EXTPHY_RESET);
    }
}

void reset_phys(unsigned int phy) {
    switch(phy) {
        case 0:
        case 1:
            reset_external_phys(phy, 1);
            udelay(0x100);
            reset_external_phys(phy, 0);
            break;
    }
}

/*------------------------------------------------------------------------------------------*\
 * Reset the device
\*------------------------------------------------------------------------------------------*/
static int ifx_nway_reset(struct net_device *dev __attribute__ ((unused))) {
     /*TODO*/
    return 0;
}

/*------------------------------------------------------------------------------------------*\
 * Structure of the ether tool operation
\*------------------------------------------------------------------------------------------*/
static const struct ethtool_ops ifx_ethtool_ops = {
        .get_drvinfo            = ifx_get_drvinfo,
        .get_settings           = ifx_get_settings,
        .set_settings           = ifx_set_settings,
        .nway_reset             = ifx_nway_reset,
        .get_link               = ethtool_op_get_link,
        /*--- .get_perm_addr          = ethtool_op_get_perm_addr, ---*/
};

/*------------------------------------------------------------------------------------------*\
 * Get the ether addr from u-boot
\*------------------------------------------------------------------------------------------*/
static int __init ethaddr_setup(char *line) {
    char *ep;
    int i;
    memset(my_ethaddr, 0, MAX_ADDR_LEN);
    /* there should really be routines to do this stuff */
    for (i = 0; i < 6; i++) {
        my_ethaddr[i] = line ? simple_strtoul(line, &ep, 16) : 0;
        if (line)
            line = (*ep) ? ep+1 : ep;
    }
    DEB_INFO("mac address %2x-%2x-%2x-%2x-%2x-%2x \n",
             my_ethaddr[0],
             my_ethaddr[1],
             my_ethaddr[2],
             my_ethaddr[3],
             my_ethaddr[4],
             my_ethaddr[5]);
    return 0;
}
__setup("ethaddr=", ethaddr_setup);

/*------------------------------------------------------------------------------------------*\
 * Turn On RX DMA channels
\*------------------------------------------------------------------------------------------*/
static int enable_dma_level = 0 ;
static void enable_dma_channel(void) {
    struct dma_device_info* dma_dev=g_dma_device;
    int i;

    if(enable_dma_level++ > 0) {
        return;
    }

    /*--- printk("[%s] dma_enable\n", __FUNCTION__); ---*/
    for(i = 0 ; i < dma_dev->max_rx_chan_num; i++) {
        if ((dma_dev->rx_chan[i])->control == IFX_DMA_CH_ON)
            dma_dev->rx_chan[i]->open(dma_dev->rx_chan[i]);
    }
}

/*------------------------------------------------------------------------------------------*\
 * Turn Off RX DMA channels
\*------------------------------------------------------------------------------------------*/
static void disable_dma_channel(void) {
    struct dma_device_info* dma_dev=g_dma_device;
    int i;

    if(--enable_dma_level > 0) {
        return;
    }
    /*--- printk("[%s] dma_disable\n", __FUNCTION__); ---*/

    for (i=0; i<dma_dev->max_rx_chan_num; i++)
        dma_dev->rx_chan[i]->close(dma_dev->rx_chan[i]);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int ifx_control_mac(unsigned int mii_nr, unsigned int enable) {
    unsigned int cfg_reg;
    switch(mii_nr) {
        case 0:
            cfg_reg = SW_READ_REG32(MII_CFG_0_REG);
            cfg_reg &= ~MII_CFG_EN;
            if(enable)
                cfg_reg |=  MII_CFG_EN;
            SW_WRITE_REG32( cfg_reg, MII_CFG_0_REG );
            break;
        case 1:
            cfg_reg = SW_READ_REG32(MII_CFG_1_REG);
            cfg_reg &= ~MII_CFG_EN;
            if(enable)
                cfg_reg |=  MII_CFG_EN;
            SW_WRITE_REG32( cfg_reg, MII_CFG_1_REG );
            break;
        case 2:
            cfg_reg = SW_READ_REG32(MII_CFG_2_REG);
            cfg_reg &= ~MII_CFG_EN;
            if(enable)
                cfg_reg |=  MII_CFG_EN;
            SW_WRITE_REG32( cfg_reg, MII_CFG_2_REG );
            break;
        case 4:
            cfg_reg = SW_READ_REG32(MII_CFG_4_REG);
            cfg_reg &= ~MII_CFG_EN;
            if(enable)
                cfg_reg |=  MII_CFG_EN;
            SW_WRITE_REG32( cfg_reg, MII_CFG_4_REG );
            break;
#ifdef CONFIG_MAC5
        case 5:
            cfg_reg = SW_READ_REG32(MII_CFG_5_REG);
            cfg_reg &= ~MII_CFG_EN;
            if(enable)
                cfg_reg |=  MII_CFG_EN;
            SW_WRITE_REG32( cfg_reg, MII_CFG_5_REG );
            break;
#endif // CONFIG_MAC5
        default:
            DEB_ERR("[%s] Called with illegal mii_nr: 0x%02x\n", __FUNCTION__, mii_nr);
            break;
    }
    
    return 0;
}

/*------------------------------------------------------------------------------------------*\
 * open the network device interface
\*------------------------------------------------------------------------------------------*/
static int ifx_eth_open(struct net_device *dev) {
    IFX_switch_priv_t *priv = netdev_priv(dev);
    unsigned int net_dev_number, phy_addr, phy_nr, mii_nr;

    DEB_ERR("[%s] '%s'\n", __FUNCTION__, dev ? dev->name : "NONE");
    if(!dev)
        return IFX_SUCCESS;

    net_dev_number = get_network_dev_num(dev);
    phy_nr   = get_phy_nr_from_network_index(net_dev_number);
    phy_addr = get_mdio_addr_from_phy_nr(phy_nr);
    mii_nr   = get_mii_nr_from_phy_nr(phy_nr);

    netif_dormant_off(dev);
    netif_carrier_off(dev);

    cpmac_ar9_open(dev);

    ifx_control_mac(mii_nr, 1);
    napi_enable(&priv->napi);
    enable_dma_channel();
    netif_start_queue(dev);
    
    set_mdio_polling(1);

    ifx_avm_7port_power_ctrl(net_dev_number, 1);
    return IFX_SUCCESS;
}

/*------------------------------------------------------------------------------------------*\
 * Close the network device interface
\*------------------------------------------------------------------------------------------*/
static int ifx_eth_close(struct net_device *dev)
{
    IFX_switch_priv_t *priv = netdev_priv(dev);
    unsigned int net_dev_number, phy_addr, phy_nr, mii_nr;

    net_dev_number = get_network_dev_num(dev);
    phy_nr   = get_phy_nr_from_network_index(net_dev_number);
    phy_addr = get_mdio_addr_from_phy_nr(phy_nr);
    mii_nr   = get_mii_nr_from_phy_nr(phy_nr);

    ifx_avm_7port_power_ctrl(net_dev_number, 0);

    disable_dma_channel();
    /*--- printk("[%d:%s] '%s'\n", smp_processor_id(), __FUNCTION__, dev->name); ---*/
    ifx_control_mac(mii_nr, 0);

    DEB_ERR("[%s/%d] netif_stop_queue(%s) called\n", __FUNCTION__, __LINE__, dev->name);
    netif_dormant_off(dev);
    netif_carrier_off(dev);
    napi_disable(&priv->napi);
    netif_stop_queue(dev);
    return IFX_SUCCESS;
}

/*------------------------------------------------------------------------------------------*\
 * Send the packet to netwrok rx queue, used by  switch_hw_receive function
\*------------------------------------------------------------------------------------------*/
static void ifx_eth_rx(struct net_device *dev, int len, struct sk_buff* skb) {
    IFX_switch_priv_t *priv = netdev_priv(dev);

    /*--- printk("[%d:%s] len = %d\n", smp_processor_id(), __FUNCTION__, len); ---*/

    if(!skb) {
        DEB_ERR("[%s] argument skb is NULL!\n", __FUNCTION__);
        return;
    }
    skb->dev = dev;
#   ifdef CONFIG_AVM_PA
    if(avm_pa_dev_receive(AVM_PA_DEVINFO(dev), skb) == 0)
      return;
#   endif /*--- #ifdef CONFIG_AVM_PA ----*/

    priv->stats.rx_packets++;
    priv->stats.rx_bytes += len;
    if(cpmac_global.instance[get_network_dev_num(dev)].is_in_testmode) {
        cpmac_test_interprete_packet(skb, get_network_dev_num(dev));
        return;
    }

#   if defined(CONFIG_IP_MULTICAST_FASTFORWARD)
    skb->uniq_id &= 0xffffff;
    skb->uniq_id |= 1;
    /*--- DEB_TEST("[%s] snooping\n", __FUNCTION__); ---*/
    mcfw_snoop_recv(&cpmac_global.instance[get_network_dev_num(dev)].mcfw_netdrv, 1, skb);
#   endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) ---*/
    skb->protocol = eth_type_trans(skb, dev);
    netif_receive_skb(skb);
}

/*------------------------------------------------------------------------------------------*\
 * This function is called in dma intr handler (DMA RCV_INT interrupt).
 * This function get the data from the DMA device.
 *   if the packet is valid then it will send to upper layer based on  criteria.
 *       The switch CPU port PMAC status header is enabled, then remove the header and
 *           look from which port the packet comes and send to relative network device.
         if PMAC status header is not enabled, then send the packets eth0 interafce
\*------------------------------------------------------------------------------------------*/
static void switch_hw_receive(struct dma_device_info* dma_dev) {
    unsigned char* buf = NULL;
    struct sk_buff *skb = NULL;
    struct net_device *dev = NULL;
    unsigned short *hdr_ptr;
    unsigned short vlanId;
    int len=0 ;

    len = dma_device_read(dma_dev, &buf, (void**)&skb);
    dma_cache_inv((unsigned long)buf, 64);

    if(skb != NULL) {
        DEB_DEBUG("[%s] %u - % *B\n", __FUNCTION__, len, min(len, 128), skb->data);
    }

    if (len >= 0x600) {
        DEB_ERR("[%s] Packet is too large (%d)!\n", __FUNCTION__, len);
        goto rx_err_exit;
    }
    if (len < 64) {
        DEB_ERR("[%s] Packet is a little too small (%d): %*pB ...\n", __FUNCTION__, len, len, buf);
    }
    if (skb == NULL  ) {
        DEB_ERR("[%s] Can't restore the Packet !\n", __FUNCTION__);
        goto rx_err_exit;
    }
    /* remove CRC */
    len -= 4;
    if (len > (skb->end - skb->tail)) {
        DEB_ERR("[%s] len:%d end:%p tail:%p Err!!!\n", __FUNCTION__,(len+4), skb->end, skb->tail);
        goto rx_err_exit;
    }
    skb_put(skb, len);
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    {
        int remaininglen = len - 64;
        if(remaininglen > 0) {
            dma_cache_inv((unsigned long) buf + 64, remaininglen);
        }
    }
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    hdr_ptr = (unsigned short*) skb->data;
    if(hdr_ptr[6] != 0x8100){
        DEB_ERR("[%s] no vlan tag in ethernet frame\n", __FUNCTION__);
        goto rx_err_exit;
    }
    vlanId = remove_vlan_tag((void*) hdr_ptr);
    dev = get_network_dev_from_vlan_tag(vlanId);
    if(dev == NULL) {
        DEB_ERR("[%s] unknown vlan id %d\n", __FUNCTION__, vlanId);
        goto rx_err_exit;
    }
    skb->dev = dev;

    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    skb_pull(skb, 4);
    len -= 4;

    if(netif_running(dev) && netif_carrier_ok(dev)) {
        ifx_eth_rx(dev, len, skb);
        return;
    } else {
        DEB_INFO("[%s] netif_running(dev)=%d, netif_carrier_ok(dev)=%d \n", __FUNCTION__, netif_running(dev) , netif_carrier_ok(dev));
    }

rx_err_exit:
    
    if(skb){
        if(dev) {
            IFX_switch_priv_t *priv    = netdev_priv(dev);
            priv->stats.rx_dropped++;
            DEB_INFO("[%s] Dropping Frame (%s)", __FUNCTION__, dev->name);
        } else {
            DEB_INFO("[%s] Dropping Frame (can't log it -> no device distinguished, yet)", __FUNCTION__);
        }
        DEB_INFO("[%s] (%d bytes): %*pB\n", __FUNCTION__, len, len > 128 ? 128 : len, skb->data);
        dev_kfree_skb_any(skb);
    }
    return ;
}


/*------------------------------------------------------------------------------------------*\
 * Get the network device stats information
\*------------------------------------------------------------------------------------------*/
static struct net_device_stats *ifx_get_stats (struct net_device *dev) {
    IFX_switch_priv_t *priv = netdev_priv(dev);
    return &priv->stats;
}


/*------------------------------------------------------------------------------------------*\
 * Trasmit timeout
\*------------------------------------------------------------------------------------------*/
static void ifx_tx_timeout(struct net_device *dev) {
    IFX_switch_priv_t *priv        = netdev_priv(dev);
    /*--- struct dma_device_info* dma_dev = g_dma_device; ---*/

    DEB_ERR("[%s] dev '%s' transmit timed out, disable the dma channel irq\n", __FUNCTION__, dev->name);

    priv->stats.tx_errors++;

    /*--- for (i = 0; i < dma_dev->max_tx_chan_num; i++) { ---*/
        /*--- dma_dev->tx_chan[i]->disable_irq(dma_dev->tx_chan[i]); ---*/
    /*--- } ---*/
    /*--- DEB_ERR("[%s/%d] netif_wake_queue(%s) called\n", __FUNCTION__, __LINE__, dev->name); ---*/
    netif_wake_queue(dev);
}

/*------------------------------------------------------------------------------------------*\
 * Set the MAC address
\*------------------------------------------------------------------------------------------*/
static int ifx_set_mac_address(struct net_device *dev, void *p __attribute__ ((unused))) {
    struct sockaddr addr;
    unsigned int a[6];
    char *name;
    int number;
    char *env;

    if (netif_running(dev)) {
        /*--- DEB_ERR("[%s] net device '%s' busy\n", __FUNCTION__, dev->name); ---*/
        return -EBUSY;
    }

    /*--------------------------------------------------------------------------------------*\
     * macdsl, macwlan, sind genutzt
    \*--------------------------------------------------------------------------------------*/
    number = get_network_dev_num(dev);
    switch(number) {
        case 0:
            name = "maca";
            break;
        case 1:
            name = "macb";
            break;
        case 2:
            name = "usb_board_mac";
            break;
        case 3:
            name = "usb_rndis_mac";
            break;
        case 4:
            name = "mac_offload";
            break;
        default:
            return -EINVAL;
    }

    env = prom_getenv(name);
    if(env == NULL) {
        addr.sa_data[0] = 0x00;
        addr.sa_data[1] = 0x04;
        addr.sa_data[2] = 0x0E;
        addr.sa_data[3] = 0xFF;
        addr.sa_data[4] = 0x00;
        addr.sa_data[5] = 0x01 + number;
    } else {
        sscanf(env, "%x:%x:%x:%x:%x:%x", &a[0], &a[1], &a[2], &a[3], &a[4], &a[5]);
        addr.sa_data[0] = a[0];
        addr.sa_data[1] = a[1];
        addr.sa_data[2] = a[2];
        addr.sa_data[3] = a[3];
        addr.sa_data[4] = a[4];
        addr.sa_data[5] = a[5];
    }

    if (!is_valid_ether_addr(addr.sa_data)) {
        /*--- DEB_ERR("[%s] invalid addr %:6B for device '%s'\n", __FUNCTION__, a, dev->name); ---*/
        return -EINVAL;
    }

    memcpy(dev->dev_addr, addr.sa_data, dev->addr_len);
    /*--- DEB_ERR("[%s] set addr %:6B to device '%s'.\n", __FUNCTION__, a, dev->name); ---*/
    return 0;
}

/*------------------------------------------------------------------------------------------*\
 * Change the MTU value of the netwrok device interfaces
\*------------------------------------------------------------------------------------------*/
static int ifx_change_mtu (struct net_device *dev, int new_mtu) {
    if(new_mtu < ETH_ZLEN || new_mtu > ETH_DATA_LEN)
        return -EINVAL;
    dev->mtu = new_mtu;
    return 0;
}

/*------------------------------------------------------------------------------------------*\
 * select the DMA channel numbers, refer dma_setup_init function
\*------------------------------------------------------------------------------------------*/
static int ifx_select_tx_chan(struct sk_buff *skb __attribute__ ((unused)),
                              struct net_device *dev __attribute__ ((unused))) {
#if 1
    return 0;
#else
     int chan_no;
/*TODO: select the channel based on  criteria*/
    int dev_index = (!strcmp(dev->name, "eth0") ? 0 : 1);
    if (dev_index)
        chan_no = 1;
    else
        chan_no = 0;
    return chan_no;
#endif
}


/*------------------------------------------------------------------------------------------*\
 * Transmit packet over DMA, which comes from the Tx Queue
 * Note: Tx packet based on the interface queue.
 *       if packet comes from eth0, then sendout the packet over Tx DMA channel 0
 *       if packet comes from eth1, then sendout the packet over Tx DMA channel 1
 * refer the function "ifx_select_tx_chan" selection of dma channels
 * if switch CPU port PMAC status header is enabled, then set the status header
 *   based on criteria and push the status header infront of header.
 * if head room is not availabe for status header(4-bytes), reallocate the head room
 *   and push status header  infront of the header
\*------------------------------------------------------------------------------------------*/
int ifx_start_xmit(struct sk_buff *skb, struct net_device *dev) {
    IFX_switch_priv_t *priv = netdev_priv(dev);
    int len , rc = NETDEV_TX_OK;
    char *data;
    struct dma_device_info* dma_dev = g_dma_device;
    static struct {
        volatile int len;
        volatile char *data;
        volatile struct sk_buff *skb;
        volatile struct net_device *dev;
        volatile unsigned long *p_tx_packets;
        volatile unsigned long *p_tx_bytes;
    } pending_tx_request;

#   define DMA_NOT_START       (1U << 31)
#   define DMA_NOT_END         (1U << 30)

    /*----------------------------------------------------------------------*\
     * noch nicht abgeschlossene Transfers auf jeden Fall vorziehen
    \*----------------------------------------------------------------------*/
    if(unlikely(pending_tx_request.len)) {
        if(pending_tx_request.skb != skb) {
            DEB_TRC("[%s] Pending skb %p (dev %u) differs from current skb %p (dev %u).\n",
                    __FUNCTION__, (struct sk_buff *) pending_tx_request.skb,
                    get_network_dev_num((struct net_device *) pending_tx_request.dev),
                    skb, get_network_dev_num(dev));
            return NETDEV_TX_BUSY;
        }
        dma_dev->current_tx_chan = ifx_select_tx_chan((struct sk_buff *) pending_tx_request.skb, 
                                                      (struct net_device *) pending_tx_request.dev);
        if(   dma_device_write(dma_dev, 
                               (char *) pending_tx_request.data,
                               (int) pending_tx_request.len | DMA_NOT_START,
                               (struct sk_buff *) pending_tx_request.skb)
           != pending_tx_request.len) {
            DEB_TRC("[%s] pending second part (skb = %p) could not be transmitted\n", 
                     __FUNCTION__, pending_tx_request.skb);
            return NETDEV_TX_BUSY;
        }
        DEB_TRC("[%s] pending second part (skb = %p) was transmitted\n", __FUNCTION__, pending_tx_request.skb);
        (*pending_tx_request.p_tx_packets)++;
        *pending_tx_request.p_tx_bytes  += pending_tx_request.len;
        pending_tx_request.len = 0;
        return NETDEV_TX_OK;
    }

    if(!skb) {
        DEB_ERR("[%s] skb is NULL!\n", __FUNCTION__);
        return -EFAULT;
    }

    if(get_network_dev_num(dev) >= cpmac_global.ports) {
        dev_kfree_skb_any(skb);
        return -ENXIO;
    }
    if(   (cpmac_global.instance[get_network_dev_num(dev)].is_in_testmode)
       && (skb->uniq_id != 0x47110815)) {
        DEB_DEBUG("[%s] '%s': Dropping skb with uniq_id %#lx\n", __FUNCTION__, skb->dev->name, skb->uniq_id);
        dev_kfree_skb_any(skb);
        priv->stats.tx_errors++;
        priv->stats.tx_dropped++;
        return NETDEV_TX_OK;
    }

    /*----------------------------------------------------------------------*\
     * add vlan tag
    \*----------------------------------------------------------------------*/
    skb_orphan(skb);
    priv->skb = skb;
    dev->trans_start = jiffies;

    /*----------------------------------------------------------------------*\
     * If the skb is cloned or used by more than one user, it would have to
     * be copied on modification. To prevent this, the transfer is divided
     * into two DMA requests
    \*----------------------------------------------------------------------*/
    if(unlikely(skb_header_cloned(skb)) || unlikely((atomic_read(&(skb->users)) > 1)) || unlikely(skb_headroom(skb) < 4)) {
        char *vlan_tag;
#if 0
        DEB_ERR("[%s] frame is %scloned users %d\n", 
                __FUNCTION__, skb_header_cloned(skb) ? "" : "not ", atomic_read(&(skb->users)));
        DEB_ERR("[%s] xmit %d bytes on device %s from %:6B to %:6B %s\n", __FUNCTION__, skb->len, dev->name, skb->data + 6, skb->data,
                skb->data[12] == 0x81 ? "with TAG" : "");
#endif
        vlan_tag = kmalloc(64, GFP_ATOMIC);
        if(vlan_tag == NULL) {
            dev_kfree_skb_any(skb);
            priv->stats.tx_errors++;
            priv->stats.tx_dropped++;
            return -ENOMEM;
        }
        data = skb->data;
        memcpy(vlan_tag, data, 12);

        write_vlan_header((void *) (vlan_tag + 12), get_vlan_tag_from_network_dev(dev));
#if 0
        DEB_ERR("[%s] split %d bytes on device %s from %:6B to %:6B TAG % 4B\n", 
                __FUNCTION__, skb->len, dev->name, vlan_tag + 6, vlan_tag, vlan_tag + 12);
#endif
        /*select the tx channel*/
        dma_dev->current_tx_chan = ifx_select_tx_chan(skb, dev);
        len = skb->len < ETH_ZLEN ? ETH_ZLEN : skb->len;
        if(unlikely(dma_device_write(dma_dev, vlan_tag, (12 + 4) | DMA_NOT_END, NULL) != (12 + 4))) {
            kfree(vlan_tag);
            dev_kfree_skb_any(skb);
            priv->stats.tx_errors++;
            priv->stats.tx_dropped++;
            return rc;
        }
        if(unlikely(dma_device_write(dma_dev, data + 12, (len - 12) | DMA_NOT_START, skb) != len - 12)) {
            DEB_TRC("[%s] second part of frame (skb = %p) could not be transmitted, leave it pending\n", 
                    __FUNCTION__, skb);
            pending_tx_request.data = data + 12;
            pending_tx_request.len  = len  - 12;
            pending_tx_request.skb  = skb;
            pending_tx_request.dev  = dev;
            pending_tx_request.p_tx_packets = &priv->stats.tx_packets;
            pending_tx_request.p_tx_bytes   = &priv->stats.tx_bytes;
            return NETDEV_TX_BUSY;
        } else {
            priv->stats.tx_packets++;
            priv->stats.tx_bytes += len;
        }
#ifdef DUMP_PACKET
        DEB_TRC("[%d:%s/%d] tx len:%d\n", smp_processor_id(), __FUNCTION__, __LINE__, len);
        dump_skb(len, data);
#endif
        return rc;
    }
    /*--- DEB_DEBUG("[%s] xmit %d bytes on device %s from %*pB to %*pB %s\n", ---*/ 
              /*--- __FUNCTION__, skb->len, dev->name, 6, skb->data + 6, 6, skb->data, ---*/
              /*--- skb->data[12] == 0x81 ? "with TAG" : ""); ---*/
#if 0
    DEB_ERR("[%s] xmit %d bytes on device %s from %:6B to %:6B %s\n", __FUNCTION__, skb->len, dev->name, skb->data + 6, skb->data,
            skb->data[12] == 0x81 ? "with TAG" : "");
#endif

    skb_push(skb, 4);
    insert_vlan_tag((void *) skb->data, get_vlan_tag_from_network_dev(dev));
    skb->protocol = htons(ETH_P_8021Q);
    
    /*select the tx channel*/
    dma_dev->current_tx_chan = ifx_select_tx_chan (skb, dev);
    len = skb->len < ETH_ZLEN ? ETH_ZLEN : skb->len;
    data = skb->data;

    DEB_DEBUG("[%s] '%s' %d - %*pB\n", __FUNCTION__, dev->name, len, len < 64 ? len : 64, skb->data);

    if (unlikely(dma_device_write(dma_dev, data, len, skb) != (len))) {
        dev_kfree_skb_any(skb);
        priv->stats.tx_errors++;
        priv->stats.tx_dropped++;
  /*      rc = NETDEV_TX_BUSY;  */
    } else {
        priv->stats.tx_packets++;
        priv->stats.tx_bytes += len;
    }

#   ifdef DUMP_PACKET
    DEB_TRC("[%d:%s/%d] tx len:%d\n", smp_processor_id(), __FUNCTION__, __LINE__, len);
    dump_skb(len, data);
#   endif
    return rc;
}


/*------------------------------------------------------------------------------------------*\
 * Platform specific IOCTL's handler
\*------------------------------------------------------------------------------------------*/
static int ifx_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd) {
    /*--- int dev_index = get_network_dev_num(dev); ---*/
    int err = -EINVAL;

    if(!netif_running(dev))
        goto out;
    switch(cmd){
        case SET_VLAN_COS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "SET_VLAN_COS"); ---*/
            break;

        case SET_DSCP_COS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "SET_DSCP_COS"); ---*/
            break;

        case ENABLE_VLAN_COS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "ENABLE_VLAN_COS"); ---*/
            break;

        case DISABLE_VLAN_COS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "DISABLE_VLAN_COS"); ---*/
            break;
            
        case ENABLE_DSCP_COS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "ENABLE_DSCP_COS"); ---*/
            break;

        case DISABLE_DSCP_COS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "DISABLE_DSCP_COS"); ---*/
            break;

        case PASS_UNICAST_PACKETS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "PASS_UNICAST_PACKETS"); ---*/
            break;

        case FILTER_UNICAST_PACKETS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "FILTER_UNICAST_PACKETS"); ---*/
            break;

        case KEEP_BROADCAST_PACKETS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "KEEP_BROADCAST_PACKETS"); ---*/
            break;

        case DROP_BROADCAST_PACKETS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "DROP_BROADCAST_PACKETS"); ---*/
            break;

        case KEEP_MULTICAST_PACKETS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "KEEP_MULTICAST_PACKETS"); ---*/
            break;

        case DROP_MULTICAST_PACKETS:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "DROP_MULTICAST_PACKETS"); ---*/
            break;

        case AVM_CPMAC_IOCTL_GENERIC:
            /*--- DEB_ERR("[%s] device '%s' (index %d) '%s'\n", __FUNCTION__, dev->name, dev_index, "AVM_CPMAC_IOCTL_GENERIC"); ---*/
            err = cpmac_ar9_ioctl(dev, ifr, cmd);
            break;
        default:
            /*--- DEB_ERR("[%s] device '%s' (index %d) 'unknown 0x%x'\n", __FUNCTION__, dev->name, dev_index, cmd); ---*/
            break;
    }
out:
    return err;
}

/*------------------------------------------------------------------------------------------*\
 * DMA Pseudo Interrupt handler.
 * This function handle the DMA pseudo interrupts to handle the data packets Rx/Tx over DMA channels
 * It will handle the following interrupts
 *   RCV_INT: DMA receive the packet interrupt,So get from the PPE peripheral
 *   TX_BUF_FULL_INT: TX channel descriptors are not availabe, so, stop the transmission
         and enable the Tx channel interrupt.
 *   TRANSMIT_CPT_INT: Tx channel descriptors are availabe and resume the transmission and
         disable the Tx channel interrupt.
\*------------------------------------------------------------------------------------------*/
int dma_intr_handler(struct dma_device_info* dma_dev, int status) {
    struct net_device* dev;
    int net_device_number;
    int i;

    /*--- DEB_ERR("[%d:%s] status=0x%x\n", smp_processor_id(), __FUNCTION__, status); ---*/
    
    switch(status) {
        case RCV_INT:
            switch_hw_receive(dma_dev);
            break;
        case TX_BUF_FULL_INT:
            for(net_device_number=0; net_device_number < cpmac_global.ports ; net_device_number++){
                dev = ifx_eth_dev[net_device_number];
                DEB_DEBUG("[%s] netif_stop_queue(%s) called\n", __FUNCTION__, dev->name);
                netif_stop_queue(dev);
            }
            for(i = 0 ; i < dma_dev->max_tx_chan_num; i++) {
                if((dma_dev->tx_chan[i])->control == IFX_DMA_CH_ON) {
                    DEB_TRC("[%s] tx_chan[%u] dma enable irq called\n", __FUNCTION__, i);
                    dma_dev->tx_chan[i]->enable_irq(dma_dev->tx_chan[i]);
                }
            }
            break;
        case TRANSMIT_CPT_INT:
            for(i = 0 ; i < dma_dev->max_tx_chan_num; i++) {
                DEB_TRC("[%s] tx_chan[%u] dma disable irq called\n", __FUNCTION__, i);
                dma_dev->tx_chan[i]->disable_irq(dma_dev->tx_chan[i]);
            }
            for(net_device_number = 0; net_device_number < cpmac_global.ports; net_device_number++) {
                dev = ifx_eth_dev[net_device_number];
                DEB_DEBUG("[%s] netif_wake_queue(%s) called\n", __FUNCTION__, dev->name);
                netif_wake_queue(dev);
            }
            break;
    }
    return IFX_SUCCESS;
}

/*------------------------------------------------------------------------------------------*\
 * Allocates the buffer for ethernet packet.
 * This function is invoke when DMA callback function to be called
 *   to allocate a new buffer for Rx packets.
\*------------------------------------------------------------------------------------------*/
unsigned char* sw_dma_buffer_alloc(int len __attribute__ ((unused)),
                                   int* byte_offset,
                                   void** opt) {
    unsigned char *buffer=NULL;
    struct sk_buff *skb=NULL;
    int offset = 0;
    int dma_burst_len = g_dma_device->rx_burst_len <<2;
    /* for reserving 2 bytes in skb buffer, so, set offset 2 bytes infront of data pointer */
    /*--- DEB_ERR("[%d:%s] \n", smp_processor_id(), __FUNCTION__); ---*/
    *byte_offset= 2 + 16;
    skb = dev_alloc_skb(ETH_PKT_BUF_SIZE+dma_burst_len + 16);
    if (skb == NULL) {
        DEB_ERR("[%s] Buffer allocation failed!!!\n", __FUNCTION__);
        return NULL;
    }
    if(likely(skb)) {
        if(((u32)skb->data & (dma_burst_len -1)) != 0) {
            offset = ~((u32)skb->data+(dma_burst_len -1)) & (dma_burst_len -1);
        }
        if(offset != 0 ) {
            buffer = (unsigned char *)(skb->data+offset);
            skb_reserve(skb, offset);
        } else {
            buffer = (unsigned char*)(skb->data);
        }
        skb_reserve(skb, 2);
        *(int*)opt=(int)skb;
    }
    /*--- DEB_ERR("[%s] buffer %p skb->data %p skb->head %p skb->tail %p skb->data_len %d skb->len %d\n", ---*/
            /*--- __FUNCTION__, buffer, skb->data, skb->head, skb->tail, skb->data_len, skb->len); ---*/
    return buffer;
}

/*------------------------------------------------------------------------------------------*\
 * Free skb buffer
 * This function frees a buffer previously allocated by the DMA buffer
 *   alloc callback function. 
\*------------------------------------------------------------------------------------------*/
int sw_dma_buffer_free(unsigned char* dataptr, void* opt) {
    struct sk_buff *skb=NULL;

    if(opt == NULL){
        if(dataptr)
            kfree(dataptr);
    }else {
        skb = (struct sk_buff *) opt;
        if(skb) {
            dev_kfree_skb_any(skb);
        }
    }

    return IFX_SUCCESS;
}

/*------------------------------------------------------------------------------------------*\
 * This function scheduled from upper layer when the NAPI is enabled
\*------------------------------------------------------------------------------------------*/
static int switch_poll(struct napi_struct *napi, int budget) {
    IFX_switch_priv_t *priv = container_of(napi, IFX_switch_priv_t, napi);
    struct net_device *dev = (struct net_device*)priv->owner;
    struct dma_device_info* dma_dev = g_dma_device;
    int work_done, ret;
    /*--- printk (KERN_ERR "[%d:%s/%d]\n", smp_processor_id(), __FUNCTION__, __LINE__); ---*/
    work_done = 0;
    ret = 0;

    if (netif_running(dev)) {
        /*--- DEB_ERR("[%s/%d] dma_device_poll called\n", __FUNCTION__, __LINE__); ---*/
        ret = dma_device_poll(dma_dev, budget, &work_done);
        /*--- DEB_ERR("[%s/%d] dma_device_poll returned %d\n", __FUNCTION__, __LINE__, ret); ---*/
    } else {
        /*--- DEB_ERR("[%s] dma_device_poll device %s not running\n", __FUNCTION__, dev->name); ---*/
    }

    /* poll should return 0 if finished, 1 if not */        
    return work_done;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void switch_activate_poll(struct dma_device_info* dma_dev __attribute__ ((unused))) {
    IFX_switch_priv_t *priv;
    struct net_device *dev;
    unsigned int net_device_number;
    
    for(net_device_number=0; net_device_number < cpmac_global.ports ; net_device_number++) {
        dev = ifx_eth_dev[net_device_number];
        priv = netdev_priv(dev);
        /*--- DEB_ERR("[%s/%d] netif_rx_schedule_prep \n", __FUNCTION__, __LINE__); ---*/
        if ( netif_running(dev)) {
            if ( napi_schedule_prep(&priv->napi) ) {
                /*--- DEB_ERR("[%s/%d] __netif_rx_schedule called\n", __FUNCTION__, __LINE__); ---*/
                __napi_schedule(&priv->napi);
            }
        }
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void switch_inactivate_poll(struct dma_device_info* dma_dev __attribute__ ((unused))) {
    IFX_switch_priv_t *priv;
    struct net_device *dev;
    unsigned int net_dev_number;
    /*--- printk (KERN_ERR "[%d:%s/%d]\n", smp_processor_id(), __FUNCTION__, __LINE__); ---*/
    
    for(net_dev_number = 0; net_dev_number < cpmac_global.ports ; net_dev_number++) {
        dev = ifx_eth_dev[net_dev_number];
        priv = netdev_priv(dev);
        if(netif_running(dev) ) {
            /*--- DEB_ERR("[%s/%d] netif_rx_complete called\n", __FUNCTION__, __LINE__); ---*/
            napi_complete(&priv->napi);
       } else {
            /*--- DEB_ERR("[%s/%d] netif %s not running\n", __FUNCTION__, __LINE__, dev->name); ---*/
       }
    }
}

/*------------------------------------------------------------------------------------------*\
 * Unregister with DMA device core driver
\*------------------------------------------------------------------------------------------*/
static void dma_setup_uninit(void) {
    struct dma_device_info* dma_dev=g_dma_device;
    printk (KERN_ERR "[%d:%s/%d]\n", smp_processor_id(), __FUNCTION__, __LINE__);
    if( dma_dev ) {
        dma_device_unregister(dma_dev);
        dma_device_release(dma_dev);
    }
}

/*------------------------------------------------------------------------------------------*\
 * Register with DMA device core driver
\*------------------------------------------------------------------------------------------*/
static int dma_setup_init(void) {
    int i, ret;

    /*--- printk (KERN_ERR "[%d:%s/%d]\n", smp_processor_id(), __FUNCTION__, __LINE__); ---*/
    g_dma_device = dma_device_reserve("PPE");
    if(!g_dma_device) {
        /*--- DEB_ERR("%s[%d]: Reserved with DMA core driver failed!!!\n", __func__,__LINE__); ---*/
        return -ENODEV;
    }
    g_dma_device->buffer_alloc              = &sw_dma_buffer_alloc;
    g_dma_device->buffer_free               = &sw_dma_buffer_free;
    g_dma_device->intr_handler              = &dma_intr_handler;
    g_dma_device->num_rx_chan               = 4;
    g_dma_device->num_tx_chan               = 2;
    g_dma_device->tx_burst_len              = DMA_TX_BURST_LEN;
    g_dma_device->rx_burst_len              = DMA_RX_BURST_LEN;
    g_dma_device->tx_endianness_mode        = IFX_DMA_ENDIAN_TYPE3;
    g_dma_device->rx_endianness_mode        = IFX_DMA_ENDIAN_TYPE3;
    g_dma_device->port_packet_drop_enable   = IFX_DISABLE;

    for (i = 0; i < g_dma_device->num_rx_chan; i++) {
        g_dma_device->rx_chan[i]->packet_size       = ETH_PKT_BUF_SIZE;
        g_dma_device->rx_chan[i]->control           = IFX_DMA_CH_ON;
    }
    for (i = 0; i < g_dma_device->num_tx_chan; i++) {
        if ( (i == 0) || (i == 1)) /* eth0 --> DMA Tx0 channel, eth1--> DMA Tx1 channel*/
            g_dma_device->tx_chan[i]->control       = IFX_DMA_CH_ON;
        else
            g_dma_device->tx_chan[i]->control       = IFX_DMA_CH_OFF;
    }
    g_dma_device->activate_poll     = switch_activate_poll;
    g_dma_device->inactivate_poll   = switch_inactivate_poll;
    ret = dma_device_register(g_dma_device);
    if ( ret != IFX_SUCCESS)
        DEB_ERR("[%s] Register with DMA core driver Failed!!!\n", __FUNCTION__);

    return ret;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void set_mdio_polling(int on){
    uint32_t flags;
    printk("[%s] %s\n", __FUNCTION__, on ? "on" : "off");

    if(on) {
        if(SW_READ_REG32(MDC_CFG_0_REG)){
            DEB_TRC("[%s] polling runs already!\n", __FUNCTION__);
        } else {
            flags = (1 << 4)|(1 << 2)|(1 << 1)|(1 << 0);
#ifdef CONFIG_MAC5
            // TODO: determine this from offload_cfg
            // no auto-polling on WASP
            // flags |= (1 << 5);
#endif
            SW_WRITE_REG32(flags, MDC_CFG_0_REG);
        }
    } else {
        SW_WRITE_REG32(0, MDC_CFG_0_REG);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void ifx_avm_7port_power_ctrl(unsigned char port, unsigned char power_on) {
    unsigned int phy_nr = get_phy_nr_from_network_index(port);
    unsigned short phy_addr = get_mdio_addr_from_phy_nr(phy_nr);

    if(unlikely(port > cpmac_global.ports)) {
        DEB_ERR("[%s] illegal port %u given!\n", __FUNCTION__, port);
        return;
    }

    if(offload_cfg.offload_mac >= 0 && offload_cfg.port_nr == port){
        DEB_TRC("[%s] skipping power control on offload port\n", __func__);
        phy_is_power_down[offload_cfg.phy_nr] = !power_on;
        return;
    }

    if(power_on) {
        vr9_write_mdio_sched(phy_addr , 0x0, vr9_read_mdio_sched(phy_addr, 0) & ~(1 << 11));
        phy_is_power_down[phy_nr] = 0;
    } else {
        phy_is_power_down[phy_nr] = 1;
        vr9_write_mdio_sched(phy_addr , 0x0, vr9_read_mdio_sched(phy_addr, 0) | (1 << 11));
    }
    DEB_TRC("[%s] Power %s\n", __FUNCTION__, power_on ? "on" : "off");
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void ifx_avm_7port_speed_throttle(unsigned char port, unsigned char do_throttle) {
    unsigned short RegData, mdio_addr;

    // skip throttling of offload port
    if(offload_cfg.offload_mac >= 0 && offload_cfg.port_nr == port){
        DEB_TRC("[%s] skipping speed throttle on offload port\n", __func__);
        return;
    }

    mdio_addr = get_mdio_addr_from_phy_nr(get_phy_nr_from_network_index(port));

    RegData = vr9_read_mdio_sched(mdio_addr, MII_REG_1000BASET_CTRL);
    if(do_throttle) {
        /* Disable 1000 Base T negotiation capability */
        DEB_INFO("[%s] Throttling port %u speeds to 100 MBit/s\n", __FUNCTION__, port);
        RegData &= ~(MII_REG_1000BASET_CTRL_ADV_HD | MII_REG_1000BASET_CTRL_ADV_FD);
    } else {
        /* Enable 1000 Base T negotiation capability */
        DEB_INFO("[%s] Unthrottling port %u speeds\n", __FUNCTION__, port);
        RegData |= MII_REG_1000BASET_CTRL_ADV_HD | MII_REG_1000BASET_CTRL_ADV_FD;
    }
    vr9_write_mdio_sched(mdio_addr, MII_REG_1000BASET_CTRL, RegData);

    /* Now restart autonegotiation to let the change take effect */
    RegData = vr9_read_mdio_sched(mdio_addr, MII_REG_CONTROL);
    RegData |= MII_REG_CONTROL_AUTONEG_RESTART;
    vr9_write_mdio_sched(mdio_addr, MII_REG_CONTROL, RegData);
    /* Work around the problem, that PHY11G might not do the first restart correctly */
    msleep(50);
    vr9_write_mdio_sched(mdio_addr, MII_REG_CONTROL, RegData);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if 0
static void ifx_change_rx_flags(struct net_device *dev, int change)
{
    IFX_switch_priv_t *priv = netdev_priv(dev);

    if (change & IFF_ALLMULTI)
        ifx_dev_set_allmulti(dev, dev->flags & IFF_ALLMULTI ? 1 : -1);
}
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void ifx_print_port_config(IFX_ETHSW_portCfg_t *pPar __attribute__ ((unused)),
                           const char *func __attribute__ ((unused))) {
#if 0
    DEB_ERR("---------------------------------------------------------\n");
    DEB_ERR("[%s] PortEnable (%d)       0x%x\n", func, pPar->nPortId, pPar->eEnable);
    DEB_ERR("[%s] LearningLimit         0x%x\n", func, pPar->nLearningLimit);
    DEB_ERR("[%s] Aging                 0x%x\n", func, pPar->bAging);
    DEB_ERR("[%s] bUnicastUnknownDrop   0x%x\n", func, pPar->bUnicastUnknownDrop);
    DEB_ERR("[%s] bMulticastUnknownDrop 0x%x\n", func, pPar->bMulticastUnknownDrop );
    DEB_ERR("[%s] eFlowCtrl             %s\n",   func, 
         pPar->eFlowCtrl == IFX_ETHSW_FLOW_AUTO  ? "IFX_ETHSW_FLOW_AUTO" :
         pPar->eFlowCtrl == IFX_ETHSW_FLOW_OFF   ? "IFX_ETHSW_FLOW_OFF" :
         pPar->eFlowCtrl == IFX_ETHSW_FLOW_RX    ? "IFX_ETHSW_FLOW_RX" :
         pPar->eFlowCtrl == IFX_ETHSW_FLOW_TX    ? "IFX_ETHSW_FLOW_TX" :
         pPar->eFlowCtrl == IFX_ETHSW_FLOW_RXTX  ? "IFX_ETHSW_FLOW_RXTX" : "unkwnown");
    DEB_ERR("[%s] ePortMonitor          %s\n",   func, 
         pPar->ePortMonitor == IFX_ETHSW_PORT_MONITOR_RXTX ? "IFX_ETHSW_PORT_MONITOR_RXTX" :
         pPar->ePortMonitor == IFX_ETHSW_PORT_MONITOR_RX   ? "IFX_ETHSW_PORT_MONITOR_RX" :
         pPar->ePortMonitor == IFX_ETHSW_PORT_MONITOR_TX   ? "IFX_ETHSW_PORT_MONITOR_TX" : "IFX_ETHSW_PORT_MONITOR_NONE" );
    DEB_ERR("---------------------------------------------------------\n");
#endif
}

/*------------------------------------------------------------------------------------------*\
 * fx_set_rx_mode - Secondary Unicast, Multicast and Promiscuous mode set
 * @netdev: network interface device structure
 *
 * The set_rx_mode entry point is called whenever the unicast or multicast
 * address lists or the network interface flags are updated. This routine is
 * responsible for configuring the hardware for proper unicast, multicast,
 * promiscuous mode, and all-multi behavior.
\*------------------------------------------------------------------------------------------*/
static void ifx_set_rx_mode(struct net_device *dev __attribute__ ((unused))) {
#if 0
    int net_dev_number = get_network_dev_num(dev);
    IFX_ETHSW_portCfg_t Par;

    DEB_ERR("[%s] device '%s'\n", __FUNCTION__, dev->name);
    Par.nPortId = 6; /*--- cpu Port ---*/
    IFX_FLOW_PortCfgGet(NULL, &Par);
    ifx_print_port_config(&Par, __FUNCTION__);

    Par.nPortId = (unsigned int)net_dev_number;
    IFX_FLOW_PortCfgGet(NULL, &Par);
    ifx_print_port_config(&Par, __FUNCTION__);

    if ((dev->flags & IFF_PROMISC) || (dev->flags & IFF_ALLMULTI)) {
        DEB_ERR("[%s] device '%s': disable UnicastDrop und MulticastDrop\n", __FUNCTION__, dev->name);
        Par.bUnicastUnknownDrop   = 1;
        Par.bMulticastUnknownDrop = 1;
    } else {
        DEB_ERR("[%s] device '%s': enable UnicastDrop und MulticastDrop\n", __FUNCTION__, dev->name);
        Par.bUnicastUnknownDrop   = 1;
        Par.bMulticastUnknownDrop = 1;
    }
    IFX_FLOW_PortCfgSet(NULL, &Par);
#endif
}

/*------------------------------------------------------------------------------------------*\
 * init of the phy device 
\*------------------------------------------------------------------------------------------*/
static int phy_init_port(struct net_device *dev) {
    unsigned int net_dev_number;
    unsigned int phy_addr;
    unsigned int phy_nr;
    net_dev_number = get_network_dev_num(dev);
    phy_nr   = get_phy_nr_from_network_index(net_dev_number);
    phy_addr = get_mdio_addr_from_phy_nr(phy_nr);

    DEB_INFOTRC("[%d:%s/%d] %s up\n", smp_processor_id(), __FUNCTION__, __LINE__, dev->name);

    if(offload_cfg.offload_mac >= 0 && offload_cfg.phy_nr == phy_nr){
        DEB_INFO("[%s]: setting MDIO address of offload port\n", __func__);
        *(volatile uint32_t *)PHY_ADDR_5 |= PHY_ADDR_ADDR(offload_cfg.mdio_addr); 
        return 0;
    }

    ifx_avm_7port_power_ctrl(net_dev_number, 1);

    if((((u32) vr9_read_mdio_sched(phy_addr, 0x1e)) & 0x7FFF) >= 0x20A) { /* Check certain FW version. Only enable compatibility */
        /* The following 4 MMD Register accesses are required to set the Compatibility bit for GPHYv1.4 and later */
        vr9_write_mdio_sched(phy_addr, 0x0D, 0x001F);  /* make an MMD access to internal registers */
        vr9_write_mdio_sched(phy_addr, 0x0E, 0x01FF);  /* Address of the internal register */
        vr9_write_mdio_sched(phy_addr, 0x0D, 0x401F);  /* Make a Data Access to it */
        vr9_write_mdio_sched(phy_addr, 0x0E, 0x0001);  /* Set the Compatibility Bit */
    }

    switch(phy_nr) {
        case 0:
            /* Enable auto downshift from GBit/s to 100 MBit/s after three autoneg failures */
            vr9_write_mdio_sched(phy_addr, 0x14, 0x8006);
            mdelay(10);
            *(volatile unsigned int*)(0xbe108000 + 4 * (0xc40 + 0x15)) = 0x1800 | phy_addr; /* phy mdio addressen in switch eintragen (Register: PHY_ADDR_0) */
            break;

        case 1:
            /* Enable auto downshift from GBit/s to 100 MBit/s after three autoneg failures */
            vr9_write_mdio_sched(phy_addr, 0x14, 0x8006);
            mdelay(10);
            *(volatile unsigned int*)(0xbe108000 + 4 * (0xc40 + 0x14)) = 0x1800 | phy_addr; /* phy mdio addressen in switch eintragen (Register: PHY_ADDR_1) */
            break;

        case 2:
            vr9_write_mdio_sched(phy_addr, 0xd, 0x1F);
            mdelay(10);
            vr9_write_mdio_sched(phy_addr, 0xe, 0x1FF);
            mdelay(10);
            vr9_write_mdio_sched(phy_addr, 0xd, 0x401F);
            mdelay(10);
            vr9_write_mdio_sched(phy_addr, 0xe, 0x1);
            mdelay(10);
            /* Enable STICKY functionality for internal GPHY */
            vr9_write_mdio_sched(phy_addr, 0x14, 0x8106);

            /* Issue GPHY reset */
            vr9_write_mdio_sched(phy_addr, MII_REG_CONTROL,   MII_REG_CONTROL_RESET 
                                                            | MII_REG_CONTROL_AUTONEG_ENABLE);
            mdelay(30);
            *(volatile unsigned int*)(0xbe108000 + 4 * (0xc40 + 0x13)) = 0x1800 | phy_addr; /* phy mdio addressen in switch eintragen (Register: PHY_ADDR_2) */
            break;

        case 3:
            vr9_write_mdio_sched(phy_addr, 0xd, 0x1F);
            mdelay(10);
            vr9_write_mdio_sched(phy_addr, 0xe, 0x1FF);
            mdelay(10);
            vr9_write_mdio_sched(phy_addr, 0xd, 0x401F);
            mdelay(10);
            vr9_write_mdio_sched(phy_addr, 0xe, 0x1);
            mdelay(10);
            /* Enable STICKY functionality for internal GPHY */
            vr9_write_mdio_sched(phy_addr, 0x14, 0x8106);

            /* Issue GPHY reset */
            vr9_write_mdio_sched(phy_addr, MII_REG_CONTROL,   MII_REG_CONTROL_RESET 
                                                            | MII_REG_CONTROL_AUTONEG_ENABLE);
            mdelay(30);
            *(volatile unsigned int*)(0xbe108000 + 4 * (0xc40 + 0x11)) = 0x1800 | phy_addr; /* phy mdio addressen in switch eintragen (Register: PHY_ADDR_4) */
            break;

        case 4:  /*--- offload wlan ---*/
            return 0;
            break;

        default:
            dump_stack();
            return -ENXIO;
    }
    
    DEB_INFOTRC("[%s] automode enabled\n", __FUNCTION__);

    /*------------------------------*\
     * setup autoneg support
    \*------------------------------*/
    mdelay(10);
    vr9_write_mdio_sched(phy_addr, 
                         MII_REG_AUTONEG_ADV,
                           MII_REG_AUTONEG_ADV_SF(1)
                         | MII_REG_AUTONEG_ADV_10BT_HDX
                         | MII_REG_AUTONEG_ADV_10BT_FDX
                         | MII_REG_AUTONEG_ADV_100BT_HDX
                         | MII_REG_AUTONEG_ADV_100BT_FDX
                         | MII_REG_AUTONEG_ADV_PAUSE_SYM
                         | MII_REG_AUTONEG_ADV_PAUSE_ASYM
                        );
    mdelay(10);
    vr9_write_mdio_sched(phy_addr, 
                         MII_REG_1000BASET_CTRL,
                           MII_REG_1000BASET_CTRL_ADV_HD
                         | MII_REG_1000BASET_CTRL_ADV_FD
                         | MII_REG_1000BASET_CTRL_MSPT);

    /*------------------------------*\
     * setup phy irq mask
    \*------------------------------*/
    vr9_write_mdio_sched(phy_addr, 0x19,
            (1 << 0) | /* Link state change  */
            (1 << 1) | /* Link speed change  */
            (1 << 2) | /* Duplex Mode change  */
            /*--- (1 << 3) | ---*/ /* MDIX Change  */
            (1 << 4) | /* MDI Polarity Change  */
            (1 << 5) | /* Link Speed Auto Down Shift Detect Mask */
            /*--- (1 << 6) | ---*/ /* Res */
            /*--- (1 << 7) | ---*/ /* Res */
            /*--- (1 << 8) | ---*/ /* Res */
            /*--- (1 << 9) | ---*/ /* Res */
            (1 << 10) | /* Auto Negotiation Complete Mask */
            (1 << 11) | /* Auto Negotioation Error Mask */
            /*--- (1 << 12) | ---*/ /* Next Page Transmitter Mask */
            /*--- (1 << 13) | ---*/ /* Next Page Received Mask */
            (1 << 14) | /* Master-Slave Resoltion Error Mask */
            (1 << 15) | /* Wake On Lan Event Mask */
            0);

    /*------------------------------*\
     * PHY Optimierung:
     * WRITE MDIO.PHY.TPGDATA=0x0100
    \*------------------------------*/
    DEB_INFOTRC("[%s] setting up phy_tpgdata, phy_addr=%#x, old_val=%#x new_val=%#x\n", __FUNCTION__, phy_addr, vr9_read_mdio_sched(phy_addr, 0x1D), 0x0100);
    vr9_write_mdio_sched(phy_addr, 0x1D, 0x0100 );

    ifx_avm_7port_power_ctrl(net_dev_number, 0);
    return 0;
}


/*------------------------------------------------------------------------------------------*\
 * init of the network device
\*------------------------------------------------------------------------------------------*/
static const struct net_device_ops netdev_ops = {
    .ndo_init            = switch_init,

    .ndo_open            = ifx_eth_open,
    .ndo_stop            = ifx_eth_close,
    .ndo_start_xmit      = ifx_start_xmit,
    .ndo_do_ioctl        = ifx_eth_ioctl,
    .ndo_get_stats       = ifx_get_stats,
    .ndo_change_mtu      = ifx_change_mtu,
    .ndo_set_mac_address = ifx_set_mac_address,
    .ndo_tx_timeout      = ifx_tx_timeout,
    .ndo_set_rx_mode     = ifx_set_rx_mode,
};
static int switch_init(struct net_device *dev) {
    unsigned int net_dev_number;
    int i;
    unsigned int phy_addr;
    unsigned int phy_nr;

    /*--- printk (KERN_ERR "[%d:%s/%d] %s up\n", smp_processor_id(), __FUNCTION__, __LINE__, dev->name); ---*/
    dev->watchdog_timeo     = IFX_TX_TIMEOUT;

    SET_ETHTOOL_OPS(dev, &ifx_ethtool_ops);
    //SET_MODULE_OWNER(dev);
    //
    ifx_set_mac_address (dev, NULL);

    net_dev_number = get_network_dev_num(dev);
    
    phy_nr   = get_phy_nr_from_network_index(net_dev_number);
    phy_addr = get_mdio_addr_from_phy_nr(phy_nr);

    switch(phy_nr) {
        case 0:
            /*--- DEB_ERR("[%s] set automode for speed and link detect for mii %d.\n", __FUNCTION__, 0); ---*/
            i  = SW_READ_REG32(MAC_0_CTRL_0);
            i &= ~(MAC_CTRL_0_FCON_MASK | MAC_CTRL_0_FDUP_MASK | MAC_CTRL_0_GMII_MASK);
            i |=  (MAC_CTRL_0_FCON_AUTO | MAC_CTRL_0_FDUP_AUTO | MAC_CTRL_0_GMII_AUTO);
            SW_WRITE_REG32( i, MAC_0_CTRL_0);
            i  = SW_READ_REG32(MII_CFG_0_REG);
            i &= ~(MII_CFG_MIIRATE_MASK);
            i |=  (MII_CFG_MIIRATE_AUTO);
            SW_WRITE_REG32( i, MII_CFG_0_REG );
            break;

        case 1:
            /*--- DEB_ERR("[%s] set automode for speed and link detect for mii %d.\n", __FUNCTION__, 1); ---*/
            i  = SW_READ_REG32(MAC_1_CTRL_0);
            i &= ~(MAC_CTRL_0_FCON_MASK | MAC_CTRL_0_FDUP_MASK | MAC_CTRL_0_GMII_MASK);
            i |=  (MAC_CTRL_0_FCON_AUTO | MAC_CTRL_0_FDUP_AUTO | MAC_CTRL_0_GMII_AUTO);
            SW_WRITE_REG32( i, MAC_1_CTRL_0);
            i  = SW_READ_REG32(MII_CFG_1_REG);
            i &= ~(MII_CFG_MIIRATE_MASK);
            i |=  (MII_CFG_MIIRATE_AUTO);
            SW_WRITE_REG32( i, MII_CFG_1_REG );
            break;

        case 2:
            /*--- DEB_ERR("[%s] set automode for speed and link detect for mii %d.\n", __FUNCTION__, 2); ---*/
            i  = SW_READ_REG32(MAC_2_CTRL_0);
            i &= ~(MAC_CTRL_0_FCON_MASK | MAC_CTRL_0_FDUP_MASK | MAC_CTRL_0_GMII_MASK);
            i |=  (MAC_CTRL_0_FCON_AUTO | MAC_CTRL_0_FDUP_AUTO | MAC_CTRL_0_GMII_AUTO);
            SW_WRITE_REG32( i, MAC_2_CTRL_0);
            i  = SW_READ_REG32(MII_CFG_2_REG);
            i &= ~(MII_CFG_MIIRATE_MASK);
            i |=  (MII_CFG_MIIRATE_AUTO);
            SW_WRITE_REG32( i, MII_CFG_2_REG );
            break;

        case 3:
            /*--- DEB_ERR("[%s] set automode for speed and link detect for mii %d.\n", __FUNCTION__, 4); ---*/
            i  = SW_READ_REG32(MAC_4_CTRL_0);
            i &= ~(MAC_CTRL_0_FCON_MASK | MAC_CTRL_0_FDUP_MASK | MAC_CTRL_0_GMII_MASK);
            i |=  (MAC_CTRL_0_FCON_AUTO | MAC_CTRL_0_FDUP_AUTO | MAC_CTRL_0_GMII_AUTO);
            SW_WRITE_REG32( i, MAC_4_CTRL_0);
            i  = SW_READ_REG32(MII_CFG_4_REG);
            i &= ~(MII_CFG_MIIRATE_MASK);
            i |=  (MII_CFG_MIIRATE_AUTO);
            SW_WRITE_REG32( i, MII_CFG_4_REG );
            break;
#ifdef CONFIG_MAC5
        case 4:
            /*--- DEB_ERR("[%s] set automode for speed and link detect for mii %d.\n", __FUNCTION__, 5); ---*/
            i  = SW_READ_REG32(MAC_5_CTRL_0);
            i &= ~(MAC_CTRL_0_FCON_MASK | MAC_CTRL_0_FDUP_MASK | MAC_CTRL_0_GMII_MASK);
            i |=  (MAC_CTRL_0_FCON_RXTX | MAC_CTRL_0_FDUP_EN | MAC_CTRL_0_GMII_RGMII);
            SW_WRITE_REG32( i, MAC_5_CTRL_0);
            i  = SW_READ_REG32(MII_CFG_5_REG);
            i &= ~(MII_CFG_MIIRATE_MASK | MII_CFG_RMII_OUT | MII_CFG_MIIMODE_MASK);
            i |=  (MII_CFG_MIIRATE_125MHZ | MII_CFG_MIIMODE_RGMII);
            SW_WRITE_REG32( i, MII_CFG_5_REG );
            // set up RGMII timing for WASP
            SW_WRITE_REG32(0xC183, MII_PCDU_5_REG);
            break;
#endif
        default:
            return -ENXIO;
    }

    return 0;
}

/** Driver version info */
static inline int eth_drv_ver(char *buf)
{
    return sprintf(buf, "IFX 7 Port Eth driver, version %s,(c)2009 Infineon Technologies AG\n", version);
}

#if defined (SNMP_COUNTERS_DEBUG)

int MAC_TABLE_PRINT(void) {
    union ifx_sw_param x;
    int retval=IFX_SUCCESS;
    int port;
    static unsigned int bla = 0;

    DEB_ERR("[%s] bla = %u\n", __FUNCTION__, bla++);
    DEB_ERR("--------------------------------------------------------------\n");
    DEB_ERR("   MAC Address       | port |   age    | FID | Static  \n");
    DEB_ERR("--------------------------------------------------------------\n");

    memset(&x.MAC_tableRead, 0x00, sizeof(x.MAC_tableRead));
    x.MAC_tableRead.bInitial = IFX_TRUE;
    retval = ifx_ethsw_kioctl(switch_handle, IFX_ETHSW_MAC_TABLE_ENTRY_READ, (unsigned int) &x.MAC_tableRead);
    if(retval != IFX_SUCCESS) {
        DEB_ERR("IFX_ETHSW_MAC_TABLE_ENTRY_READ Error\n");
        return -1;

    }

    if(x.MAC_tableRead.bLast == IFX_FALSE) {
        if((x.MAC_tableRead.nMAC[0] == 0) && (x.MAC_tableRead.nMAC[1] == 0) && (x.MAC_tableRead.nMAC[2] == 0) &&
           (x.MAC_tableRead.nMAC[3] == 0) && (x.MAC_tableRead.nMAC[4] == 0) && (x.MAC_tableRead.nMAC[5] == 0)) {
            /* Nothing to be done */
        } else {
            DEB_ERR("  %02x:%02x:%02x:%02x:%02x:%02x  |  %d   |  %4d    | %d   | %d\n",
                    x.MAC_tableRead.nMAC[0], x.MAC_tableRead.nMAC[1], x.MAC_tableRead.nMAC[2],
                    x.MAC_tableRead.nMAC[3], x.MAC_tableRead.nMAC[4], x.MAC_tableRead.nMAC[5],
                    x.MAC_tableRead.nPortId,x.MAC_tableRead.nAgeTimer, x.MAC_tableRead.nFId,
                    x.MAC_tableRead.bStaticEntry );
        }

        do {
            memset(&x.MAC_tableRead, 0x00, sizeof(x.MAC_tableRead));
            retval = ifx_ethsw_kioctl(switch_handle, IFX_ETHSW_MAC_TABLE_ENTRY_READ, (unsigned int) &x.MAC_tableRead);

            if(retval != IFX_SUCCESS) {
                DEB_ERR("IFX_ETHSW_MAC_TABLE_ENTRY_READ Error\n");
                return -1;
            }

            if(x.MAC_tableRead.bLast == IFX_FALSE) {
                if((x.MAC_tableRead.nMAC[0] == 0) && (x.MAC_tableRead.nMAC[1] == 0) && (x.MAC_tableRead.nMAC[2] == 0) &&
                   (x.MAC_tableRead.nMAC[3] == 0) && (x.MAC_tableRead.nMAC[4] == 0) && (x.MAC_tableRead.nMAC[5] == 0)) {
                    /* Do nothing */
                } else {
                    if(x.MAC_tableRead.bStaticEntry == 1) {
                        for(port = 0; port < 7; port++) {
                            if(((x.MAC_tableRead.nPortId >> port) & 0x1) == 1) {
                                DEB_ERR("  %02x:%02x:%02x:%02x:%02x:%02x  |  %d   |  %4d    | %d   | %d \n",
                                        x.MAC_tableRead.nMAC[0], x.MAC_tableRead.nMAC[1], x.MAC_tableRead.nMAC[2],
                                        x.MAC_tableRead.nMAC[3], x.MAC_tableRead.nMAC[4], x.MAC_tableRead.nMAC[5],
                                        port, x.MAC_tableRead.nAgeTimer, x.MAC_tableRead.nFId,
                                        x.MAC_tableRead.bStaticEntry);
                            }
                        }
                    } else {
                        DEB_ERR("  %02x:%02x:%02x:%02x:%02x:%02x  |  %d   |  %4d    | %d   | %d\n",
                                x.MAC_tableRead.nMAC[0], x.MAC_tableRead.nMAC[1], x.MAC_tableRead.nMAC[2],
                                x.MAC_tableRead.nMAC[3], x.MAC_tableRead.nMAC[4], x.MAC_tableRead.nMAC[5],
                                x.MAC_tableRead.nPortId, x.MAC_tableRead.nAgeTimer, x.MAC_tableRead.nFId,
                                x.MAC_tableRead.bStaticEntry );
                    }
                }
            }
        } while(x.MAC_tableRead.bLast == IFX_FALSE);
    }

    DEB_ERR("--------------------------------------------------------------\n");

    return IFX_SUCCESS;
}


static int stricmp(const char *p1, const char *p2)
{
    int c1, c2;

    while ( *p1 && *p2 ) {
        c1 = *p1 >= 'A' && *p1 <= 'Z' ? *p1 + 'a' - 'A' : *p1;
        c2 = *p2 >= 'A' && *p2 <= 'Z' ? *p2 + 'a' - 'A' : *p2;
        if ( (c1 -= c2) )
            return c1;
        p1++;
        p2++;
    }
    return *p1 - *p2;
}

#define RMON_COUNT_SIZE         64

static unsigned int g_rmon_counter[7][RMON_COUNT_SIZE];

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void clear_rmon_counter(int port) {
    int i;
    unsigned int  val0=0, val1=0, val2=0, val3=0, data=0;

    for(i=0;i<RMON_COUNT_SIZE;i++) {
        SW_WRITE_REG32(i, ETHSW_BM_RAM_ADDR_REG);
        SW_WRITE_REG32((0x8000| port), ETHSW_BM_RAM_CTRL_REG);
        while(  (SW_READ_REG32(ETHSW_BM_RAM_CTRL_REG) & 0x8000) == 1  ) {
            ;
        }
        val3 = SW_READ_REG32(ETHSW_BM_RAM_VAL_3_REG);
        val2 = SW_READ_REG32(ETHSW_BM_RAM_VAL_2_REG);
        val0 = SW_READ_REG32(ETHSW_BM_RAM_VAL_0_REG);
        val1 = SW_READ_REG32(ETHSW_BM_RAM_VAL_1_REG);
        data = (val1 << 16) | (val0);
        g_rmon_counter[port][i] = data;
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int print_rmon_counter(int  port) {
    int i;
    unsigned int  val0=0, val1=0, val2=0, val3=0, data=0;
    DEB_SUPPORT("RMON counter for Port: %d\n",port);
    for(i = 0 ; i < RMON_COUNT_SIZE ; i++) {
        SW_WRITE_REG32(i, ETHSW_BM_RAM_ADDR_REG);
        SW_WRITE_REG32((0x8000| port), ETHSW_BM_RAM_CTRL_REG);
        while(  (SW_READ_REG32(ETHSW_BM_RAM_CTRL_REG) & 0x8000) == 1  ) {
            ;
        }
        val3 = SW_READ_REG32(ETHSW_BM_RAM_VAL_3_REG);
        val2 = SW_READ_REG32(ETHSW_BM_RAM_VAL_2_REG);
        val0 = SW_READ_REG32(ETHSW_BM_RAM_VAL_0_REG);
        val1 = SW_READ_REG32(ETHSW_BM_RAM_VAL_1_REG);
        data = (val1 << 16) | (val0);
        data -= g_rmon_counter[port][i];
        switch(i) {
            case 0x1F :
                DEB_SUPPORT("Receive Frame Count                    :0x%08x\n", data);
                break;
            case 0x23 :
                DEB_SUPPORT("Receive Unicast Frame Count            :0x%08x\n",data);
                break;
            case 0x22 :
                DEB_SUPPORT("Receive Multicast Frame Count          :0x%08x\n",data);
                break;
            case 0x21 :
                DEB_SUPPORT("Receive CRC errors Count               :0x%08x\n",data);
                break;
            case 0x1D :
                DEB_SUPPORT("Receive Undersize good Count           :0x%08x\n",data);
                break;
            case 0x1E :
                DEB_SUPPORT("Receive Undersize bad Count            :0x%08x\n",data);
                break;
            case 0x1B :
                DEB_SUPPORT("Receive Oversize good Count            :0x%08x\n",data);
                break;
            case 0x1C :
                DEB_SUPPORT("Receive Oversize bad Count             :0x%08x\n",data);
                break;
            case 0x20 :
                DEB_SUPPORT("Receive Pause good Count               :0x%08x\n",data);
                break;
            case 0x1A :
                DEB_SUPPORT("Receive Alignment errors Count         :0x%08x\n",data);
                break;
            case 0x12 :
                DEB_SUPPORT("Receive size 64Bytes Frame Count       :0x%08x\n",data);
                break;
            case 0x13 :
                DEB_SUPPORT("Receive size 65-127Bytes Frame Count   :0x%08x\n",data);
                break;
            case 0x14 :
                DEB_SUPPORT("Receive size 128-255Bytes Frame Count  :0x%08x\n",data);
                break;
            case 0x15 :
                DEB_SUPPORT("Receive size 256-511Bytes Frame Count  :0x%08x\n",data);
                break;
            case 0x16 :
                DEB_SUPPORT("Receive size 512-1023Bytes Frame Count :%08x\n",data);
                break;
            case 0x17 :
                DEB_SUPPORT("Receive size >1024Bytes Frame Count    :%08x\n",data);
                break;
            case 0x18 :
                DEB_SUPPORT("Receive Discard (Tail-Drop) frame Count:0x%08x\n",data);
                break;
            case 0x19 :
                DEB_SUPPORT("Receive Drop ( Filter) frame Count     :0x%08x\n",data);
                break;
            case 0x24 :
                DEB_SUPPORT("Receive Good Byte Count (Low)          :%08x\n",data);
                break;
            case 0x25 :
                DEB_SUPPORT("Receive Good Byte Count(High)          :%08x\n",data);
                break;
            case 0x26 :
                DEB_SUPPORT("Receive Bad Byte Count (Low)           :%08x\n",data);
                break;
            case 0x27 :
                DEB_SUPPORT("Receive Bad Byte Count(High)           :%08x\n",data);
                break;
            case 0x11 :
                DEB_SUPPORT("Receive Discard (Acive Congestion Management) frame Count:%08x\n",data);
                break;
            case 0x0C :
                DEB_SUPPORT("Transmit Frame Count                        :%08x\n",data);
                break;
            case 0x06 :
                DEB_SUPPORT("Transmit Unicast Frame Count                :%08x\n",data);
                break;
            case 0x07 :
                DEB_SUPPORT("Transmit Multicast Frame Count              :%08x\n",data);
                break;
            case 0x00 :
                DEB_SUPPORT("Transmit size 64Bytes Frame Count           :%08x\n",data);
                break;
            case 0x01 :
                DEB_SUPPORT("Transmit size 65-127Bytes Frame Count       :%08x\n",data);
                break;
            case 0x02 :
                DEB_SUPPORT("Transmit size 128-255Bytes Frame Count      :%08x\n",data);
                break;
            case 0x03 :
                DEB_SUPPORT("Transmit size 256-511Bytes Frame Count      :%08x\n",data);
                break;
            case 0x04 :
                DEB_SUPPORT("Transmit size 512-1023Bytes Frame Count     :%08x\n",data);
                break;
            case 0x05 :
                DEB_SUPPORT("Transmit size >1024Bytes Frame Count        :%08x\n",data);
                break;
            case 0x08 :
                DEB_SUPPORT("Transmit Single Collision Count             :%08x\n",data);
                break;
            case 0x09 :
                DEB_SUPPORT("Transmit Multiple Collision Count           :%08x\n",data);
                break;
            case 0x0A :
                DEB_SUPPORT("Transmit Late Collision Count               :%08x\n",data);
                break;
            case 0x0B :
                DEB_SUPPORT("Transmit Excessive Collision Count          :%08x\n",data);
                break;
            case 0x0D :
                DEB_SUPPORT("Transmit Pause Frame Count                  :%08x\n",data);
                break;
            case 0x10 :
                DEB_SUPPORT("Transmit Dropped Frame Count                :%08x\n",data);
                break;
            case 0x0E :
                DEB_SUPPORT("Transmit Good Byte Count (Low)              :%08x\n",data);
                break;
            case 0x0F :
                DEB_SUPPORT("Transmit Good Byte Count(High)              :%08x\n",data);
                break;
        }
    }
    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int get_token(char **p1, char **p2, int *len, int *colon) {
    int tlen = 0;
    while ( *len && !((**p1 >= 'A' && **p1 <= 'Z')
            || (**p1 >= 'a' && **p1<= 'z')
            || (**p1 >= '0' && **p1<= '9')) ) {
        (*p1)++;
        (*len)--;
    }
    if ( !*len )
        return 0;
    if ( *colon ) {
        *colon = 0;
        *p2 = *p1;
        while ( *len && **p2 > ' ' && **p2 != ',' ) {
            if ( **p2 == ':' ) {
                *colon = 1;
                break;
            }
            (*p2)++;
            (*len)--;
            tlen++;
        }
        **p2 = 0;
    } else {
        *p2 = *p1;
        while ( *len && **p2 > ' ' && **p2 != ',' ) {
            (*p2)++;
            (*len)--;
            tlen++;
        }
        **p2 = 0;
    }
    return tlen;
}

static struct semaphore ifx_7port_eth_sema = __SEMAPHORE_INITIALIZER(ifx_7port_eth_sema, 1);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int get_number(char **p, int *len, int is_hex) {
    int ret = 0;
    int n = 0;

    while (((*p)[0] == '=' || (*p)[0] == ' ') && ( *len )){
        (*p)++;
        (*len)--;
    }

    if ( (*p)[0] == '0' && (*p)[1] == 'x' )
    {
        is_hex = 1;
        (*p) += 2;
        (*len) -= 2;
    }

    if ( is_hex )
    {
        while ( *len && ((**p >= '0' && **p <= '9') || (**p >= 'a' && **p <= 'f') || (**p >= 'A' && **p <= 'F')) )
        {
            if ( **p >= '0' && **p <= '9' )
                n = **p - '0';
            else if ( **p >= 'a' && **p <= 'f' )
               n = **p - 'a' + 10;
            else if ( **p >= 'A' && **p <= 'F' )
                n = **p - 'A' + 10;
            ret = (ret << 4) | n;
            (*p)++;
            (*len)--;
        }
    }
    else
    {
        while ( *len && **p >= '0' && **p <= '9' )
        {
            n = **p - '0';
            ret = ret * 10 + n;
            (*p)++;
            (*len)--;
        }
    }

    return ret;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

#if defined(CONFIG_RELOAD_PHY_FW)
static int load_file(char *file, unsigned char **buf) {
    unsigned int bytesRead;
    unsigned char *buffer;
    loff_t filesize;
    struct file *fp;
    mm_segment_t  oldfs;  

    *buf = NULL;
    fp = filp_open(file, O_RDONLY, 00);
    if(IS_ERR(fp)) {
        printk(KERN_ERR "[%s] Failed: Could not open bitfile\n", __FUNCTION__);
        return 0;
    }
    oldfs = get_fs();
    set_fs(KERNEL_DS);
    filesize = fp->f_op->llseek(fp, 0LL, SEEK_END);
    if(filesize < 0LL) {
        printk(KERN_ERR "[%s]Could not get filesize\n", __FUNCTION__);
        return 0;
    }
    buffer = vmalloc(filesize);
    if(buffer == NULL) {
        printk(KERN_ERR "[%s] Could not allocate temporary buffer space\n", __FUNCTION__);
        return 0;
    }
    fp->f_pos = 0; /* Von Anfang an lesen */
    bytesRead = fp->f_op->read(fp, buffer, (unsigned int)filesize, &fp->f_pos);
    if(bytesRead != (unsigned int)filesize) {
        vfree(buffer);
        printk( KERN_ERR "[%s] Could not read all bytes\n", __FUNCTION__);
        return 0;
    }
    set_fs(oldfs);
    filp_close(fp, NULL);
    *buf = buffer;
    printk( KERN_ERR "[%s] successfully read %d bytes from %s to %p\n", __FUNCTION__, bytesRead, file, buffer );
    return bytesRead;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#define fw_len 65536
extern const char gphy_ge_fw_data[fw_len];
extern int vr9_gphy_reset_released_mask(int phy_num_mask);
extern int vr9_gphy_reset_mask(int phy_num_mask);

static void load_fw_to_int_phys(char *filename) {
    unsigned char *fw = NULL;
    unsigned int fw_aligned = (unsigned int)&gphy_ge_fw_data[0] ;
    unsigned char *fw_aligned_p = (unsigned char*)fw_aligned;
    unsigned int fw_aligned_phys = virt_to_phys(fw_aligned_p);
    int new_fw_len;
    int reset_mask = 0x1 | 0x2;
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] put phy to reset \n", __FUNCTION__);
    vr9_gphy_reset_mask( reset_mask);
    msleep_interruptible(200);

    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] trying to open '%s'\n", __FUNCTION__, filename);
    new_fw_len = load_file(filename , &fw);

    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] copy firmware to memory \n", __FUNCTION__);
    if (new_fw_len != fw_len ){
        printk("[%s] wrong length %d, just fw with length = %d allowed \n", __FUNCTION__, new_fw_len, fw_len );
        return; 
    }
    memcpy(fw_aligned_p, fw, new_fw_len);
    vfree(fw);
    dma_cache_wback_inv( fw_aligned, fw_len);
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] disable switch mdio polling \n", __FUNCTION__);
    set_mdio_polling(0);
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] disable phy port at switch \n", __FUNCTION__);
    ifx_control_mac(2, 0);
    ifx_control_mac(4, 0);

    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] init phy\n", __FUNCTION__);
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] load firmware\n", __FUNCTION__);
    SW_WRITE_REG32(fw_aligned_phys, GFS_ADD0 );
    SW_WRITE_REG32(fw_aligned_phys, GFS_ADD1 );
    printk("GPHY FIRMWARE LOAD SUCCESSFULLY AT ADDR : %x (GFS_ADD0 = %#x, GFS_ADD1 = %#x )\n",fw_aligned_phys, GFS_ADD0, GFS_ADD1);
    /*--- printk(" %:16B ...\n", fw_aligned_p); ---*/
    msleep_interruptible(200);
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] release reset  \n", __FUNCTION__);
    vr9_gphy_reset_released_mask( reset_mask );
    msleep_interruptible(200);
    
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] enable phy port at switch\n", __FUNCTION__);
    ifx_control_mac(2, 1);
    ifx_control_mac(4, 1);

    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] enable switch mdio polling \n", __FUNCTION__ );
    set_mdio_polling(1);
    
    /*------------------------------------------------------------*\
    \*------------------------------------------------------------*/
    printk("[%s] configure phy\n", __FUNCTION__);
    phy_init_port(ifx_eth_dev[0]);
    phy_init_port(ifx_eth_dev[1]);
    msleep_interruptible(200);

    printk("[%s] power on phy\n", __FUNCTION__);
    ifx_avm_7port_power_ctrl(0, 1);
    ifx_avm_7port_power_ctrl(1, 1);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int reload_phy_help(char *page, 
                         char **start __attribute__ ((unused)),
                         off_t off,
                         int count __attribute__ ((unused)),
                         int *eof __attribute__ ((unused)),
                         void *data __attribute__ ((unused))) {
    int len = 0;
    len += sprintf(page + off + len, "usage: echo /abs/path/to/filename > /proc/driver/ifx_7port_sw/reload_phyX_fw\n");
    return len;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static char filename_buf[2048];
static int reload_phy_fw(struct file *file __attribute__ ((unused)),
                         const char *buf,
                         unsigned long count,
                         void *data __attribute__ ((unused))) {

    int len;
    
    filename_buf[0] = 0;
    len = sizeof(filename_buf) < count ? sizeof(filename_buf) - 1 : count;
    len = len - copy_from_user(filename_buf, buf, len);
    if (len > 1) {
        if ( filename_buf[len-1] == '\n') {
            len --;
        }
        filename_buf[len] = 0;
        load_fw_to_int_phys(filename_buf);
    }
    return count;
}

#endif /*--- #if defined(CONFIG_RELOAD_PHY_FW) ---*/
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static int proc_phy_dump(char *page, 
                         char **start __attribute__ ((unused)),
                         off_t off,
                         int count __attribute__ ((unused)),
                         int *eof __attribute__ ((unused)),
                         void *data __attribute__ ((unused))) {
    int len = 0;
    int reg_offset;
    unsigned int phy_nr;
    for (phy_nr = 0; phy_nr < cpmac_global.ports ; phy_nr++){
        int mdio_addr = get_mdio_addr_from_phy_nr(phy_nr);
        len += sprintf(page + off + len, "----------------------------------------\n");
        len += sprintf(page + off + len, "dumping regs for phy %d [mdio:%#x]\n",phy_nr, mdio_addr); 
            if(offload_cfg.offload_mac >= 0 && offload_cfg.phy_nr == phy_nr){
                len += sprintf(page + off + len, "    no registers for pseudo-PHY on offload-MAC\n");
                continue;
            }
        for (reg_offset = 0; reg_offset < 0x20; reg_offset++){
            len += sprintf(page + off + len, "    reg[%#x]: %#x\n",reg_offset, vr9_read_mdio_sched(mdio_addr, reg_offset) ); 
        }
    }
    return len;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int proc_read_h3(char *page,
                        char **start __attribute__ ((unused)),
                        off_t off,
                        int count __attribute__ ((unused)),
                        int *eof,
                        void *data __attribute__ ((unused)))
{
    int len = 0;

    len += sprintf(page + off + len, "   resv_7_3               = %#x\n", default_hdr.resv_7_3     ); 
    len += sprintf(page + off + len, "   SPID                   = %#x\n", default_hdr.SPID         ); 
    len += sprintf(page + off + len, "   CRCGEN_DIS             = %#x\n", default_hdr.CRCGEN_DIS   ); 
    len += sprintf(page + off + len, "   resv_14_11             = %#x\n", default_hdr.resv_14_11   ); 
    len += sprintf(page + off + len, "   DPID                   = %#x\n", default_hdr.DPID         ); 
    len += sprintf(page + off + len, "   Port_Map_EN            = %#x\n", default_hdr.Port_Map_EN  ); 
    len += sprintf(page + off + len, "   Port_Map_Sel           = %#x\n", default_hdr.Port_Map_Sel ); 
    len += sprintf(page + off + len, "   LRN_DIS                = %#x\n", default_hdr.LRN_DIS      ); 
    len += sprintf(page + off + len, "   CLASS_EN               = %#x\n", default_hdr.CLASS_EN     ); 
    len += sprintf(page + off + len, "   Ig_Tr_CLASS            = %#x\n", default_hdr.Ig_Tr_CLASS  ); 
    len += sprintf(page + off + len, "   resv_31                = %#x\n", default_hdr.resv_31      ); 
    len += sprintf(page + off + len, "   Dest_Port_Map          = %#x\n", default_hdr.Dest_Port_Map); 
    len += sprintf(page + off + len, "   DPID_EN                = %#x\n", default_hdr.DPID_EN      ); 

    *eof = 1;

    return len;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static const char *hdr_names[] = {
   "resv_7_3", "SPID", "CRCGEN_DIS", "resv_14_11", "DPID", "Port_Map_EN", 
   "Port_Map_Sel", "LRN_DIS", "CLASS_EN", "Ig_Tr_CLASS",  "resv_31",  "Dest_Port_Map", "DPID_EN"};

static int proc_write_h3(struct file *file __attribute__ ((unused)),
                         const char *buf,
                         unsigned long count,
                         void *data __attribute__ ((unused))) {
    static char local_buf[2048];
    int len;
    char *s;

    len = sizeof(local_buf) < count ? sizeof(local_buf) - 1 : count;
    len = len - copy_from_user(local_buf, buf, len);
    local_buf[len] = 0;
    s = local_buf;

    if ( strncmp(s, hdr_names[0] , sizeof(hdr_names[0])) == 0 ) {
        s += sizeof(hdr_names[0]);
        len -=sizeof(hdr_names[0]); 
        default_hdr.resv_7_3 = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[1] , sizeof(hdr_names[1])) == 0 ) {
        s += sizeof(hdr_names[1]);
        len -=sizeof(hdr_names[1]); 
        default_hdr.SPID = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[2] , sizeof(hdr_names[2])) == 0 ) {
        s += sizeof(hdr_names[2]);
        len -=sizeof(hdr_names[2]); 
        default_hdr.CRCGEN_DIS = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[3] , sizeof(hdr_names[3])) == 0 ) {
        s += sizeof(hdr_names[3]);
        len -=sizeof(hdr_names[3]); 
        default_hdr.resv_14_11 = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[5] , sizeof(hdr_names[5])) == 0 ) {
        s += sizeof(hdr_names[5]);
        len -=sizeof(hdr_names[5]); 
        default_hdr.Port_Map_EN = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[6] , sizeof(hdr_names[6])) == 0 ) {
        s += sizeof(hdr_names[6]);
        len -=sizeof(hdr_names[6]); 
        default_hdr.Port_Map_Sel = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[7] , sizeof(hdr_names[7])) == 0 ) {
        s += sizeof(hdr_names[7]);
        len -=sizeof(hdr_names[7]); 
        default_hdr.LRN_DIS = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[8] , sizeof(hdr_names[8])) == 0 ) {
        s += sizeof(hdr_names[8]);
        len -=sizeof(hdr_names[8]); 
        default_hdr.CLASS_EN = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[9] , sizeof(hdr_names[9])) == 0 ) {
        s += sizeof(hdr_names[9]);
        len -=sizeof(hdr_names[9]); 
        default_hdr.Ig_Tr_CLASS = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[10] , sizeof(hdr_names[10])) == 0 ) {
        s += sizeof(hdr_names[10]);
        len -=sizeof(hdr_names[10]); 
        default_hdr.resv_31 = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[11] , sizeof(hdr_names[11])) == 0 ) {
        s += sizeof(hdr_names[11]);
        len -=sizeof(hdr_names[11]); 
        default_hdr.Dest_Port_Map = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[12] , sizeof(hdr_names[12])) == 0 ) {
        s += sizeof(hdr_names[12]);
        len -=sizeof(hdr_names[12]); 
        default_hdr.DPID_EN = get_number(&s, &len, 0);

    } else if ( strncmp(s, hdr_names[4] , sizeof(hdr_names[4])) == 0 ) {
        s += sizeof(hdr_names[4]);
        len -=sizeof(hdr_names[4]); 
        default_hdr.DPID = get_number(&s, &len, 0);

    } else {
        DEB_ERR("unknown command\n");
    }
    return count;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int proc_read_rmon(struct file *file __attribute__ ((unused)),
                          const char *buf,
                          unsigned long count,
                          void *data __attribute__ ((unused))) {
    static char local_buf[2048];
    int len, colon = 0, port = -1;
    char *p1, *p2;

    down(&ifx_7port_eth_sema);
    len = sizeof(local_buf) < count ? sizeof(local_buf) - 1 : count;
    len = len - copy_from_user(local_buf, buf, len);
    local_buf[len] = 0;
    p1 = local_buf;
    while ( get_token(&p1, &p2, &len, &colon) ) {
        if ( stricmp(p1, "p0") == 0 ) {
            port= 0;
        } else if ( stricmp(p1, "p1") == 0 ) {
            port= 1;
        } else if ( stricmp(p1, "p2") == 0 ) {
            port= 2;
        } else if ( stricmp(p1, "p3") == 0 ) {
            port= 3;
        } else if ( stricmp(p1, "p4") == 0 ) {
            port= 4;
        } else if ( stricmp(p1, "p5") == 0 ) {
            port= 5;
        } else if ( stricmp(p1, "p6") == 0 ) {
            port= 6;
        } else if ( stricmp(p1, "clear") == 0 || stricmp(p1, "clean") == 0 ) {
            int i;

            for ( i = 0; i < 7; i++ )
                clear_rmon_counter(i);
            break;
        } else if ( stricmp(p1, "help") == 0 || strcmp(p1, "?") == 0 ) {
            printk("echo px[x:0~6] > /proc/driver/ifx_7port_sw/read_rmon_counters\n");
            printk("echo clean > /proc/driver/ifx_7port_sw/read_rmon_counters\n");
            break;
        }
        p1 = p2;
    }
    
    if (port >= 0 || port <= 6)
        print_rmon_counter(port);

    up(&ifx_7port_eth_sema);
    return count;
}
#endif /*SNMP_COUNTERS_DEBUG*/

/** Displays the version of ETH module via proc file */
static int eth_proc_version(IFX_char_t *buf,
                            IFX_char_t **start __attribute__ ((unused)),
                            off_t offset,
                            int count __attribute__ ((unused)),
                            int *eof,
                            void *data __attribute__ ((unused)))
{
    int len = 0;

    *eof = 1;
    /* E.g. cat calls until it receives zero bytes */
    if(offset != 0) {
        return 0;
    }
    /* No sanity check cos length is smaller than one page */
    len += eth_drv_ver(buf + len);
    return len;
}

/** create proc for debug  info, ifx_eth_module_init */
static int eth_proc_create(void)
{
    /* procfs */
    g_ifx_eth_proc_dir = proc_mkdir("driver/ifx_7port_sw",NULL);
    if (g_ifx_eth_proc_dir == NULL) {
        DEB_ERR("%s: Create proc directory (/driver/ifx_7port_sw) failed!!!\n", __func__);
        return IFX_ERROR;
    }
    create_proc_read_entry("version", 0, g_ifx_eth_proc_dir, eth_proc_version,  NULL);
#if defined (SNMP_COUNTERS_DEBUG)
    {
        struct proc_dir_entry *res;
        res = create_proc_entry("read_rmon_counters", 0, g_ifx_eth_proc_dir );
        if ( res ) {
            res->read_proc  = NULL;
            res->write_proc = proc_read_rmon;
        }
    }
#endif /* SNMP_COUNTERS_DEBUG */
    {
        struct proc_dir_entry *res;
        res = create_proc_entry("setup_h3", 0, g_ifx_eth_proc_dir );
        if ( res ) {
            res->read_proc  = proc_read_h3;
            res->write_proc = proc_write_h3;
        }
    }
    {
        struct proc_dir_entry *res;
        res = create_proc_entry("dump_phy_regs", 0, g_ifx_eth_proc_dir );
        if ( res ) {
            res->read_proc  = proc_phy_dump;
            res->write_proc = NULL;
        }
    }
#if defined (CONFIG_RELOAD_PHY_FW)
    {
        struct proc_dir_entry *res;
        res = create_proc_entry("reload_phy_fw", 0, g_ifx_eth_proc_dir );
        if ( res ) {
            res->read_proc  = reload_phy_help;
            res->write_proc = reload_phy_fw;
        }
    }
#endif /*--- #if defined (CONFIG_RELOAD_PHY_FW) ---*/
    return IFX_SUCCESS;
}

/** remove of the proc entries, ifx_eth_module_exit */
static void eth_proc_delete(void) {
    remove_proc_entry("version", g_ifx_eth_proc_dir);
#   if defined (SNMP_COUNTERS_DEBUG)
    remove_proc_entry("read_rmon_counters", g_ifx_eth_proc_dir);
#   endif
    remove_proc_entry("driver/ifx_7port_sw",  NULL);
}

/*------------------------------------------------------------------------------------------*\
* Initilization  Ethernet module
\*------------------------------------------------------------------------------------------*/
#ifdef CONFIG_AVM_PA
static void ifx_eth_pa_dev_transmit(void *arg, struct sk_buff *skb) {
    skb->dev = (struct net_device *) arg;
    dev_queue_xmit(skb);
}
#endif

int ifx_eth_module_init (void) {
    int  err = 0;
    unsigned int phy_nr, net_dev_number;
    unsigned int reg;
    char ver_str[128] = {0};
    IFX_switch_priv_t* priv;
    IFX_ETHSW_portCfg_t Par;
    g_dma_pmac = 0;

    DEB_INFOTRC("[%s] support %d network devices\n", __FUNCTION__, cpmac_global.ports );

    init_MUTEX(&mdio_semaphore);

    init_offload_cfg();

    for(phy_nr = 0; phy_nr < cpmac_global.ports ; phy_nr++) {
        unsigned int mii_nr = get_mii_nr_from_phy_nr(phy_nr);
        ifx_control_mac(mii_nr, 0);  /*--- mac abschalten bevor wir das Netzwerk device erzeugen ---*/
    }

    /** Register with DM core driver */
    init_gpio();
    dma_setup_init();
     /* HW init of the Switch */
    for (phy_nr = 0; phy_nr < cpmac_global.ports ; phy_nr++) {
        reset_phys(phy_nr);
    }

    vr9_7port_sw_hw_init();
    switch_handle = ifx_ethsw_kopen(switch_dev);
    if(switch_handle == (unsigned int) IFX_ERROR) {
        DEB_ERR("[%s] Failed: Could not open switch API\n", __FUNCTION__);
        goto err_out_free_res;
    }
    init_virt_phy_irq_workqueue();

    for(net_dev_number = 0; net_dev_number < cpmac_global.ports ; net_dev_number++) {
        unsigned int phy_nr = get_phy_nr_from_network_index(net_dev_number);
        int mii_nr          = get_mii_nr_from_phy_nr(phy_nr);
        ifx_eth_dev[net_dev_number] = alloc_etherdev(sizeof(IFX_switch_priv_t));
        if (!ifx_eth_dev[net_dev_number]) {
            DEB_ERR("%s[%d]: no memory for ifx_eth_dev!!!\n", __func__,__LINE__);
            err = -ENOMEM;
            goto err_out_free_res;
        }
        ifx_eth_dev[net_dev_number]->netdev_ops = &netdev_ops;
        ifx_eth_dev[net_dev_number]->hard_header_len += 48;
        ifx_eth_dev[net_dev_number]->needed_headroom  = 48;
        ifx_eth_dev[net_dev_number]->tx_queue_len = 64;
        /*--- ifx_eth_dev[net_dev_number]->features = NETIF_F_NO_CSUM; ---*/
        ifx_eth_dev[net_dev_number]->features = 0;
        /*--- ifx_eth_dev[net_dev_number]->features |= NETIF_F_GSO_SOFTWARE; ---*/
        /*--- ifx_eth_dev[net_dev_number]->features |= NETIF_F_GSO_ROBUST; ---*/
        /*--- ifx_eth_dev[net_dev_number]->features |= NETIF_F_SG; ---*/
        /*--- ifx_eth_dev[net_dev_number]->features |= NETIF_F_HIGHDMA; ---*/
        /*--- ifx_eth_dev[net_dev_number]->features |= NETIF_F_FRAGLIST; ---*/
        DEB_ERR("[%s] network device %u has feature mask 0x%08lx\n", __FUNCTION__, net_dev_number, ifx_eth_dev[net_dev_number]->features);

        priv                 = netdev_priv(ifx_eth_dev[net_dev_number]);
        priv->owner          = ifx_eth_dev[net_dev_number];
        priv->phy_addr       = get_mdio_addr_from_phy_nr(phy_nr);
        priv->dev_index      = net_dev_number;
        cpmac_global.instance[net_dev_number].dev = ifx_eth_dev[net_dev_number];
        DEB_ERR("[%s] phy_addr 0x%x device_index 0x%x\n", __FUNCTION__, priv->phy_addr, priv->dev_index);

        /* By default, advertise supported  speed/duplex settings. */
        if(offload_cfg.offload_mac == mii_nr){
            priv->flags |= IFX_FLAG_ADV_1000FULL;
        } else {
            priv->flags |= (IFX_FLAG_ADV_10HALF
                            | IFX_FLAG_ADV_10FULL
                            | IFX_FLAG_ADV_100HALF
                            | IFX_FLAG_ADV_100FULL
                            /*--- | IFX_FLAG_ADV_1000HALF ---*/
                            | IFX_FLAG_ADV_1000FULL
                            );
        }

        switch(phy_nr) {
            case 0:
            case 1:
            case 4:  /*--- TODO: wlan offload ---*/
                break;
            case 2:
            case 3:
                priv->flags |= IFX_FLAG_INTERNAL_PHY;
                break;
        }

        /*--- Init Phys Setup MDIO Setup needed for PHY-FW >= 0x20c ---*/ 
        if(phy_init_port(ifx_eth_dev[net_dev_number])) {
            goto err_out_free_res;
        }

        /* By default, auto-negotiate PAUSE. */
        priv->flags |= IFX_FLAG_PAUSE_AUTO;
        spin_lock_init(&priv->lock);
        memset(&priv->napi, 0, sizeof(priv->napi));
        ifx_control_mac(mii_nr, 0);  /*--- mac abschalten bevor wir das Netzwerk device erzeugen ---*/
        netif_napi_add(ifx_eth_dev[net_dev_number], &priv->napi, switch_poll, 64);
        
        DEB_ERR("%s[%d]: Try to Register network device %s\n", __func__,__LINE__, ifx_eth_dev[net_dev_number]->name);
        if((err = register_netdev(ifx_eth_dev[net_dev_number]))) {
            DEB_ERR("%s[%d]: Register with network device failed!!!\n", __func__,__LINE__);
            goto err_out_free_res;
        }
        /*--- napi_enable(&priv->napi); ---*/

        /*----------------------------------------------------------------------------------*\
         * Port konfigurieren
         \*----------------------------------------------------------------------------------*/
        memset(&Par, 0, sizeof(Par));
        Par.nPortId   = mii_nr;
        Par.eEnable = 1;
        Par.nLearningLimit = 0xFF;
        Par.bAging = 0;
        Par.bUnicastUnknownDrop = 0;
        Par.bMulticastUnknownDrop = 0;
        /*--- Par.bLearningLimitAction = 1; ---*/ /* Forward frames */
        Par.eFlowCtrl = IFX_ETHSW_FLOW_AUTO;
        /*--- Par.ePortMonitor = IFX_ETHSW_PORT_MONITOR_RX; ---*/
        IFX_FLOW_PortCfgSet(avm_switch_api_device, &Par);
        ifx_print_port_config(&Par, __FUNCTION__);

        ifx_eth_dev_check_link_status(phy_nr);
        DEB_ERR("[%s]: Network device '%s' registered\n", __FUNCTION__, ifx_eth_dev[net_dev_number]->name);
#   ifdef CONFIG_AVM_PA
        {
        struct net_device *dev = ifx_eth_dev[net_dev_number];
        {
            struct avm_pa_pid_cfg cfg;
            snprintf(cfg.name, sizeof(cfg.name), "%s", dev->name);
            cfg.framing = avm_pa_framing_ether;
            cfg.default_mtu = 1500;
            cfg.tx_func = ifx_eth_pa_dev_transmit;
            cfg.tx_arg = dev;
            if(avm_pa_dev_pid_register(AVM_PA_DEVINFO(dev), &cfg) < 0)
                printk(KERN_ERR "%s: failed to register PA PID\n", cfg.name);
        }
        {
            struct avm_pa_vpid_cfg cfg;
            snprintf(cfg.name, sizeof(cfg.name), "%s", dev->name);
            cfg.v4_mtu = 1500;
            cfg.v6_mtu = 1500;
            if(avm_pa_dev_vpid_register(AVM_PA_DEVINFO(dev), &cfg) < 0)
                printk(KERN_ERR "%s: failed to register PA VPID\n", cfg.name);
        }
        }
#   endif
    }

    reg = SW_READ_REG32(PMAC_HD_CTL_REG );
    reg &= ~PMAC_HD_CTL_RXSH;
    SW_WRITE_REG32(reg, PMAC_HD_CTL_REG );

    if (reg  & PMAC_HD_CTL_RXSH ) {
        DEB_ERR("[%s] enable g_dma_pmac \n", __FUNCTION__);
        g_dma_pmac = 1;
    } else {
        DEB_ERR("[%s] disabled g_dma_pmac \n", __FUNCTION__);
    }

    /*----------------------------------------------------------------------------------*\
     * CPU Port konfigurieren
    \*----------------------------------------------------------------------------------*/
    DEB_INFOTRC("[%s] config cpu port...\n", __FUNCTION__);

    memset(&Par, 0, sizeof(Par));
    Par.nPortId   = 6;
    Par.eEnable = 1;
    Par.nLearningLimit = 0xFF;
    Par.bAging = 0;
    Par.bUnicastUnknownDrop = 0;
    Par.bMulticastUnknownDrop = 0;
    Par.eFlowCtrl = IFX_ETHSW_FLOW_AUTO;
    Par.ePortMonitor = IFX_ETHSW_PORT_MONITOR_NONE;
    /*--- Par.bLearningLimitAction = 1; ---*/ /* Forward frames */
    IFX_FLOW_PortCfgSet(avm_switch_api_device, &Par);
    ifx_print_port_config(&Par, __FUNCTION__);

    /*--- *((volatile unsigned int *)(0xbe108000 + (4 * 0x453))) = 1 << 6; ---*/


    if(eth_proc_create() == IFX_ERROR )
        goto err_out_free_res;
 /* Print the driver version info */
    eth_drv_ver(ver_str);
    DEB_INFOTRC("[%s] Driver version %s", __FUNCTION__, ver_str);

    return  0;

err_out_free_res:
/*Unregister with DMA core driver */
    dma_setup_uninit();
    /* unregister the network devices */
    for (net_dev_number = 0; net_dev_number < cpmac_global.ports ; net_dev_number++) {
        if(ifx_eth_dev[net_dev_number])
            free_netdev(ifx_eth_dev[net_dev_number]);
    }
    ifx_ethsw_kclose(switch_handle);

    return err;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void  ifx_eth_module_exit (void) {
    int net_dev_number;

    /*Unregister with DMA core driver */
    dma_setup_uninit();
    /* unregister the network devices */
    for (net_dev_number = 0; net_dev_number < cpmac_global.ports ; net_dev_number++) {
        unregister_netdev(ifx_eth_dev[net_dev_number]);
        free_netdev(ifx_eth_dev[net_dev_number]);
    }
    /** remove of the proc entries */
    eth_proc_delete();
}

late_initcall(ifx_eth_module_init);
module_exit(ifx_eth_module_exit);

MODULE_AUTHOR("Reddy Mallikarjuna");
MODULE_DESCRIPTION("IFX ethernet driver (Supported VR9)");
MODULE_LICENSE("GPL");
MODULE_VERSION(IFX_DRV_MODULE_VERSION);

